using System;
using Microsoft.DirectX.Direct3D;
using TexCache=System.Collections.Generic.Dictionary<string, Microsoft.DirectX.Direct3D.Texture>;

namespace MGEgui.DirectX {
    class LTEX {
        private static readonly TexCache texCache=new TexCache();

        public static void ReleaseCache() {
            foreach(Texture t in texCache.Values) { try { if(!t.Disposed) t.Dispose(); } catch { } }
            texCache.Clear();
        }

        private string filePath;
        public string FilePath {
            set { filePath=value.ToLower(); }
            get { return filePath; }
        }
        public int index;
        public Texture tex;

        public void LoadTexture() {
            if(texCache.ContainsKey(filePath)) {
                tex=texCache[filePath];
            } else {
                try {
                    byte[] data=MGEgui.DistantLand.BSA.GetTexture(filePath);
                    tex=TextureLoader.FromStream(DXMain.device, new System.IO.MemoryStream(data));
                    texCache[filePath]=tex;
                } catch {
                    tex=null;
                    throw;
                }
            }
        }
    }

    class StaticTexCreator {
        private readonly System.Collections.Generic.List<string> texCache;
        private readonly int div;

        public StaticTexCreator(int skip) {
            div=(1<<skip);
            texCache=new System.Collections.Generic.List<string>();
        }

        public void Dispose() {
            texCache.Clear();
        }

        public void LoadTexture(string path) {
            if(texCache.Contains(path)) return;
            texCache.Add(path);

            byte[] data=MGEgui.DistantLand.BSA.GetTexture(path);
            System.IO.MemoryStream ms=new System.IO.MemoryStream(data);
            path=System.IO.Path.ChangeExtension(path, ".dds");
            
            Texture t=TextureLoader.FromStream(DXMain.device, ms, 0, 0, 0, Usage.None, Format.Unknown, Pool.Scratch, Filter.Triangle|Filter.Dither, Filter.Box, 0);
            SurfaceDescription sd=t.GetLevelDescription(0);
            t.Dispose();

            Format format;
            if(sd.Format==Format.Dxt1||sd.Format==Format.X8R8G8B8) format=Format.Dxt1;
            else format=Format.Dxt3;

            ms.Position=0;
            System.IO.Directory.CreateDirectory(System.IO.Path.GetDirectoryName(System.IO.Path.Combine(@"data files\distantland\statics\textures\", path)));
            t=TextureLoader.FromStream(DXMain.device, ms, sd.Width/div, sd.Height/div, 0, Usage.None, format, Pool.Scratch, Filter.Triangle|Filter.Dither, Filter.Box, 0);
            TextureLoader.Save(System.IO.Path.Combine(@"data files\distantland\statics\textures\", path), ImageFileFormat.Dds, t);
            t.Dispose();

            ms.Close();
        }
    }

    class CellTexCreator {
        private struct CellVertex {
        public float x, y, z, w;
        public float u1, v1, u2, v2, u3, v3;

        public void SetTex(int x2, int y2, byte pos) {
            u1=(float)x2/4.0f;//+0.25
            v1=(float)y2/4.0f;
            if(u1>1) u1-=1;
            if(u1==1||u1==0) if(pos==0||pos==2) u1=0; else u1=1;
            u2=(float)x2/4.0f + 0.25f;//0.5
            v2=(float)y2/4.0f + 0.75f;//0.75
            if(u2>1) u2-=1;
            if(v2>1) v2-=1;
            if(u2==1||u2==0) if(pos==0||pos==2) u2=0; else u2=1;
            if(v2==1||v2==0) if(pos==0||pos==1) v2=0; else v2=1;
            switch(pos) {
                case 0:
                    u3=v3=0;
                    break;
                case 1:
                    u3=1;
                    v3=0;
                    break;
                case 2:
                    u3=0;
                    v3=1;
                    break;
                case 3:
                    u3=v3=1;
                    break;
            }
        }

        public const VertexFormats Format=VertexFormats.Transformed|VertexFormats.Texture3;
        public const int Stride=40;
    }

        private const string EffectPath=@"data files\shaders\distantland\CellTexBlend.fx";
        private VertexBuffer[,] vBuffers;
        private Texture RenderTargetTex;
        private Texture CompressedTex;
        private Surface RenderTarget;
        private Effect effect;
        private bool vcolor;

        private EffectHandle t1h;
        private EffectHandle t2h;
        private EffectHandle t3h;
        private EffectHandle t4h;
        private EffectHandle c1h;
        private EffectHandle c2h;
        private EffectHandle c3h;
        private EffectHandle c4h;

        public CellTexCreator(int Res, bool vColor) {
            vBuffers=new VertexBuffer[64, 64];
            for(int x=0;x<64;x++) {
                for(int y=0;y<64;y++) {
                    vBuffers[x,y]=new VertexBuffer(typeof(CellVertex), 4, DXMain.device, Usage.WriteOnly, CellVertex.Format, Pool.Managed);
                    CellVertex[] CellData=(CellVertex[])vBuffers[x, y].Lock(0, LockFlags.None);
                    int x2=x%4;
                    int y2=y%4;
                    CellData[0].SetTex(x2, y2, 0);
                    CellData[1].SetTex(x2+1, y2, 1);
                    CellData[2].SetTex(x2, y2+1, 2);
                    CellData[3].SetTex(x2+1, y2+1, 3);
                    float mult=(float)(Res/64);
                    CellData[0].x=x*mult;
                    CellData[0].y=y*mult;
                    CellData[0].z=0.5f;
                    CellData[0].w=1;
                    CellData[1].x=(x+1)*mult;
                    CellData[1].y=y*mult;
                    CellData[1].z=0.5f;
                    CellData[1].w=1;
                    CellData[2].x=x*mult;
                    CellData[2].y=(y+1)*mult;
                    CellData[2].z=0.5f;
                    CellData[2].w=1;
                    CellData[3].x=(x+1)*mult;
                    CellData[3].y=(y+1)*mult;
                    CellData[3].z=0.5f;
                    CellData[3].w=1;
                    vBuffers[x, y].Unlock();
                }
            }
            RenderTargetTex=new Texture(DXMain.device, Res, Res, 0, Usage.RenderTarget, Format.X8R8G8B8, Pool.Default);
            CompressedTex=new Texture(DXMain.device, Res, Res, 0, Usage.None, Format.Dxt1, Pool.SystemMemory);
            RenderTarget=RenderTargetTex.GetSurfaceLevel(0);
            effect=Effect.FromFile(DXMain.device, EffectPath, null, null, ShaderFlags.NotCloneable, null);

            t1h=effect.GetParameter(null, "t1");
            t2h=effect.GetParameter(null, "t2");
            t3h=effect.GetParameter(null, "t3");
            t4h=effect.GetParameter(null, "t4");
            if(!vColor) {
                ColorValue white=ColorValue.FromColor(System.Drawing.Color.White);
                effect.SetValue("c1", white);
                effect.SetValue("c2", white);
                effect.SetValue("c3", white);
                effect.SetValue("c4", white);
                vcolor=false;
            } else {
                c1h=effect.GetParameter(null, "c1");
                c2h=effect.GetParameter(null, "c2");
                c3h=effect.GetParameter(null, "c3");
                c4h=effect.GetParameter(null, "c4");
                vcolor=true;
            }

            DXMain.device.RenderState.CullMode=Cull.CounterClockwise;
            //DXMain.device.RenderState.CullMode=Cull.None;
            DXMain.device.RenderState.Clipping=false;
            DXMain.device.SetRenderTarget(0, RenderTarget);
            DXMain.device.VertexFormat=CellVertex.Format;
        }

        public void Dispose() {
            DXMain.device.SetRenderTarget(0, DXMain.BackBuffer);
            for(int x=0;x<64;x++) {
                for(int y=0;y<64;y++) {
                    vBuffers[x, y].Dispose();
                }
            }
            RenderTarget.Dispose();
            RenderTargetTex.Dispose();
            CompressedTex.Dispose();
            effect.Dispose();
        }

        public void Begin() {
            DXMain.device.BeginScene();
            effect.Begin(FX.None);
            effect.BeginPass(0);
        }

        public void RenderStep(LTEX t1, LTEX t2, LTEX t3, LTEX t4, int c1, int c2, int c3, int c4, int x, int y) {
            effect.SetValue(t1h, t1.tex);
            effect.SetValue(t2h, t2.tex);
            effect.SetValue(t3h, t3.tex);
            effect.SetValue(t4h, t4.tex);
            if(vcolor) {
                effect.SetValue(c1h, ColorValue.FromArgb(c1));
                effect.SetValue(c2h, ColorValue.FromArgb(c2));
                effect.SetValue(c3h, ColorValue.FromArgb(c3));
                effect.SetValue(c4h, ColorValue.FromArgb(c4));
            }
            effect.CommitChanges();
            DXMain.device.SetStreamSource(0, vBuffers[x,y], 0, CellVertex.Stride);
            DXMain.device.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 2);
        }

        public void End(string path) {
            effect.EndPass();
            effect.End();
            DXMain.device.EndScene();

            Surface s1=CompressedTex.GetSurfaceLevel(0);
            SurfaceLoader.FromSurface(s1, RenderTarget, Filter.None, 0);
            for(int i=1;i<CompressedTex.LevelCount;i++) {
                Surface s2=CompressedTex.GetSurfaceLevel(i);
                SurfaceLoader.FromSurface(s2, s1, Filter.Linear, 0);
                s1=s2;
            }
            TextureLoader.Save(path, ImageFileFormat.Dds, CompressedTex);
        }
    }

    class WorldTexCreator {
        private const string DefaultTex=@"data files\distantland\default.dds";

        private Texture CompressedTex;
        private Surface RenderTarget;
        private Surface DefaultSurf;

        private readonly System.Drawing.Rectangle cellRect;
        private Filter surfFilter;

        private readonly int[] Offsets;

        public WorldTexCreator(int Res, int Cells, int CellRes, bool filter) {
            float frac=(float)Res/(float)Cells;
            cellRect=new System.Drawing.Rectangle(0, 0, CellRes, CellRes);

            if(filter) surfFilter=Filter.Triangle|Filter.Dither; else surfFilter=Filter.Point;
            
            CompressedTex=new Texture(DXMain.device, Res, Res, 0, Usage.None, Format.Dxt1, Pool.Scratch);
            RenderTarget=DXMain.device.CreateOffscreenPlainSurface(Res, Res, Format.X8R8G8B8, Pool.Scratch);
            DefaultSurf=DXMain.device.CreateOffscreenPlainSurface(CellRes, CellRes, Format.X8R8G8B8, Pool.Scratch);
            SurfaceLoader.FromFile(DefaultSurf, DefaultTex, Filter.None, 0);

            Offsets=new int[Cells+1];
            for(int i=0;i<Cells;i++) Offsets[i]=(int)(i*frac);
            Offsets[Cells]=Res-1;
        }

        public void Render(string path, int x, int y) {
            System.Drawing.Rectangle rect=new System.Drawing.Rectangle();
            rect.X=Offsets[x];
            rect.Y=Offsets[y];
            rect.Width=Offsets[x+1]-Offsets[x]+1;
            rect.Height=Offsets[y+1]-Offsets[y]+1;

            if(path!=null) {
                SurfaceLoader.FromFile(RenderTarget, rect, path, cellRect, surfFilter, 0);
            } else {
                SurfaceLoader.FromSurface(RenderTarget, rect, DefaultSurf, cellRect, surfFilter, 0);
            }
        }

        public void Finish(string path) {
            Surface output=CompressedTex.GetSurfaceLevel(0);
            SurfaceLoader.FromSurface(output, RenderTarget, Filter.None, 0);
            TextureLoader.FilterTexture(CompressedTex, 0, Filter.Box);
            TextureLoader.Save(path, ImageFileFormat.Dds, CompressedTex);

            DefaultSurf.Dispose();
            RenderTarget.Dispose();
            output.Dispose();
            CompressedTex.Dispose();
        }
    }
}
