//#define DEBUGSTATICS

using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.IO;
using MGEgui.DirectX;

namespace MGEgui.DistantLand {
    public partial class DistantLandForm : Form {
        private static void DumpError(Exception ex) {
            string str=ex.ToString();
            while((ex=ex.InnerException)!=null) {
                str+=Environment.NewLine+Environment.NewLine+str.ToString();
            }
            File.WriteAllText("mge3\\distant-land-crash-dump.txt", str);
        }

        public DistantLandForm(bool exists) {
            Exists=exists;
            if(Exists) StaticsExist=File.Exists(@"data files\distantland\statics\usage.data");

            InitializeComponent();
            Directory.CreateDirectory("data files\\distantland");
            statusText.Text="Initializing BSAs";
            statusProgress.Style = ProgressBarStyle.Marquee;
            backgroundWorker.DoWork+=workerInitBSAs;
            backgroundWorker.RunWorkerCompleted+=workerFInitBSAs;
            statusWarnings.Enabled=false;

            cmbCellTexResolution.SelectedIndex=2;
            cmbWorldTexResolution.SelectedIndex=4;
            lMaxTexSize.Text+=DXMain.mCaps.MaxTexSize.ToString();

            DXMain.CreateDevice(this);
            LTEX.ReleaseCache();

            backgroundWorker.RunWorkerAsync();
        }

        private bool ChangingPage;
        private bool Exists;
        private bool StaticsExist;

        private int MapSize;
        private int CellCount;
        public int VeryHigh=int.MinValue;
        public int VeryLow=int.MaxValue;

        private List<string> warnings;

        private LAND[,] map;
        private readonly LAND defaultland=new LAND();

        private LTEX DefaultTex;

        private string[] files;

        private static string ReadString(BinaryReader br) {
            return new string(br.ReadChars(4));
        }
        private static string ReadCString(BinaryReader br) {
            List<char> chars=new List<char>();
            while(true) {
                char c=br.ReadChar();
                if(c=='\0') return new string(chars.ToArray());
                else chars.Add(c);
            }
        }

        private void statusWarnings_ButtonClick(object sender, EventArgs e) {
            try {
                File.WriteAllLines("mge3\\distantlandlog.txt", warnings.ToArray());
                System.Diagnostics.Process p=System.Diagnostics.Process.Start("mge3\\distantlandlog.txt");
                p.Close();
            } catch {
                MessageBox.Show(string.Join("\n", warnings.ToArray()));
            }
        }
        private void UpdateProgress(object sender, System.ComponentModel.ProgressChangedEventArgs e) {
            statusProgress.Value=e.ProgressPercentage;
            if(e.UserState!=null) {
                statusText.Text=(string)e.UserState;
            }
        }

        private void tabControl_Selecting(object sender, TabControlCancelEventArgs e) {
            if(ChangingPage) {
                ChangingPage=false;
            } else {
                e.Cancel=true;
            }
        }
        private void IgnoreKeyPress(object sender, KeyPressEventArgs e) {
            e.Handled=true;
        }
        private void DistantLandForm_FormClosing(object sender, FormClosingEventArgs e) {
            if(!ChangingPage) {
                if(backgroundWorker.IsBusy) {
                    e.Cancel=true;
                    MessageBox.Show("Please wait for the current stage of distant land generation to complete before exiting", "Error");
                } else {
                    if(MessageBox.Show("Distant land generation is not complete.\n"+
                    "Exiting now will cause you to loose any files created so far.\n"+
                    "Are you sure you wish to exit?", "Warning", MessageBoxButtons.YesNo)!=DialogResult.Yes) e.Cancel=true;
                }
            }
        }

        private void workerInitBSAs(object sender, System.ComponentModel.DoWorkEventArgs e) {
            BSA.InitBSAs();
        }
        private void workerFInitBSAs(object sender, System.ComponentModel.RunWorkerCompletedEventArgs e) {
            if(e.Error!=null) {
                DumpError(e.Error);
                MessageBox.Show("A fatal error occured loading the BSA's. Distant land setup cannot continue\n"+
                    e.Error.Message);
                ChangingPage=true;
                Close();
                return;
            }
            if(e.Result!=null) {
                warnings=(List<string>)e.Result;
                statusWarnings.Enabled=true;
            }
            statusText.Text="BSA's loaded.";
            statusProgress.Style = ProgressBarStyle.Continuous;
            backgroundWorker.DoWork-=workerInitBSAs;
            backgroundWorker.RunWorkerCompleted-=workerFInitBSAs;
            bRunStep1.Enabled=true;
        }

        private void bLoadModsFromINI_Click(object sender, EventArgs e) {
            StreamReader sr=new StreamReader("morrowind.ini", System.Text.Encoding.Default);
            bool InModList=false;
            List<string> ActivePlugins=new List<string>();
            while(!sr.EndOfStream) {
                string s=sr.ReadLine().Trim().ToLower();
                if(s.Length==0) continue;
                if(!InModList) {
                    if(s=="[game files]") InModList=true;
                } else {
                    if(s[0]=='['&&s.EndsWith("]")) {
                        InModList=false;
                        continue;
                    }
                    if(s.StartsWith("gamefile")) {
                        int index=s.IndexOf(';');
                        if(index!=-1) s=s.Remove(index, s.Length-index);
                        index=s.IndexOf('=');
                        if(index!=-1) {
                            s=s.Remove(0, index+1).Trim();
                            ActivePlugins.Add("data files\\"+s);
                        }
                    }
                }
            }
            sr.Close();
            lbModList.Items.Clear();
            foreach(string s in ActivePlugins) {
                lbModList.Items.Add(Path.GetFullPath(s).ToLower());
            }
        }

        private void bAddMod_Click(object sender, EventArgs e) {
            if(openFileDialog.ShowDialog()==DialogResult.OK) {
                foreach(string s in openFileDialog.FileNames) lbModList.Items.Add(Path.GetFullPath(s).ToLower());
            }
        }

        private void bRemoveMod_Click(object sender, EventArgs e) {
            while(lbModList.SelectedIndices.Count>0) lbModList.Items.RemoveAt(lbModList.SelectedIndices[0]);
        }

        private static bool ComparePlugins(FileInfo fi1, FileInfo fi2) {
            string ext1=fi1.Extension.ToLower();
            string ext2=fi2.Extension.ToLower();
            if(ext1!=ext2) {
                if(ext1==".esm") return false; else return true;
            }
            if(fi1.LastWriteTime>fi2.LastWriteTime) return true; else return false;
        }
        private void bRunStep1_Click(object sender, EventArgs e) {
            //Check that at least one mod is selected
            if(lbModList.Items.Count==0) {
                MessageBox.Show("You haven't selected any plugins", "Error");
                return;
            }
            //Create list of files
            FileInfo[] fi=new FileInfo[lbModList.Items.Count];
            files=new string[lbModList.Items.Count];
            for(int i=0;i<lbModList.Items.Count;i++) {
                files[i]=((string)lbModList.Items[i]).ToLower();
                fi[i]=new FileInfo(files[i]);
            }
            //Check for non-plugins
            for(int i=0;i<files.Length;i++) {
                string ext=Path.GetExtension(files[i]);
                if(ext!=".esm"&&ext!=".esp") {
                    MessageBox.Show("Plugin '"+lbModList.Items[i]+"' is not a valid plugin", "Error");
                    return;
                }
            }
            //Check for duplicate file paths
            for(int i=0;i<files.Length-1;i++) {
                for(int j=i+1;j<files.Length;j++) {
                    if(files[i]==files[j]) {
                        MessageBox.Show("Plugin '"+lbModList.Items[i]+"' has been added twice", "Error");
                        return;
                    }
                }
            }
            //Sort into load order
            bool b=true;
            while(b) {
                b=false;
                for(int i=0;i<files.Length-1;i++) {
                    if(ComparePlugins(fi[i], fi[i+1])) {
                        string s=files[i];
                        files[i]=files[i+1];
                        files[i+1]=s;
                        FileInfo f=fi[i];
                        fi[i]=fi[i+1];
                        fi[i+1]=f;
                        b=true;
                    }
                }
            }
            //Continue
            ChangingPage=true;
            tabControl.SelectedIndex=1;
            statusText.Text="Loading plugin land data";
            statusProgress.Maximum=lbModList.Items.Count;
            backgroundWorker.WorkerReportsProgress=true;
            backgroundWorker.DoWork+=workerLoadPlugins;
            backgroundWorker.RunWorkerCompleted+=workerFLoadPlugins;
            statusWarnings.Enabled=false;

            backgroundWorker.RunWorkerAsync();
        }

        private void workerLoadPlugins(object sender, System.ComponentModel.DoWorkEventArgs e) {
            map=(LAND[,])Array.CreateInstance(typeof(LAND), new int[] { 501, 501 }, new int[] { -250, -250 });
            
            DefaultTex=new LTEX();
            DefaultTex.FilePath="_land_default.tga";
            DefaultTex.index=0;
            try {
                DefaultTex.LoadTexture();
            } catch(Exception ex) {
                throw new ApplicationException("Failed to load default ground texture ("+ex.Message+")",ex);
            }

            CellCount=0;

            int progress=0;

            List<string> errors=new List<string>();
            foreach(string file in files) {
                BinaryReader br=new BinaryReader(File.OpenRead(file), System.Text.Encoding.Default);
                List<LTEX> Textures=new List<LTEX>();
                //Add the default texture first, because all vtex id's are +1 compared to the ltex ids
                Textures.Add(DefaultTex);
                string s="";
                while(br.BaseStream.Position<br.BaseStream.Length) {
                    s=ReadString(br);
                    if(s=="LAND") {
                        int size=br.ReadInt32();
                        int read=0;
                        br.BaseStream.Position+=8;
                        LAND land=new LAND();
                        land.Textures=Textures;
                        int lx=-999;
                        int ly=-999;
                        while(read<size) {
                            s=ReadString(br);
                            int size2=br.ReadInt32();
                            read+=size2+8;
                            switch(s) {
                                case "INTV":
                                    lx=br.ReadInt32();
                                    ly=br.ReadInt32();
                                    land.xpos=lx;
                                    land.ypos=ly;
                                    break;
                                case "VHGT":
                                    int offset=(int)br.ReadSingle();
                                    for(int y=0;y<65;y++) {
                                        offset+=br.ReadSByte();
                                        land.Heights[0, y]=offset;
                                        int pos=offset;
                                        for(int x=1;x<65;x++) {
                                            pos+=br.ReadSByte();
                                            land.Heights[x, y]=pos;
                                            if(pos>VeryHigh) VeryHigh=pos;
                                            if(pos<VeryLow) VeryLow=pos;
                                        }
                                    }
                                    br.BaseStream.Seek(3, SeekOrigin.Current);
                                    break;
                                case "VNML":
                                    for(int y=0;y<65;y++) {
                                        for(int x=0;x<65;x++) {
                                            sbyte vx, vy, vz;
                                            vx=br.ReadSByte();
                                            vy=br.ReadSByte();
                                            vz=br.ReadSByte();
                                            land.Normals[x, y]=new Normal(vx, vy, vz);
                                        }
                                    }
                                    break;
                                case "VCLR":
                                    for(int y=0;y<65;y++) {
                                        for(int x=0;x<65;x++) {
                                            land.Color[x, y]=new RGB(br.ReadByte(), br.ReadByte(), br.ReadByte());
                                        }
                                    }
                                    break;
                                case "VTEX":
                                    for(int y1=0;y1<4;y1++) {
                                        for(int x1=0;x1<4;x1++) {
                                            for(int y2=0;y2<4;y2++) {
                                                for(int x2=0;x2<4;x2++) {
                                                    land.Tex[x1*4+x2, y1*4+y2]=br.ReadInt16();
                                                }
                                            }
                                        }
                                    }
                                    break;
                                default:
                                    br.BaseStream.Position+=size2;
                                    break;
                            }
                        }
                        if(lx!=-999&&ly!=-999) {
                            int max=Math.Max(Math.Abs(lx), Math.Abs(ly));
                            if(max>MapSize) MapSize=max;
                            if(map[lx, ly]==null) CellCount++;
                            map[lx, ly]=land;
                        }
                    } else if(s=="LTEX") {
                        //TODO: Fix!
                        //Presumably, not all LTEX's have the subrecords in an identical order
                        LTEX tex=new LTEX();
                        int size=br.ReadInt32();
                        long pos=br.BaseStream.Position+8;
                        br.BaseStream.Position+=16;
                        ReadCString(br);
                        br.BaseStream.Position+=8;
                        tex.index=br.ReadInt32()+1;
                        br.BaseStream.Position+=8;
                        tex.FilePath=ReadCString(br).ToLower();

                        try {
                            tex.LoadTexture();
                        } catch(Exception ex) {
                            errors.Add("Texture '"+tex.FilePath+"' in plugin '"+file+"' failed to load ("+ex.Message+")");
                            tex.tex=DefaultTex.tex;
                        }
                        Textures.Add(tex);
                        br.BaseStream.Position=pos+size;
                    } else {
                        int size=br.ReadInt32();
                        br.BaseStream.Position+=size+8;
                    }
                }
                br.Close();
                Textures.Sort(new LTEXSorter());
                backgroundWorker.ReportProgress(++progress);
            }
        }
        private void workerFLoadPlugins(object sender, System.ComponentModel.RunWorkerCompletedEventArgs e) {
            if(e.Error!=null) {
                DumpError(e.Error);
                MessageBox.Show("A fatal error occured loading land data from your plugins. Distant land setup cannot continue\n"+
                    e.Error.Message);
                ChangingPage=true;
                Close();
                return;
            }
            statusProgress.Value=0;
            if(e.Result!=null) {
                warnings=(List<string>)e.Result;
                statusWarnings.Enabled=true;
            }
            statusText.Text="Plugin land data loaded";
            backgroundWorker.DoWork-=workerLoadPlugins;
            backgroundWorker.RunWorkerCompleted-=workerFLoadPlugins;

            bSkipStep2.Enabled=true;
            bRunStep2.Enabled=true;
        }

        private void bSkipStep2_Click(object sender, EventArgs e) {
            if(!Exists) {
                MessageBox.Show("Landscape texture and mesh generation can only be skipped if the files already exist.", "Error");
                return;
            }

            LTEX.ReleaseCache();
            ChangingPage=true;
            tabControl.SelectedIndex=3;
            statusText.Text="Landscape texture and mesh generation skipped";
            statusWarnings.Enabled=false;

            workerFCreateMeshes(null, null);
        }
        private void bRunStep2_Click(object sender, EventArgs e) {
            if(128 << Math.Max(cmbCellTexResolution.SelectedIndex, cmbWorldTexResolution.SelectedIndex) > DXMain.mCaps.MaxTexSize) {
                MessageBox.Show("You have selected a texture resolution above that which your graphics card supports.", "Error");
                return;
            }
            if(Exists) {
                if(MessageBox.Show("Landscape textures and meshes already exist. If you recreate them, the existing files will be deleted.\n"+
                    "Do you wish to continue?", "Warning", MessageBoxButtons.YesNo)!=DialogResult.Yes) return;
                foreach(string s in Directory.GetFiles(@"data files\distantland")) File.Delete(s);
            }

            //Set max mesh detail here
            int VertsRequired=(MapSize*2+1)*(MapSize*2+1)*64*64 + (MapSize*2+1)*128+1;
            int PrimsRequired=(MapSize*2+1)*(MapSize*2+1)*64*64*2;
            int index=0;
            while(VertsRequired>DXMain.mCaps.MaxIndicies||PrimsRequired>DXMain.mCaps.MaxPrimitives) {
                VertsRequired/=4;
                PrimsRequired/=4;
                index++;
            }
            lMeshDetail.Text+=(string)cmbWorldMeshDetail.Items[index];
            cmbWorldMeshDetail.SelectedIndex=index;

            ChangingPage=true;
            tabControl.SelectedIndex=2;
            statusText.Text="Creating cell textures";
            statusProgress.Maximum=(MapSize*2 + 1)*(MapSize*2 + 2);
            backgroundWorker.WorkerReportsProgress=true;
            backgroundWorker.DoWork+=workerCreateTextures;
            backgroundWorker.RunWorkerCompleted+=workerFCreateTextures;
            statusWarnings.Enabled=false;

            CreateTextureArgs args=new CreateTextureArgs();
            args.CellRes=128 << cmbCellTexResolution.SelectedIndex;
            args.WorldRes=128 << cmbWorldTexResolution.SelectedIndex;
            args.UseVertexColour=cbVColour.Checked;
            args.filter=cbFilter.Checked;

            backgroundWorker.RunWorkerAsync(args);
        }

        private static void ModCell(ref int cell, ref int tex) {
            if(tex<0) { cell--; tex+=16; } else if(tex>=16) { cell++; tex-=16; }
        }
        private LTEX GetTex(int cellx, int celly, int texx, int texy) {
            if(map[cellx, celly]!=null)
                return map[cellx, celly].Textures[map[cellx, celly].Tex[texx, texy]];
            else return DefaultTex;
        }
        private void GenerateTexture(CellTexCreator ctc, LAND land, string path, bool vColor) {
            ctc.Begin();
            if(land==defaultland) {
                int white=System.Drawing.Color.White.ToArgb();
                for(int y = 0;y < 64;y++) {
                    for(int x = 0;x < 64;x++) {
                        ctc.RenderStep(DefaultTex, DefaultTex, DefaultTex, DefaultTex, white, white, white, white, x, y);
                    }
                }
            } else if(vColor) {
                for(int y = 0;y < 64;y++) {
                    for(int x = 0;x < 64;x++) {
                        int top    = (int)Math.Floor(((float)y-1.0f)/4.0f);//-2.0f
                        int left   = (int)Math.Floor(((float)x-3.0f)/4.0f);//-2.0
                        int bottom = top+1;
                        int right  = left+1;
                        LTEX t1, t2, t3, t4;
                        int c1, c2, c3, c4;

                        int leftc=land.xpos, rightc=land.xpos;
                        int topc=land.ypos, bottomc=land.ypos;
                        ModCell(ref leftc, ref left);
                        ModCell(ref rightc, ref right);
                        ModCell(ref topc, ref top);
                        ModCell(ref bottomc, ref bottom);
                        t1=GetTex(leftc, topc, left, top);
                        t2=GetTex(rightc, topc, right, top);
                        t3=GetTex(leftc, bottomc, left, bottom);
                        t4=GetTex(rightc, bottomc, right, bottom);

                        c1=land.Color[x, y];
                        c2=land.Color[x+1, y];
                        c3=land.Color[x, y+1];
                        c4=land.Color[x+1, y+1];

                        ctc.RenderStep(t1, t2, t3, t4, c1, c2, c3, c4, x, y);
                    }
                }
            } else {
                int white=System.Drawing.Color.White.ToArgb();
                for(int y = 0;y < 64;y++) {
                    for(int x = 0;x < 64;x++) {
                        int top    = (int)Math.Floor(((float)y-1.0f)/4.0f);//-2.0f
                        int left   = (int)Math.Floor(((float)x-3.0f)/4.0f);//-2.0
                        int bottom = top+1;
                        int right  = left+1;
                        LTEX t1, t2, t3, t4;

                        int leftc=land.xpos, rightc=land.xpos;
                        int topc=land.ypos, bottomc=land.ypos;
                        ModCell(ref leftc, ref left);
                        ModCell(ref rightc, ref right);
                        ModCell(ref topc, ref top);
                        ModCell(ref bottomc, ref bottom);
                        t1=GetTex(leftc, topc, left, top);
                        t2=GetTex(rightc, topc, right, top);
                        t3=GetTex(leftc, bottomc, left, bottom);
                        t4=GetTex(rightc, bottomc, right, bottom);

                        ctc.RenderStep(t1, t2, t3, t4, white, white, white, white, x, y);
                    }
                }
            }
            ctc.End(path);
        }

        private struct CreateTextureArgs {
            public int CellRes;
            public int WorldRes;
            public bool UseVertexColour;
            public bool filter;
        }
        void workerCreateTextures(object sender, System.ComponentModel.DoWorkEventArgs e) {
            CreateTextureArgs args=(CreateTextureArgs)e.Argument;

            DirectX.CellTexCreator ctc=new CellTexCreator(args.CellRes, args.UseVertexColour);

            GenerateTexture(ctc, defaultland, @"data files\distantland\default.dds", args.UseVertexColour);

            int count=0;
            for(int y=-MapSize;y<=MapSize;y++) {
                for(int x=-MapSize;x<=MapSize;x++) {
                    count++;
                    if(map[x, y]!=null) {
                        backgroundWorker.ReportProgress(count);
                        GenerateTexture(ctc, map[x, y], "data files\\distantland\\__"+x+"_"+y+".dds", args.UseVertexColour);
                    }
                }
            }
            ctc.Dispose();
            LTEX.ReleaseCache();

            backgroundWorker.ReportProgress(count, "Creating world texture");

            WorldTexCreator wtc=new WorldTexCreator(args.WorldRes, MapSize*2+1, args.CellRes, args.filter);
            for(int y=-MapSize;y<=MapSize;y++) {
                backgroundWorker.ReportProgress(++count);
                for(int x=-MapSize;x<=MapSize;x++) {
                    if(File.Exists("data files\\distantland\\__"+x+"_"+y+".dds")) {
                        wtc.Render("data files\\distantland\\__"+x+"_"+y+".dds", x+MapSize,y+MapSize);
                    } else {
                        wtc.Render(null, x+MapSize,y+MapSize);
                    }
                }
            }
            wtc.Finish(@"data files\distantland\world.dds");
        }
        void workerFCreateTextures(object sender, System.ComponentModel.RunWorkerCompletedEventArgs e) {
            if(e.Error!=null) {
                DumpError(e.Error);
                MessageBox.Show("A fatal error occured creating the landscape textures. Distant land setup cannot continue\n"+
                    e.Error.Message);
                ChangingPage=true;
                Close();
                return;
            }
            statusProgress.Value=0;
            if(e.Result!=null) {
                warnings=(List<string>)e.Result;
                statusWarnings.Enabled=true;
            }
            statusText.Text="Landscape textures created";
            backgroundWorker.DoWork-=workerCreateTextures;
            backgroundWorker.RunWorkerCompleted-=workerFCreateTextures;
            bRunStep3.Enabled=true;
        }

        private void bRunStep3_Click(object sender, EventArgs e) {
            int VertsRequired=(MapSize*2+1)*(MapSize*2+1)*64*64 + (MapSize*2+1)*128+1;
            int PrimsRequired=(MapSize*2+1)*(MapSize*2+1)*64*64*2;
            int index=0;
            while(VertsRequired>DXMain.mCaps.MaxIndicies||PrimsRequired>DXMain.mCaps.MaxPrimitives) {
                VertsRequired/=4;
                PrimsRequired/=4;
                index++;
            }
            if(cmbWorldMeshDetail.SelectedIndex<index) {
                if(MessageBox.Show("You have selected a mesh detail above that which your graphics card apparently supports.\n"+
                    "If you continue, morrowind may crash, or the distant landscape may not appear.\n"+
                    "If this happens, simply recreate the land mesh at a lower detail.\n"+
                    "Are you sure you wish to continue?", "Warning", MessageBoxButtons.YesNo)!=DialogResult.Yes) return;
            }

            ChangingPage=true;
            tabControl.SelectedIndex=3;
            statusText.Text="Creating cell meshes";
            statusProgress.Maximum=(MapSize*2 + 1)*2+1;
            backgroundWorker.WorkerReportsProgress=true;
            backgroundWorker.DoWork+=workerCreateMeshes;
            backgroundWorker.RunWorkerCompleted+=workerFCreateMeshes;
            statusWarnings.Enabled=false;

            CreateMeshArgs cma=new CreateMeshArgs();
            cma.AverageHeights=cbAverageVertexHeights.Checked;
            cma.MeshDetail=cmbWorldMeshDetail.SelectedIndex;

            backgroundWorker.RunWorkerAsync(cma);
        }

        private static void GenerateMesh(LAND land, string path) {
            Directory.CreateDirectory(Path.GetDirectoryName(path));
            BinaryWriter bw=new BinaryWriter(File.Create(path));
            for(int x=0;x<65;x++) {
                for(int y=0;y<65;y++) {
                    bw.Write((float)(x*128));
                    bw.Write((float)(y*128));
                    bw.Write((float)(land.Heights[x, y]*8));
                    bw.Write(land.Normals[x, y].X);
                    bw.Write(land.Normals[x, y].Y);
                    bw.Write(land.Normals[x, y].Z);
                    //Need to offset the border texcoords slightly to avoid lines between meshes
                    if(x==0) bw.Write(0.005f);
                    else if(x==64) bw.Write(0.997f);
                    else bw.Write((float)x/64.0f);
                    if(y==0) bw.Write(0.005f);
                    else if(y==64) bw.Write(0.997f);
                    else bw.Write((float)y/64.0f);
                }
            }
            bw.Close();
        }
        private void GenerateWorldMesh(int detail, bool avgVerts, string path) {
            BinaryWriter bw=new BinaryWriter(File.Create(path));
            int vCount=64/(1<<detail);
            int vSpacing=128*(1<<detail);
            int VertsAcross=(MapSize*2+1)*vCount;
            int tFaces=(VertsAcross-1)*(VertsAcross-1)*2;
            int tVerts=VertsAcross*VertsAcross;
            float texMulti=1/(float)VertsAcross;

            bw.Write(tVerts);
            bw.Write(tFaces);

            //Write out verticies
            for(int y1=-MapSize;y1<=MapSize;y1++) {
                for(int y2=0;y2<vCount;y2++) {
                    for(int x1=-MapSize;x1<=MapSize;x1++) {
                        for(int x2=0;x2<vCount;x2++) {

                            bw.Write((float)(((float)(x1*vCount+x2)+0.5f)*vSpacing));
                            bw.Write((float)(((float)(y1*vCount+y2)+0.5f)*vSpacing));

                            if(x1>map.GetLowerBound(0)&&x1<map.GetUpperBound(0)&&y1>map.GetLowerBound(1)&&y1<map.GetUpperBound(1)&&map[x1, y1]!=null) {
                                //Average height
                                int avg;
                                if(avgVerts) {
                                    avg=0;
                                    for(int y3=0;y3<(1<<detail);y3++) {
                                        for(int x3=0;x3<(1<<detail);x3++) {
                                            int New=map[x1, y1].Heights[x2*(1<<detail)+x3, y2*(1<<detail)+y3];
                                            avg+=New;
                                        }
                                    }
                                    avg/=((1<<detail)*(1<<detail));
                                } else {
                                    avg=int.MaxValue;
                                    for(int y3=0;y3<(1<<detail);y3++) {
                                        for(int x3=0;x3<(1<<detail);x3++) {
                                            int New=map[x1, y1].Heights[x2*(1<<detail)+x3, y2*(1<<detail)+y3];
                                            if(New<avg) avg=New;
                                        }
                                    }
                                }

                                bw.Write((float)avg*8);

                                //Average normal
                                Normal navg=new Normal();
                                for(int y3=0;y3<(1<<detail);y3++) {
                                    for(int x3=0;x3<(1<<detail);x3++) {
                                        navg+=map[x1, y1].Normals[x2*(1<<detail)+x3, y2*(1<<detail)+y3];
                                    }
                                }
                                navg/=((1<<detail)*(1<<detail));
                                bw.Write(navg.X);
                                bw.Write(navg.Y);
                                bw.Write(navg.Z);

                            } else {
                                bw.Write((float)-2048);

                                bw.Write(Normal.Up.X);
                                bw.Write(Normal.Up.Y);
                                bw.Write(Normal.Up.Z);
                            }
                            //Calc tex coord
                            bw.Write((float)((x1+MapSize)*vCount+x2)*texMulti);
                            bw.Write((float)((y1+MapSize)*vCount+y2)*texMulti);
                        }
                    }
                }
            }

            //Write out indicies
            if(tFaces<0xFFFF&&tVerts<0xFFFF) {
                for(int y=0;y<VertsAcross-1;y++) {
                    for(int x=0;x<VertsAcross-1;x++) {
                        bw.Write((short)((x+0)*VertsAcross+(y+0)));
                        bw.Write((short)((x+0)*VertsAcross+(y+1)));
                        bw.Write((short)((x+1)*VertsAcross+(y+0)));
                        bw.Write((short)((x+0)*VertsAcross+(y+1)));
                        bw.Write((short)((x+1)*VertsAcross+(y+1)));
                        bw.Write((short)((x+1)*VertsAcross+(y+0)));
                    }
                }
            } else {
                for(int y=0;y<VertsAcross-1;y++) {
                    for(int x=0;x<VertsAcross-1;x++) {
                        bw.Write((int)((x+0)*VertsAcross+(y+0)));
                        bw.Write((int)((x+0)*VertsAcross+(y+1)));
                        bw.Write((int)((x+1)*VertsAcross+(y+0)));
                        bw.Write((int)((x+0)*VertsAcross+(y+1)));
                        bw.Write((int)((x+1)*VertsAcross+(y+1)));
                        bw.Write((int)((x+1)*VertsAcross+(y+0)));
                    }
                }
            }

            bw.Close();
        }

        private struct CreateMeshArgs {
            public int MeshDetail;
            public bool AverageHeights;
        }
        void workerCreateMeshes(object sender, System.ComponentModel.DoWorkEventArgs e) {
#if DEBUGSTATICS
            return;
#endif
            CreateMeshArgs cma=(CreateMeshArgs)e.Argument;

            int count=0;
            GenerateMesh(defaultland, @"data files\distantland\default");
            for(int y=-MapSize;y<=MapSize;y++) {
                backgroundWorker.ReportProgress(++count);
                for(int x=-MapSize;x<=MapSize;x++) {
                    if(map[x, y]!=null) {
                        GenerateMesh(map[x, y], "data files\\distantland\\__"+x+"_"+y);
                    }
                }
            }

            backgroundWorker.ReportProgress(count, "Creating world mesh");

            GenerateWorldMesh(cma.MeshDetail, cma.AverageHeights, @"data files\distantland\world");
        }
        void workerFCreateMeshes(object sender, System.ComponentModel.RunWorkerCompletedEventArgs e) {
            if(e!=null) {
                if(e.Error!=null) {
                    DumpError(e.Error);
                    MessageBox.Show("A fatal error occured creating the landscape meshes. Distant land setup cannot continue\n"+
                    e.Error.Message);
                    ChangingPage=true;
                    Close();
                    return;
                }
                statusProgress.Value=0;
                if(e.Result!=null) {
                    warnings=(List<string>)e.Result;
                    statusWarnings.Enabled=true;
                }
                statusText.Text="Landscape meshes created";
                backgroundWorker.DoWork-=workerCreateMeshes;
                backgroundWorker.RunWorkerCompleted-=workerFCreateMeshes;
            }

            bExportStatics.Enabled=true;
            bRunStep4.Enabled=true;
            bSkipStep4.Enabled=true;
        }

        private void bImportStatics_Click(object sender, EventArgs e) {
            if(openStaticListDialog.ShowDialog()!=DialogResult.OK) return;
            tbStaticList.Text=openStaticListDialog.FileName;
        }

        private void bExportStatics_Click(object sender, EventArgs e) {
            if(saveStaticListDialog.ShowDialog()!=DialogResult.OK) return;

            bExportStatics.Enabled=false;
            bSkipStep4.Enabled=false;
            bRunStep4.Enabled=false;

            ExportStaticsArgs args=new ExportStaticsArgs();
            args.filename=saveStaticListDialog.FileName;

            ushort temp;
            if(!ushort.TryParse(tbStaticCellSize.Text, out temp)) {
                MessageBox.Show("Invalid value specified for short range static size.\n"+
                    "Must be a positive integer", "Error");
                return;
            }
            args.CellSize=temp;

            if(!ushort.TryParse(tbStaticWorldSize.Text, out temp)) {
                MessageBox.Show("Invalid value specified for long range static size.\n"+
                    "Must be a positive integer", "Error");
                return;
            }
            args.WorldSize=temp;

            if(args.WorldSize<args.CellSize) {
                MessageBox.Show("The long range size must be greater than or equal to the short range size", "Error");
                return;
            }

            statusText.Text="Exporting list of statics";
            statusProgress.Style = ProgressBarStyle.Marquee;
            backgroundWorker.WorkerReportsProgress=false;

            backgroundWorker.DoWork+=workerExportStatics;
            backgroundWorker.RunWorkerCompleted+=workerFExportStatics;

            backgroundWorker.RunWorkerAsync(args);
        }
        private void bSkipStep4_Click(object sender, EventArgs e) {
            if(!StaticsExist&&MessageBox.Show("If you skip this step, all distant static features will be unavailable.\n"+
                "Skip anyway?", "Warning", MessageBoxButtons.YesNo)!=DialogResult.Yes) return;
            ChangingPage=true;
            tabControl.SelectedIndex=4;
            statusText.Text="Distant static generation skipped";
            statusWarnings.Enabled=false;

            workerFCreateStatics(null, null);
        }
        private void bRunStep4_Click(object sender, EventArgs e) {
            if(StaticsExist) {
                if(MessageBox.Show("Distant statics files already exist. If you recreate them, the existing files will be deleted.\n"+
                    "Do you wish to continue?", "Warning", MessageBoxButtons.YesNo)!=DialogResult.Yes) return;
                Directory.Delete(@"data files\distantland\statics", true);
            }

            ushort temp;
            CreateStaticsArgs csa=new CreateStaticsArgs();
            if(!ushort.TryParse(tbStaticCellSize.Text, out temp)) {
                MessageBox.Show("Invalid value specified for short range static size.\n"+
                    "Must be a positive integer", "Error");
                return;
            }
            csa.CellSize=temp;

            if(!ushort.TryParse(tbStaticWorldSize.Text, out temp)) {
                MessageBox.Show("Invalid value specified for long range static size.\n"+
                    "Must be a positive integer", "Error");
                return;
            }
            csa.WorldSize=temp;

            if(csa.WorldSize<csa.CellSize) {
                MessageBox.Show("The long range size must be greater than or equal to the short range size", "Error");
                return;
            }

            csa.MipSkip=cmbSkipMipLevels.SelectedIndex;
            if(tbStaticList.Text!="None") csa.OverrideList=tbStaticList.Text;

            ChangingPage=true;
            tabControl.SelectedIndex=4;
            statusText.Text="Parsing plugins for possible statics";
            statusProgress.Style = ProgressBarStyle.Continuous;
            statusProgress.Maximum=6;
            backgroundWorker.WorkerReportsProgress=true;
            backgroundWorker.DoWork+=workerCreateStatics;
            backgroundWorker.RunWorkerCompleted+=workerFCreateStatics;
            statusWarnings.Enabled=false;

            backgroundWorker.RunWorkerAsync(csa);
        }

        private class Static {
            public readonly string name;
            public readonly string mesh;
            public float size;

            public Static(string Name, string Mesh) {
                name=Name.ToLower();
                mesh=Mesh.ToLower();
                size=0;
            }
        }
        private class StaticReference {
            public string name;
            public float x, y, z;
            public float yaw, pitch, roll;
            public float scale;

            private uint staticID;
            public uint StaticID { get { return staticID; } }

            public void SetID(Dictionary<string, Static> StaticsList, Dictionary<string, uint> StaticMap) {
                string file=StaticsList[name].mesh;
                if(StaticMap.ContainsKey(file)) {
                    staticID=StaticMap[file];
                } else {
                    staticID=(uint)StaticMap.Count;
                    StaticMap[file]=StaticID;
                }
            }

            public void Write(BinaryWriter bw, uint ID) {
                //for(int i=0;i<name.Length;i++) bw.Write((byte)name[i]);
                //bw.Write((byte)0);
                bw.Write(StaticID);
                //bw.Write(ID);
                bw.Write(x);
                bw.Write(y);
                bw.Write(z);
                bw.Write(yaw);
                bw.Write(pitch);
                bw.Write(roll);
                bw.Write(scale);
            }
        }
        private struct CellReference {
            public short x, y;
            public int ID;

            public override bool Equals(object obj) {
                if(obj is CellReference) return this==(CellReference)obj;
                return false;
            }

            public override int GetHashCode() {
                return x.GetHashCode()+(y.GetHashCode()<<5)+(ID.GetHashCode()<<11);
            }

            public static bool operator==(CellReference a, CellReference b) {
                return a.x==b.x&&a.y==b.y&&a.ID==b.ID;
            }
            public static bool operator!=(CellReference a, CellReference b) {
                return !(a==b);
            }
        }
        private readonly Dictionary<string, Static> StaticsList=new Dictionary<string, Static>();
        private readonly Dictionary<CellReference, StaticReference> UsedStaticsList=new Dictionary<CellReference, StaticReference>();
        private readonly Dictionary<string, uint> StaticMap=new Dictionary<string, uint>();

        private void ParseFileForStatics(BinaryReader br) {
            string s="";
            while(br.BaseStream.Position<br.BaseStream.Length) {
                s=ReadString(br);
                int size=br.ReadInt32();
                br.BaseStream.Position+=8;
                if(s=="STAT") {
                    int read=0;
                    string name=null;
                    string model=null;
                    while(read<size) {
                        s=ReadString(br);
                        int size2=br.ReadInt32();
                        read+=size2+8;
                        switch(s) {
                            case "NAME":
                                name=ReadCString(br).ToLower();
                                break;
                            case "MODL":
                                model=ReadCString(br);
                                break;
                            default:
                                br.BaseStream.Position+=size2;
                                break;
                        }
                    }
                    if(name!=null&&model!=null) StaticsList[name]=new Static(name, model);
                } else {
                    br.BaseStream.Position+=size;
                }
            }
        }
        private void ParseFileForCells(BinaryReader br) {
            string s="";
            while(br.BaseStream.Position<br.BaseStream.Length) {
                s=ReadString(br);
                int size=br.ReadInt32();
                br.BaseStream.Position+=8;
                if(s=="CELL") {
                    bool IsValidCell=false;
                    bool InReference=false;
                    bool ReferenceHasScale=false;
                    bool ReadData=false;
                    CellReference cr=new CellReference();
                    StaticReference sr=new StaticReference();
                    int read=0;
                    while(read<size) {
                        s=ReadString(br);
                        int size2=br.ReadInt32();
                        read+=size2+8;
                        switch(s) {
                            case "DATA":
                                if(!InReference) {
                                    if(ReadData) goto default;
                                    uint flags=br.ReadUInt32();
                                    if((flags&0x1)==0) IsValidCell=true;
                                    cr.x=(short)br.ReadInt32();
                                    cr.y=(short)br.ReadInt32();
                                    ReadData=true;
                                } else {
                                    sr.x=br.ReadSingle();
                                    sr.y=br.ReadSingle();
                                    sr.z=br.ReadSingle();
                                    sr.yaw=br.ReadSingle();
                                    sr.pitch=br.ReadSingle();
                                    sr.roll=br.ReadSingle();
                                }
                                break;
                            case "FRMR":
                                if(!IsValidCell) goto default;
                                if(InReference) {
                                    if(!ReferenceHasScale) sr.scale=1;
                                    if(StaticsList.ContainsKey(sr.name)) {
                                        sr.SetID(StaticsList, StaticMap);
                                        UsedStaticsList[cr]=sr;
                                    }
                                }
                                cr.ID=br.ReadInt32();
                                sr=new StaticReference();
                                InReference=true;
                                ReferenceHasScale=false;
                                break;
                            case "NAME":
                                if(!InReference) goto default;
                                sr.name=ReadCString(br).ToLower();
                                break;
                            case "XSCL":
                                if(!InReference) goto default;
                                sr.scale=br.ReadSingle();
                                ReferenceHasScale=true;
                                break;
                            default:
                                br.BaseStream.Position+=size2;
                                break;
                        }
                    }
                    if(InReference) {
                        if(!ReferenceHasScale) sr.scale=1;
                        if(StaticsList.ContainsKey(sr.name)) {
                            sr.SetID(StaticsList, StaticMap);
                            UsedStaticsList[cr]=sr;
                        }
                    }
                } else {
                    br.BaseStream.Position+=size;
                }
            }
        }
        private void ParseFileForCells2(BinaryReader br) {
            string s="";
            while(br.BaseStream.Position<br.BaseStream.Length) {
                s=ReadString(br);
                int size=br.ReadInt32();
                br.BaseStream.Position+=8;
                if(s=="CELL") {
                    bool IsValidCell=false;
                    bool InReference=false;
                    bool ReadData=false;
                    int read=0;
                    long end=br.BaseStream.Position+size;
                    while(read<size) {
                        s=ReadString(br);
                        int size2=br.ReadInt32();
                        read+=size2+8;
                        switch(s) {
                            case "DATA":
                                if(!InReference&&!ReadData) {
                                    uint flags=br.ReadUInt32();
                                    if((flags&0x1)==0) IsValidCell=true;
                                    else {
                                        read=size;
                                        br.BaseStream.Position=end;
                                        break;
                                    }
                                    br.BaseStream.Position+=8;
                                    ReadData=true;
                                    break;
                                } else goto default;
                            case "FRMR":
                                if(!IsValidCell) goto default;
                                br.ReadInt32();
                                InReference=true;
                                break;
                            case "NAME":
                                if(!InReference) goto default;
                                string sname=ReadCString(br).ToLower();
                                if(StaticsList.ContainsKey(sname)) StaticMap[sname]=1;
                                break;
                            default:
                                br.BaseStream.Position+=size2;
                                break;
                        }
                    }
                } else {
                    br.BaseStream.Position+=size;
                }
            }
        }

        private struct ExportStaticsArgs {
            public string filename;
            public float WorldSize;
            public float CellSize;
        }
        void workerExportStatics(object sender, System.ComponentModel.DoWorkEventArgs e) {
            ExportStaticsArgs args=(ExportStaticsArgs)e.Argument;
            StaticsList.Clear();
            UsedStaticsList.Clear();
            StaticMap.Clear();

            foreach(string file in files) {
                BinaryReader br=new BinaryReader(File.OpenRead(file));
                ParseFileForStatics(br);
                br.Close();
            }
            List<string> refsToRemove=new List<string>();
            foreach(KeyValuePair<string, Static> pair in StaticsList) {
                byte[] data;
                try {
                    data=BSA.GetNif(pair.Value.mesh);
                } catch { data=null; }
                if(data==null) refsToRemove.Add(pair.Key);
                else {
                    try {
                        float size=NativeMethods.ProcessNif(data, data.Length, null, 1, 1);
                        if(size<0) refsToRemove.Add(pair.Key);
                        else pair.Value.size=size;
                    } catch(Exception) {
                        //This is soooo going to result in memory leaks :(
                        //*wishes for a C# version of niflib
                        refsToRemove.Add(pair.Key);
                    }
                }
            }
            foreach(string s in refsToRemove) StaticsList.Remove(s);

            foreach(string file in files) {
                BinaryReader br=new BinaryReader(File.OpenRead(file));
                try {
                    ParseFileForCells2(br);
                } catch { }
                br.Close();
            }

            StreamWriter sw=new StreamWriter(args.filename, false);
            sw.WriteLine("#MGE distant statics list");
            sw.WriteLine();
            sw.WriteLine("#Syntax is [static edid]=[type]");
            sw.WriteLine("#If type=0 the static will not be considered for inclusion by MGE");
            sw.WriteLine("#If type=1 the static will be visable when near regardless of the short range size you specify");
            sw.WriteLine("#If type=2 the static will be visable when far regardless of the long range size you specify");
            sw.WriteLine("#If type=3 the static will obey the normal cutoff rules");
            sw.WriteLine();
            sw.WriteLine("#Lines beinging with '#' are comments");
            sw.WriteLine("#Comments must start on the first character of a line");
            sw.WriteLine("#All comments and blank lines will be completely ignored by MGE");
            sw.WriteLine();
            sw.WriteLine("#If the same edid is specified multiple times, the entry closest towards the end of this file will be used");
            sw.WriteLine("#If an edid is not specified at all, that static will obey the normal cutoff rules");
            sw.WriteLine();
            foreach(KeyValuePair<string, Static> pair in StaticsList) {
                if(StaticMap.ContainsKey(pair.Key)) {
                    sw.Write(pair.Key);
                    if(pair.Value.size>=args.WorldSize) sw.WriteLine("=2");
                    else if(pair.Value.size>=args.CellSize) sw.WriteLine("=1");
                    else sw.WriteLine("=0");
                }
            }
            sw.Close();
        }
        void workerFExportStatics(object sender, System.ComponentModel.RunWorkerCompletedEventArgs e) {
            if(e.Error!=null) {
                DumpError(e.Error);
                MessageBox.Show("A fatal error occured while exporting a list of statics. The exported list may not be complete.\n"+
                    e.Error.Message);
            }
            statusProgress.Value=0;
            statusProgress.Style= ProgressBarStyle.Continuous;
            if(e.Result!=null) {
                warnings=(List<string>)e.Result;
                statusWarnings.Enabled=true;
            }
            statusText.Text="Statics list exported";
            backgroundWorker.DoWork-=workerExportStatics;
            backgroundWorker.RunWorkerCompleted-=workerFExportStatics;

            bExportStatics.Enabled=true;
            bRunStep4.Enabled=true;
            bSkipStep4.Enabled=true;
        }

        private struct CreateStaticsArgs {
            public float WorldSize;
            public float CellSize;
            public int MipSkip;
            public string OverrideList;
        }
        void workerCreateStatics(object sender, System.ComponentModel.DoWorkEventArgs e) {
            CreateStaticsArgs args=(CreateStaticsArgs)e.Argument;
            List<string> warnings=new List<string>();
            Dictionary<string, byte> OverrideList=new Dictionary<string, byte>();
            StaticsList.Clear();
            UsedStaticsList.Clear();
            StaticMap.Clear();

            if(args.OverrideList!=null) {
                StreamReader sr=null;
                try { sr=new StreamReader(File.OpenRead(args.OverrideList)); } catch { warnings.Add("Error: Could not import statics list '"+args.OverrideList+"'"); }
                if(sr!=null) {
                    while(!sr.EndOfStream) {
                        string line=sr.ReadLine();
                        if(line.Length>0&&line[0]!='#') {
                            int index=line.LastIndexOf('=');
                            if(index!=-1) {
                                byte type;
                                if(byte.TryParse(line.Substring(index+1), out type)) {
                                    OverrideList[line.Substring(0, index)]=type;
                                    if(type>3) warnings.Add("Warning: Invalid type specified in line '"+line+"' of statics list");
                                } else warnings.Add("Warning: Failed to parse line in statics list '"+line+"'");
                            } else warnings.Add("Warning: Failed to parse line in statics list '"+line+"'");
                        }
                    }
                    sr.Close();
                }
            }

            Directory.CreateDirectory(@"data files\distantland\statics");

            foreach(string file in files) {
                BinaryReader br=new BinaryReader(File.OpenRead(file));
                ParseFileForStatics(br);
                br.Close();
            }
            backgroundWorker.ReportProgress(1, "Checking for valid nif files");
            List<string> refsToRemove=new List<string>();
            foreach(KeyValuePair<string, Static> pair in StaticsList) {
                //if(!File.Exists("data files\\distantland\\statics\\"+pair.Value.Replace('\\', '.'))) refsToRemove.Add(pair.Key);
                byte[] data;
                try {
                    data=BSA.GetNif(pair.Value.mesh);
                } catch { data=null; }
                if(data==null) refsToRemove.Add(pair.Key);
                else {
                    try {
                        float size;
                        if(OverrideList.ContainsKey(pair.Key)) {
                            switch(OverrideList[pair.Key]) {
                                case 0: size=-1; break;
                                case 1:
                                    size=NativeMethods.ProcessNif(data, data.Length, "data files\\distantland\\statics\\"+pair.Value.mesh.Replace('\\', '.'), 1, float.MaxValue);
                                    break;
                                case 2:
                                    size=NativeMethods.ProcessNif(data, data.Length, "data files\\distantland\\statics\\"+pair.Value.mesh.Replace('\\', '.'), 1, 1);
                                    break;
                                default:
                                    size=NativeMethods.ProcessNif(data, data.Length, "data files\\distantland\\statics\\"+pair.Value.mesh.Replace('\\', '.'), args.CellSize, args.WorldSize);
                                    break;
                            }
                        } else {
                            size=NativeMethods.ProcessNif(data, data.Length, "data files\\distantland\\statics\\"+pair.Value.mesh.Replace('\\', '.'), args.CellSize, args.WorldSize);
                        }
                        if(size<0) refsToRemove.Add(pair.Key);
                        else pair.Value.size=size;
                    } catch(Exception) {
                        //This is soooo going to result in memory leaks :(
                        //*wishes for a C# version of niflib
                        refsToRemove.Add(pair.Key);
                    }
                }
            }
            foreach(string s in refsToRemove) StaticsList.Remove(s);
            backgroundWorker.ReportProgress(2, "Parsing plugins for used statics");

            foreach(string file in files) {
                BinaryReader br=new BinaryReader(File.OpenRead(file));
                try {
                    ParseFileForCells(br);
                } catch(Exception ex) {
                    warnings.Add("Non fatal error in ParseFileForCells(\""+file+"\"): "+ex.Message);
                }
                br.Close();
            }

            backgroundWorker.ReportProgress(3, "Generating list of used statics");
            foreach(KeyValuePair<string, Static> pair in StaticsList) {
                string filepath="data files\\distantland\\statics\\"+pair.Value.mesh.Replace('\\', '.');
                if(File.Exists(filepath)) {
                    if(StaticMap.ContainsKey(pair.Value.mesh)) File.Move(filepath, "data files\\distantland\\statics\\"+StaticMap[pair.Value.mesh]);
                    else File.Delete(filepath);
                }
            }

            backgroundWorker.ReportProgress(4, "Writing out placement info");
            BinaryWriter bw=new BinaryWriter(File.Create("data files\\distantland\\statics\\usage.data"));
            bw.Write(StaticMap.Count);
            bw.Write(UsedStaticsList.Count);
            uint UseRef=0;
            foreach(KeyValuePair<CellReference, StaticReference> pair in UsedStaticsList) {
                pair.Value.Write(bw, UseRef++);
            }
            bw.Close();

            if(args.MipSkip!=0) {
                StaticTexCreator stc=new StaticTexCreator(args.MipSkip);
                backgroundWorker.ReportProgress(5, "Generating low quality textures");
                foreach(KeyValuePair<string,uint> pair in StaticMap) {
                    uint i=pair.Value;
                    if(!File.Exists("data files\\distantland\\statics\\"+i)) continue;
                    BinaryReader br=new BinaryReader(File.OpenRead("data files\\distantland\\statics\\"+i), System.Text.Encoding.Default);
                    int nodes=br.ReadInt32();
                    br.BaseStream.Position+=5;
                    for(int j=0;j<nodes;j++) {
                        br.BaseStream.Position+=64; //4 - radius, 1 - Far, 64 - transform matrix
                        int verts=br.ReadInt32();
                        int faces=br.ReadInt32();
                        br.BaseStream.Position+=verts*32+faces*6;
                        short chars=br.ReadInt16();
                        string path=new string(br.ReadChars(chars-1));
                        br.BaseStream.Position+=1;
                        try {
                            stc.LoadTexture(path);
                        } catch(ArgumentException) {
                            warnings.Add("Warning: Texture '"+path+"' on subset "+j+" of mesh '"+pair.Key+"' could not be found");
                        }
                    }
                    br.Close();
                }
                stc.Dispose();
            }

            if(warnings.Count>0) e.Result=warnings;
        }
        void workerFCreateStatics(object sender, System.ComponentModel.RunWorkerCompletedEventArgs e) {
            if(e!=null) {
                if(e.Error!=null) {
                    DumpError(e.Error);
                    MessageBox.Show("A fatal error occured creating the distant statics files. Distant land setup cannot continue\n"+
                    e.Error.Message);
                    ChangingPage=true;
                    Close();
                    return;
                }

                statusProgress.Value=0;
                if(e.Result!=null) {
                    warnings=(List<string>)e.Result;
                    statusWarnings.Enabled=true;
                }
                statusText.Text="Distant statics created";
                backgroundWorker.DoWork-=workerCreateStatics;
                backgroundWorker.RunWorkerCompleted-=workerFCreateStatics;
            }
            

            DirectoryInfo di=new DirectoryInfo(@"data files\distantland");
            long fileSize=0;
            int filecount=di.GetFiles().Length;
            int cells=((filecount-2)/2);
            int statics;
            foreach(FileInfo fi in di.GetFiles()) fileSize+=fi.Length;
            di=new DirectoryInfo(@"data files\distantland\statics");
            if(di.Exists) {
                statics=di.GetFiles().Length;
                foreach(FileInfo fi in di.GetFiles()) fileSize+=fi.Length;
                filecount+=statics;
            } else statics=2;
            di=new DirectoryInfo(@"data files\distantland\statics\textures");
            if(di.Exists) {
                filecount+=di.GetFiles().Length;
                foreach(FileInfo fi in di.GetFiles()) fileSize+=fi.Length;
            }

            lStatics.Text="Distant land file generation complete"+Environment.NewLine+
                "Total files created: "+filecount+Environment.NewLine+
                "Total size: "+(fileSize/(1024*1024))+"MB"+Environment.NewLine+
                "Total processed cells: "+cells+Environment.NewLine+
                "Total unique statics: "+(statics-2);

            bFinish.Enabled=true;
        }

        private void bFinish_Click(object sender, EventArgs e) {
            File.WriteAllBytes(@"data files\distantland\version", new byte[] { Statics.DistantLandVersion });
            ChangingPage=true;
            Close();
        }
        
    }
}