using System;
using CancelEventArgs=System.ComponentModel.CancelEventArgs;
using System.Collections.Generic;
using System.Windows.Forms;
using RegistryKey=Microsoft.Win32.RegistryKey;
using Registry=Microsoft.Win32.Registry;
using System.IO;
using MGEgui.DirectX;

namespace MGEgui {
    public partial class MainForm : Form {
        private bool Save=true;

        public MainForm() {
            InitializeComponent();
            //Attach dud context menus to combo boxes
            cmbAlias.ContextMenu=DudMenu;
            cmbAnsiLevel.ContextMenu=DudMenu;
            cmbFogMode.ContextMenu=DudMenu;
            cmbMipFilter.ContextMenu=DudMenu;
            cmbScaleFilter.ContextMenu=DudMenu;
            cmbVWait.ContextMenu=DudMenu;
            cmbZBuffer.ContextMenu=DudMenu;
            cmbZoom.ContextMenu=DudMenu;
            cmbRotation.ContextMenu=DudMenu;
            //Title
            Text="Morrowind Graphics Extender ("+Statics.versionString+")";
            //Reset graphics filters to linear
            cmbScaleFilter.SelectedIndex=1;
            cmbMipFilter.SelectedIndex=2;
            //Load settings
            LoadFile();
            RegistryKey key=Registry.LocalMachine.OpenSubKey(@"Software\Bethesda Softworks\Morrowind");
            uint width, height;
            try {
            	width=(uint)(int)key.GetValue("Screen Width");
            	height=(uint)(int)key.GetValue("Screen Height");
            } catch {
            	MessageBox.Show("Unable to read screen resolution from registry","Error");
            	width=640;
            	height=480;
            }
            tbResolution.Text=width.ToString()+"*"+height.ToString();
            try {
            	cbWindowed.Checked=!Convert.ToBoolean(((byte[])key.GetValue("Fullscreen"))[0]);
            } catch {
            	MessageBox.Show("Unable to read window mode from registry","Error");
            	cbWindowed.Checked=false;
            }
            key.Close();

            //Set up tooltip text
            toolTip.SetToolTip(bReset, "Resets all options back to morrowind defaults.");
            toolTip.SetToolTip(bQuality, "Sets graphics options for the best visual quality.");
            toolTip.SetToolTip(bHelp, "Opens the help file");
            toolTip.SetToolTip(bImport, "Import an MGE settings file\n"+
                "To import morrowind.ini as well, ensure that 'import morrowind.ini is checked'");
            toolTip.SetToolTip(bExport, "Export your MGE settings and morrowind.ini");
            toolTip.SetToolTip(cbImportIni, "If this is checked, morrowind.ini will be imported along with MGE settings");
            toolTip.SetToolTip(bViewCaps, "Displays your graphics cards maximum supported antialiasing and anisotropic filtering levels.");
            toolTip.SetToolTip(cmbRefreshRate, "Sets the monitor refresh rate for fullscreen mode.\n"+
                 "This has no effect in windowed mode.");
            toolTip.SetToolTip(cmbAlias, "Set antialiasing level.\n"+
                 "If you set this above what your graphics card supports, you will get a render creation error when starting up morrowind.\n"+
                 "If you set this to anything other than none, you will not be able to create a lockable backbuffer.");
            toolTip.SetToolTip(cmbZBuffer, "Set the z-buffer format.\n"+
                 "If you set this to something that your graphics card doesn't support, you will get a redner creation error when starting up morrowind.\n"+
                 "Using D16 or D16L will cause graphical artifacts while playing morrowind.");
            toolTip.SetToolTip(cmbVWait, "Set's the vertical refresh setting.\n"+
                 "Setting this to 1,2,3 or 4 will cause an fps hit, but will remove tearing artifacts.\n"+
                 "In DX8 Default 1 refreshes immediatly. In DX9 it's equivelent to '1'\n"+
                 "This setting has no effect in windowed mode under DX8");
            toolTip.SetToolTip(bCalcRefresh, "Finds the valid refresh rates for morrowind currently selected resolution.\n"+
                 "Valid refresh rates depend on resolution, adaptor and whether 16 bit colour mode is enabled.\n"+
                 "If any of these change, be sure to recalulate the available refresh rates if you want to use a value other than default.");
            toolTip.SetToolTip(cbWindowed, "Check this to run morrowind in a window instead of fullscreen");
            toolTip.SetToolTip(tbResolution, "This is morrowinds current screen resolution.");
            toolTip.SetToolTip(bCalcResolution, "Click this to change morrowinds resolution.\n"+
                 "In windowed mode you can use any resolution you like.\n"+
                 "Non 4:3 aspect ratio's may not work correctly. These are marked with a '^'");
            toolTip.SetToolTip(cmbScaleFilter, "Set the min and mag texture filters.\n"+
                 "If you set a filter not supported by your card, morrowind will automatically use the next best available filter.\n"+
                 "Do not set this to 'point' if you are using morrowind's pixel-shaded water.");
            toolTip.SetToolTip(cmbMipFilter, "Set the mip texture filter.\n"+
                 "If you set a filter not supported by your card, morrowind will automatically use the next best available filter.");
            toolTip.SetToolTip(cmbAnsiLevel, "Set the anisotropic filtering level.\n"+
                 "This will only have an effect if you use anisotropic filtering for either scale or mip filters.\n"+
                 "If you set this to above what your card supports, it will automatically drop down to the highest supported level.");
            toolTip.SetToolTip(tbLOD, "Sets a mipmap LOD bias. Valid range is between -12 and 12.\n"+
                 "Negative values increase detail, positive reduce it.\n"+
                 "Very low values will cause aliasing of texturs in the background.");
            toolTip.SetToolTip(cmbFogMode, "Sets the type of fog that morrowind uses.\n"+
                 "Using vertex fog on some ATI cards may cause everything except the sky to become completely fogged.");
            toolTip.SetToolTip(cbFog, "Uncheck this to disable fogging in morrowind.");
            toolTip.SetToolTip(cbFpsCounter, "Enables MGE's fps counter.\n"+
                 "This is unrelated to morrowinds fps counter, and can be used instead of or as well as it.\n"+
                 "Display messages must be checked for this to work.");
            toolTip.SetToolTip(cmbZoom, "Sets the initial level of zoom.\n"+
                 "Changing this will not enable zoom. You must bind the 'toogle zoom' function to a key in the macro editor.");
            toolTip.SetToolTip(cmbRotation, "This allows you to rotate the screen in 90 degree steps.");
            toolTip.SetToolTip(cbAAFix, "Fixes the monochrome backgrounds you get when loading or saving when using antialiasing.");
            toolTip.SetToolTip(cbDisplayMessages, "Allows MGE to draw text to the screen.\n"+
                 "As well as the fps counter, MGE will also alert you when various graphics options get changed.");
            toolTip.SetToolTip(cbHWShader, "Enables fullscreen pixel shaders.\n"+
                 "Some shaders will make it difficult to view menus, so it may be better to bind use the macro editors 'toggle HW shader' function");
            toolTip.SetToolTip(cmbScreenshotFormat, "Sets the format in which MGE will save screenshots.\n"+
                 "This does not change the format in which morrowind saves screenshots.");
            toolTip.SetToolTip(cbSkipMovie, "Skips the two opening movies.");
            toolTip.SetToolTip(cbConsole, "Disables the console.");
            toolTip.SetToolTip(cbAltCombat, "Allows you to use daggerfall style comabt controls.\n"+
                 "You must assign the 'toogle combat controls' function to a key in the macro editor to use this.");
            toolTip.SetToolTip(cbLagFix, "Attempts to fix input lag.\n"+
                 "This will cause a small graphical artifact in the top left corner of the screen, and may permanently mark your minimap.\n"+
                 "Turning on MGE's fps counter is usually sufficient to remove input lag caused by AA or shadows.");
            toolTip.SetToolTip(bMacroEd, "Opens up the macro editor, allowing you to assign function and macros to keyboard keys and mouse buttons.");
            toolTip.SetToolTip(bRemapper, "Opens up the keyboard remapper, with which you can remap any key on your keyboard");
            toolTip.SetToolTip(cbZoomEffectsHUD, "Check this if you want zoom and screen rotation to effect the HUD.");
            toolTip.SetToolTip(cbIdle, "Check this to force morrowind to give up control of the cpu once a frame.\n");
            toolTip.SetToolTip(tbIdle, "Controls how long morrowind gives up the cpu. Valid range is 0 to 255.\n"+
                 "0 causes morrowind to resume immediatly if no other tasks are waiting.");
            toolTip.SetToolTip(tbMaxView, "Changes the maximum view distance that MGE will set.\n"+
                 "This setting only affects the 'keep fps constant' setting.\n"+
                 "Morrowind default is 72");
            toolTip.SetToolTip(tbMinView, "Changes the minimum view distance that MGE will set.\n"+
                 "This setting only affects the 'keep fps constant' setting.\n"+
                 "Morrowind default is 25");
            toolTip.SetToolTip(cbSetEffectVars, "If this box is unchecked, the sinvar, linvar, tickcount, eyevec, sunvec and suncol variables will be unavailable to shaders.\n"+
                 "Unchecking this will give a performance boost, especially if you run many simultanious shaders.");
            toolTip.SetToolTip(cbScreenshots, "The 'Screen Shot Enable' line.\n"+
                 "Check this to enable morrowind's inbuilt screenshot function.");
            toolTip.SetToolTip(cbYesToAll, "The 'AllowYesToAll' line.\n"+
                 "Check this to add an additional 'yes to all' option when clicking through multiple warning messages.");
            toolTip.SetToolTip(cbHQShadows, "The 'High Detail Shadows' line.\n"+
                 "Check this to enable must better quality shadows.\n"+
                 "This will cause a huge fps hit.");
            toolTip.SetToolTip(cbThreadLoad, "The 'DontThreadLoad' line.\n"+
                "Check this to allow morrowind to load new cells in a seperate thread.");
            toolTip.SetToolTip(cbShowFps, "The 'Show FPS' line.\n"+
                "Check this to enable morrowinds inbuilt fps counter.");
            toolTip.SetToolTip(cbAudio, "The 'Disable Audio' line.\n"+
                "Check this to cut out some of morrowind's audio.\n"+
                "Use the disable audio option on the sound tab to completely disable morrowind's audio.");
            toolTip.SetToolTip(cbSubtitles, "The 'Subtitles' line.\n"+
                "Check this to display subtitles where normally there would just be a voice over.");
            toolTip.SetToolTip(cbHitFader, "The 'ShowHitFader' line.\n"+
                "Uncheck this to disable the red hit fader that appears when you take damage.");
            toolTip.SetToolTip(cbMultipleCS, "The 'AllowMultipleEditors' line.\n"+
                "Check this to allow multiple copies of the construction set to run simultaniously.");
            toolTip.SetToolTip(bShaderEd, "A shader editor for working on fullscreen shaders.");
            toolTip.SetToolTip(cbFPSConst, "Check this to allow MGE to modify view distance in order to keep a constant fps.\n"+
                "MGE will only modify view distance. No other graphics options will be changed.");
            toolTip.SetToolTip(tbFPSDes, "The fps MGE should try to obtain.");
            toolTip.SetToolTip(tbFPSTol, "The amount that your fps needs to be above or below your desired amount before MGE modifies framerate.");
            toolTip.SetToolTip(cbBindAI, "Checking this box will bind the AI distance to your view distance.\n"+
                "Increasing view distance with MGE will automatically increase AI distance.");
            toolTip.SetToolTip(cbDistantLand, "Check this box to display extra cells outside of morrowinds normal view range.\n"+
                "These extra cells are landscape only. NPCs, statics, water etc will not be drawn.\n"+
                "This option conflicts with vertex fog settings.");
            toolTip.SetToolTip(tbDLDist, "This controls how many cells MGE draws at high detail before switching to a low detail mesh.\n"+
                "Increasing this will exponentially increase load times when moving between cells, as well as reducing fps.");
            toolTip.SetToolTip(tbDLFogDistance, "This controls how many cells you can see with distant land enabled.\n"+
                "This must be greater than the land distance setting.");
            toolTip.SetToolTip(cbHDR, "Check this box to supply HDR information to shaders.\n"+
                "Note that checking this box by itself will not enable HDR. You also need to add a HDR shader to the active shader chain.\n"+
                "Checking this box may cause an fps hit.");
            toolTip.SetToolTip(tbHDR, "This box controls how fast the HDR reacts to brightness changes.\n"+
                "Higher numbers react faster.");
            toolTip.SetToolTip(cbHook, "Enables MGE texture hooking. Required for mods which attach effects to existing morrowind objects.\n"+
                "Checking this option may increase loading times.");
        }
        
        private void LoadINIData() {
            StreamReader sr=null;
                //INI file (Do this first so graphics settings correctly modify them
                sr=new StreamReader("morrowind.ini");
                List<string> inifileTemp=new List<string>();
                int index;
                while(sr.Peek()!=-1) {
                    string s=sr.ReadLine();
                    index=s.IndexOf(';');
                    if(index!=-1) s=s.Remove(index,s.Length-index);
                    s=s.Trim().ToLower();
                    if(s!="") inifileTemp.Add(s);
                }
                sr.Close();
                string[] inifile=(string[])inifileTemp.ToArray();
                foreach(string s in inifile) {
                    index=s.IndexOf('=');
                    if(index!=-1) {
                        string start=s.Remove(index,s.Length-index).Trim();
                        if(start=="screen shot enable") {
                            if(s[index+1]=='0') {
                                cbScreenshots.Checked=false;
                            } else {
                                cbScreenshots.Checked=true;
                            }
                        } else if(start=="dontthreadload") {
                            if(s[index+1]=='0') {
                                cbThreadLoad.Checked=true;
                            } else {
                                cbThreadLoad.Checked=false;
                            }
                        } else if(start=="allowyestoall") {
                            if(s[index+1]=='0') {
                                cbYesToAll.Checked=false;
                            } else {
                                cbYesToAll.Checked=true;
                            }
                        } else if(start=="high detail shadows") {
                            if(s[index+1]=='0') {
                                cbHQShadows.Checked=false;
                            } else {
                                cbHQShadows.Checked=true;
                            }
                        } else if(start=="show fps") {
                            if(s[index+1]=='0') {
                                cbShowFps.Checked=false;
                            } else {
                                cbShowFps.Checked=true;
                            }
                        } else if(start=="disable audio") {
                            if(s[index+1]=='0') {
                                cbAudio.Checked=false;
                            } else {
                                cbAudio.Checked=true;
                            }
                        } else if(start=="subtitles") {
                            if(s[index+1]=='0') {
                                cbSubtitles.Checked=false;
                            } else {
                                cbSubtitles.Checked=true;
                            }
                        } else if(start=="showhitfader") {
                            if(s[index+1]=='0') {
                                cbHitFader.Checked=false;
                            } else {
                                cbHitFader.Checked=true;
                            }
                        } else if(start=="allowmultipleeditors") {
                            if(s[index+1]=='0') {
                                cbMultipleCS.Checked=false;
                            } else {
                                cbMultipleCS.Checked=true;
                            }
                        }
                    }
                }
            
        }

        private void LoadGraphicsSaveFile() {
            BinaryReader br=new BinaryReader(File.OpenRead("MGE3/settings"));
            byte version=br.ReadByte();
            if(version<41) throw new Exception("The graphics save file is too old to import");
            if(version>Statics.SaveVersion) throw new Exception("The graphics save file is newer than this version of MGE");
            byte b;
            cbBindAI.Checked=br.ReadBoolean();
            switch(b=br.ReadByte()) {
                case 0: cmbAlias.SelectedIndex=0; break;
                default: cmbAlias.SelectedIndex=b-1; break;
            }
            switch(br.ReadByte()) {
                case 75: cmbZBuffer.SelectedIndex=0; break;
                case 77: cmbZBuffer.SelectedIndex=1; break;
                case 71: cmbZBuffer.SelectedIndex=2; break;
                case 80: cmbZBuffer.SelectedIndex=3; break;
                case 70: cmbZBuffer.SelectedIndex=4; break;
                case 82: cmbZBuffer.SelectedIndex=5; break;
                default: cmbZBuffer.SelectedIndex=0; break;
            }
            switch(br.ReadByte()) {
                case 3: cmbVWait.SelectedIndex=0; break;
                case 1: cmbVWait.SelectedIndex=1; break;
                case 2: cmbVWait.SelectedIndex=2; break;
                case 4: cmbVWait.SelectedIndex=3; break;
                case 8: cmbVWait.SelectedIndex=4; break;
                default: cmbVWait.SelectedIndex=5; break;
            }
            switch(b=br.ReadByte()) {
                case 0: cmbRefreshRate.SelectedIndex=0; break;
                default: cmbRefreshRate.Text=b.ToString(); break;
            }
            cbFpsCounter.Checked=br.ReadBoolean();
            cbLagFix.Checked=br.ReadBoolean();
            cbAAFix.Checked=br.ReadBoolean();
            cbDisplayMessages.Checked=br.ReadBoolean();
            cbIdle.Checked=br.ReadBoolean();
            tbIdle.Text=br.ReadByte().ToString();
            cbFog.Checked=br.ReadBoolean();
            cmbScreenshotFormat.SelectedIndex=br.ReadByte();
            cmbScaleFilter.SelectedIndex=br.ReadByte()-1;
            cmbMipFilter.SelectedIndex=br.ReadByte();
            switch(br.ReadByte()) {
                case 1: cmbAnsiLevel.SelectedIndex=0; break;
                case 2: cmbAnsiLevel.SelectedIndex=1; break;
                case 4: cmbAnsiLevel.SelectedIndex=2; break;
                case 8: cmbAnsiLevel.SelectedIndex=3; break;
                case 16: cmbAnsiLevel.SelectedIndex=4; break;
            }
            tbLOD.Text=br.ReadSByte().ToString();
            cmbFogMode.SelectedIndex=br.ReadByte();
            br.ReadByte();  //Fog pixel mode
            br.ReadByte();  //Fog vertex mode
            br.ReadByte();  //Ranged fog
            cmbZoom.SelectedIndex=br.ReadByte()-3;
            cmbRotation.SelectedIndex=br.ReadByte();
            cbHWShader.Checked=br.ReadBoolean();
            cbZoomEffectsHUD.Checked=br.ReadBoolean();
            cbFPSConst.Checked=br.ReadBoolean();
            tbFPSDes.Text=br.ReadByte().ToString();
            tbFPSTol.Text=br.ReadByte().ToString();
            tbMinView.Text=br.ReadByte().ToString();
            tbMaxView.Text=br.ReadByte().ToString();
            cbSetEffectVars.Checked=br.ReadBoolean();
            cbHook.Checked=br.ReadBoolean();
            cbDistantLand.Checked=br.ReadBoolean();
            cbDistantStatics.Checked=br.ReadBoolean();
            if(version>=43) br.ReadByte();
            if(version>=42) {
                cbDLReflectiveWater.Checked=br.ReadBoolean();
                cbDLReflectNStatics.Checked=br.ReadBoolean();
                cbDLReflectFStatics.Checked=br.ReadBoolean();
            }
            cbDLsm3Water.Checked=br.ReadBoolean();
            tbDLDist.Text=br.ReadByte().ToString();
            tbDLFogDistance.Text=br.ReadByte().ToString();
            cbHDR.Checked=br.ReadBoolean();
            tbHDR.Text=br.ReadByte().ToString();
            cbAspectZoom.Checked=br.ReadBoolean();
            br.Close();
            if(version<Statics.SaveVersion) {
                MessageBox.Show("Old graphics save file was imported successfully", "Message");
            }
        }

        private void LoadInputSaveFile() {
            byte version;
            if(!File.Exists("MGE3/DInput.data")) return;
            BinaryReader br=new BinaryReader(File.OpenRead("MGE3/DInput.data"));
            version=br.ReadByte();
            if(version<39) throw new Exception("The input save file is too old to import");
            if(version>Statics.SaveVersion) throw new Exception("The input save file is newer than this version of MGE");
            cbSkipMovie.Checked=br.ReadBoolean();
            cbConsole.Checked=br.ReadBoolean();
            cbAltCombat.Checked=br.ReadBoolean();
            try { UnserializeMacroSaves(); } catch {
                MessageBox.Show("An error occured while loading macros or triggers. Some data may have been lost.","Error");
                for(int i=0;i<266;i++) {
                    Statics.Macros[i].Type=(MacroType)br.ReadByte();
                    switch(Statics.Macros[i].Type) {
                        case MacroType.Console1:
                        case MacroType.Console2:
                            Statics.Macros[i].Console.Length=br.ReadByte();
                            for(int j=0;j<266;j++) {
                                Statics.Macros[i].Console.KeyCodes[j].Code=br.ReadByte();
                            }
                            for(int j=0;j<266;j++) {
                                Statics.Macros[i].Console.KeyCodes[j].Down=br.ReadBoolean();
                            }
                            break;
                        case MacroType.Press1:
                        case MacroType.Press2:
                        case MacroType.Unpress:
                        case MacroType.Hammer1:
                        case MacroType.Hammer2:
                        case MacroType.Unhammer:
                        case MacroType.AHammer1:
                        case MacroType.AHammer2:
                        case MacroType.AUnhammer:
                            for(int j=0;j<266;j++) {
                                Statics.Macros[i].Press.KeyStates[j]=br.ReadBoolean();
                            }
                            for(int j=0;j<267;j++) {
                                br.ReadByte();
                            }
                            break;
                        case MacroType.BeginTimer:
                        case MacroType.EndTimer:
                            Statics.Macros[i].Timer.TimerID=br.ReadByte();
                            for(int j=0;j<266*2;j++) {
                                br.ReadByte();
                            }
                            break;
                        case MacroType.Graphics:
                            Statics.Macros[i].Graphics.function=br.ReadByte();
                            for(int j=0;j<266*2;j++) {
                                br.ReadByte();
                            }
                            break;
                        default:
                            for(int j=0;j<266*2+1;j++) {
                                br.ReadByte();
                            }
                            break;
                    }
                }
                for(int i=0;i<4;i++) {
                    Statics.Triggers[i].TimeInterval=br.ReadUInt32(); ;
                    Statics.Triggers[i].Active=br.ReadBoolean();
                    for(int j=0;j<Statics.MACROS;j++) {
                        Statics.Triggers[i].data.KeyStates[j]=br.ReadBoolean();
                    }
                    br.ReadByte();  //Needed for memory alignment reasons
                }
            }
            br.Close();
            if(version<Statics.SaveVersion) {
                MessageBox.Show("Old input save file was imported successfully", "Message");
            }
        }

        private void UnserializeMacroSaves() {
            FileStream fs;
            fs=File.OpenRead("MGE3/Macros.sav");
            Statics.Macros=(Macro[])Statics.formatter.Deserialize(fs);
            fs.Close();
            fs=File.OpenRead("MGE3/Triggers.sav");
            Statics.Triggers=(Trigger[])Statics.formatter.Deserialize(fs);
            fs.Close();
        }

        private void SaveFile() {
            StreamReader sr=new StreamReader(File.OpenRead("Morrowind.ini"),System.Text.Encoding.Default);
            List<string> array=new List<string>();
            string str="";
            while(sr.Peek()!=-1) {
                str=sr.ReadLine();
                array.Add(str);
            }
            sr.Close();

            int general=0;
            bool[] bools=new bool[9];
            //Loop to mod the general section
            for(int i=0;i<array.Count;i++) {
                string s=(string)array[i];
                int index=s.IndexOf(';');
                if(index!=-1) s=s.Remove(index,s.Length-index);
                index=s.IndexOf('=');
                if(index!=-1) s=s.Remove(index,s.Length-index);
                s=s.Trim().ToLower();
                if(s=="screen shot enable") {
                    if(cbScreenshots.Checked) {
                        array[i]=((string)array[i]).Replace('0','1');
                    } else {
                        array[i]=((string)array[i]).Replace('1','0');
                    }
                    bools[0]=true;
                } else if(s=="dontthreadload") {
                    if(cbThreadLoad.Checked) {
                        array[i]=((string)array[i]).Replace('1','0');
                    } else {
                        array[i]=((string)array[i]).Replace('0','1');
                    }
                    bools[1]=true;
                } else if(s=="allowyestoall") {
                    if(cbYesToAll.Checked) {
                        array[i]=((string)array[i]).Replace('0','1');
                    } else {
                        array[i]=((string)array[i]).Replace('1','0');
                    }
                    bools[2]=true;
                } else if(s=="high detail shadows") {
                    if(cbHQShadows.Checked) {
                        array[i]=((string)array[i]).Replace('0','1');
                    } else {
                        array[i]=((string)array[i]).Replace('1','0');
                    }
                    bools[3]=true;
                } else if(s=="show fps") {
                    if(cbShowFps.Checked) {
                        array[i]=((string)array[i]).Replace('0','1');
                    } else {
                        array[i]=((string)array[i]).Replace('1','0');
                    }
                    bools[4]=true;
                } else if(s=="disable audio") {
                    if(cbAudio.Checked) {
                        array[i]=((string)array[i]).Replace('0','1');
                    } else {
                        array[i]=((string)array[i]).Replace('1','0');
                    }
                    bools[5]=true;
                } else if(s=="subtitles") {
                    if(cbSubtitles.Checked) {
                        array[i]=((string)array[i]).Replace('0','1');
                    } else {
                        array[i]=((string)array[i]).Replace('1','0');
                    }
                    bools[6]=true;
                } else if(s=="showhitfader") {
                    if(cbHitFader.Checked) {
                        array[i]=((string)array[i]).Replace('0','1');
                    } else {
                        array[i]=((string)array[i]).Replace('1','0');
                    }
                    bools[7]=true;
                } else if(s=="allowmultipleeditors") {
                    if(cbMultipleCS.Checked) {
                        array[i]=((string)array[i]).Replace('0','1');
                    } else {
                        array[i]=((string)array[i]).Replace('1','0');
                    }
                    bools[8]=true;
                } else if(s=="[general]") general=i;
                else if(s.Length>2&&s[0]=='[') break;
            }
            if(!bools[0]) {
                if(cbScreenshots.Checked) {
                    array.Insert(general+1,"Screen Shot Enable=1");
                } else {
                    array.Insert(general+1,"Screen Shot Enable=0");
                }
            }
            if(!bools[1]) {
                if(cbThreadLoad.Checked) {
                    array.Insert(general+1,"DontThreadLoad=0");
                } else {
                    array.Insert(general+1,"DontThreadLoad=1");
                }
            }
            if(!bools[2]) {
                if(cbYesToAll.Checked) {
                    array.Insert(general+1,"allowyestoall=1");
                } else {
                    array.Insert(general+1,"allowyestoall=0");
                }
            }
            if(!bools[3]) {
                if(cbHQShadows.Checked) {
                    array.Insert(general+1,"High Detail Shadows=1");
                } else {
                    array.Insert(general+1,"High Detail Shadows=0");
                }
            }
            if(!bools[4]) {
                if(cbShowFps.Checked) {
                    array.Insert(general+1,"Show FPS=1");
                } else {
                    array.Insert(general+1,"Show FPS=0");
                }
            }
            if(!bools[5]) {
                if(cbAudio.Checked) {
                    array.Insert(general+1,"Disable Audio=1");
                } else {
                    array.Insert(general+1,"Disable Audio=0");
                }
            }
            if(!bools[6]) {
                if(cbSubtitles.Checked) {
                    array.Insert(general+1,"Subtitles=1");
                } else {
                    array.Insert(general+1,"Subtitles=0");
                }
            }
            if(!bools[7]) {
                if(cbHitFader.Checked) {
                    array.Insert(general+1,"ShowHitFader=1");
                } else {
                    array.Insert(general+1,"ShowHitFader=0");
                }
            }
            if(!bools[8]) {
                if(cbMultipleCS.Checked) {
                    array.Insert(general+1,"AllowMultipleEditors=1");
                } else {
                    array.Insert(general+1,"AllowMultipleEditors=0");
                }
            }
            StreamWriter sw=new StreamWriter(File.Open("morrowind.ini",FileMode.Create),System.Text.Encoding.Default);
            foreach(string s in array) {
                sw.WriteLine(s);
            }
            sw.Close();
            //Graphics - see d3d8header.h for layout of save file
            BinaryWriter bw=new BinaryWriter(File.Open("MGE3/settings",FileMode.Create));
            bw.Write(Statics.SaveVersion);
            bw.Write(cbBindAI.Checked);
            switch(cmbAlias.Text) {
                case "None": bw.Write((byte)0); break;
                default: bw.Write((byte)(cmbAlias.SelectedIndex+1)); break;
            }
            switch(cmbZBuffer.Text) {
                case "D24S8": bw.Write((byte)75); break;
                case "D24X8": bw.Write((byte)77); break;
                case "D32": bw.Write((byte)71); break;
                case "D16": bw.Write((byte)80); break;
                case "D16L": bw.Write((byte)70); break;
                case "D32FL": bw.Write((byte)82); break;
                default: bw.Write((byte)75); break;
            }
            switch(cmbVWait.Text) {
                case "Immediate": bw.Write((byte)3); break;
                case "1": bw.Write((byte)1); break;
                case "2": bw.Write((byte)2); break;
                case "3": bw.Write((byte)4); break;
                case "4": bw.Write((byte)8); break;
                default: bw.Write((byte)0); break;
            }
            try {
                byte TempByte=Convert.ToByte(cmbRefreshRate.Text);
                bw.Write(TempByte);
            } catch {
                bw.Write((byte)0);
            }
            bw.Write(cbFpsCounter.Checked);
            bw.Write(cbLagFix.Checked);
            bw.Write(cbAAFix.Checked);
            bw.Write(cbDisplayMessages.Checked);
            bw.Write(cbIdle.Checked);
            bw.Write(Convert.ToByte(tbIdle.Text));
            bw.Write(cbFog.Checked);
            bw.Write((byte)cmbScreenshotFormat.SelectedIndex);
            bw.Write((byte)(cmbScaleFilter.SelectedIndex+1));
            bw.Write((byte)cmbMipFilter.SelectedIndex);
            switch(cmbAnsiLevel.SelectedIndex) {
                case 0: bw.Write((byte)1); break;
                case 1: bw.Write((byte)2); break;
                case 2: bw.Write((byte)4); break;
                case 3: bw.Write((byte)8); break;
                case 4: bw.Write((byte)16); break;
            }
            bw.Write(Convert.ToSByte(tbLOD.Text));
            bw.Write((byte)cmbFogMode.SelectedIndex);
            switch(cmbFogMode.SelectedIndex) {
                case 0:
                    bw.Write((byte)3);
                    bw.Write((byte)0);
                    bw.Write((byte)0);
                    break;
                case 1:
                    bw.Write((byte)0);
                    bw.Write((byte)3);
                    bw.Write((byte)0);
                    break;
                case 2:
                    bw.Write((byte)0);
                    bw.Write((byte)3);
                    bw.Write((byte)1);
                    break;
            }
            bw.Write((byte)(cmbZoom.SelectedIndex+3));
            bw.Write((byte)cmbRotation.SelectedIndex);
            bw.Write(cbHWShader.Checked);
            bw.Write(cbZoomEffectsHUD.Checked);
            bw.Write(cbFPSConst.Checked);
            bw.Write(Convert.ToByte(tbFPSDes.Text));
            bw.Write(Convert.ToByte(tbFPSTol.Text));
            bw.Write(Convert.ToByte(tbMinView.Text));
            bw.Write(Convert.ToByte(tbMaxView.Text));
            bw.Write(cbSetEffectVars.Checked);
            bw.Write(cbHook.Checked);
            bw.Write(cbDistantLand.Checked);
            bw.Write(cbDistantStatics.Checked);
            bw.Write(!DXMain.mCaps.SupportsSM2);
            bw.Write(cbDLReflectiveWater.Checked);
            bw.Write(cbDLReflectNStatics.Checked);
            bw.Write(cbDLReflectFStatics.Checked);
            bw.Write(cbDLsm3Water.Checked);
            bw.Write(Convert.ToByte(tbDLDist.Text));
            bw.Write(Convert.ToByte(tbDLFogDistance.Text));
            bw.Write(cbHDR.Checked);
            bw.Write(Convert.ToByte(tbHDR.Text));
            bw.Write(cbAspectZoom.Checked);
            bw.Close();
            //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
            //XX Input - see input.h for layout of save file XX
            //XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
            FileStream fs;
            fs=File.Open("MGE3/Macros.sav",FileMode.Create);
            Statics.formatter.Serialize(fs,Statics.Macros);
            fs.Close();
            fs=File.Open("MGE3/Triggers.sav",FileMode.Create);
            Statics.formatter.Serialize(fs,Statics.Triggers);
            fs.Close();
            bw=new BinaryWriter(File.Open("MGE3/DInput.data",FileMode.Create));
            bw.Write(Statics.SaveVersion);
            bw.Write(cbSkipMovie.Checked);
            bw.Write(cbConsole.Checked);
            bw.Write(cbAltCombat.Checked);
            for(int i=0;i<Statics.MACROS;i++) {
                bw.Write((byte)Statics.Macros[i].Type);
                switch(Statics.Macros[i].Type) {
                    case MacroType.Console1:
                    case MacroType.Console2:
                        bw.Write(Statics.Macros[i].Console.Length);
                        for(int j=0;j<Statics.MACROS;j++) {
                            bw.Write(Statics.Macros[i].Console.KeyCodes[j].Code);
                        }
                        for(int j=0;j<Statics.MACROS;j++) {
                            bw.Write((byte)(Convert.ToByte(Statics.Macros[i].Console.KeyCodes[j].Down)*0x80));
                        }
                        break;
                    case MacroType.Press1:
                    case MacroType.Press2:
                    case MacroType.Unpress:
                    case MacroType.Hammer1:
                    case MacroType.Hammer2:
                    case MacroType.Unhammer:
                    case MacroType.AHammer1:
                    case MacroType.AHammer2:
                    case MacroType.AUnhammer:
                        for(int j=0;j<Statics.MACROS;j++) {
                            bw.Write((byte)(Convert.ToByte(Statics.Macros[i].Press.KeyStates[j])*0x80));
                        }
                        for(int j=0;j<Statics.MACROS+1;j++) {
                            bw.Write((byte)0);
                        }
                        break;
                    case MacroType.BeginTimer:
                    case MacroType.EndTimer:
                        bw.Write((byte)Statics.Macros[i].Timer.TimerID);
                        for(int j=0;j<Statics.MACROS*2;j++) {
                            bw.Write((byte)0);
                        }
                        break;
                    case MacroType.Graphics:
                        bw.Write((byte)Statics.Macros[i].Graphics.function);
                        for(int j=0;j<Statics.MACROS*2;j++) {
                            bw.Write((byte)0);
                        }
                        break;
                    default:
                        for(int j=0;j<Statics.MACROS*2+1;j++) {
                            bw.Write((byte)0);
                        }
                        break;
                }
            }
            for(int i=0;i<Statics.TRIGGERS;i++) {
                bw.Write(Statics.Triggers[i].TimeInterval*1000);
                bw.Write(Statics.Triggers[i].Active);
                for(int j=0;j<Statics.MACROS;j++) {
                    bw.Write((byte)(Convert.ToByte(Statics.Triggers[i].data.KeyStates[j])*0x80));
                }
                bw.Write((byte)0);  //Needed for memory alignment reasons
            }
            bw.Close();
            bool CreateRemap=false;
            for(int i=0;i<256;i++) {
                if(Statics.Remapper[i]!=0) {
                    CreateRemap=true;
                    break;
                }
            }
            if(CreateRemap) {
                bw=new BinaryWriter(File.Open("MGE3/remap.data", FileMode.Create));
                bw.Write(Statics.Remapper);
                bw.Close();
            } else {
                File.Delete("MGE3/remap.data");
            }
        }

        private void LoadFile() {
            try {
                LoadINIData();
                cbScreenshots.Checked=false;
            } catch (Exception ex) {
                MessageBox.Show("Error loading morrowind.ini\n"+ex.Message,"Error");
            }
            if(File.Exists("MGE3/dinput.data")&&File.Exists("MGE3/settings")) {
                try {
                    LoadGraphicsSaveFile();
                } catch(Exception ex) {
                    MessageBox.Show("Error loading graphics save file\n"+ex.Message,"Error");
                }
                try {
                    LoadInputSaveFile();
                } catch(Exception ex) {
                    MessageBox.Show("Error loading input save file\n"+ex.Message,"Error");
                }
            }
            if(File.Exists(@"mge3\mwse")) cbDisableMWSE.Checked=true; else cbDisableMWSE.Checked=false;
            if(File.Exists("MGE3/remap.data")) {
                FileStream fs=File.OpenRead("MGE3/remap.data");
                fs.Read(Statics.Remapper, 0, 256);
                fs.Close();
            }
        }

        private void IgnoreKey(object sender,KeyPressEventArgs e) {
            e.Handled=true;
        }

        private void bMacroEd_Click(object sender,EventArgs e) {
            (new MacroEditorForm()).ShowDialog();
        }

        private void MainForm_FormClosing(object sender,CancelEventArgs e) {
            if(Save) SaveFile();
        }

        private void bReset_Click(object sender,EventArgs e) {
            cmbAlias.SelectedIndex=0;
            cmbAnsiLevel.SelectedIndex=0;
            cmbFogMode.SelectedIndex=0;
            cmbMipFilter.SelectedIndex=2;
            cmbRotation.SelectedIndex=0;
            cmbScaleFilter.SelectedIndex=1;
            cmbZBuffer.SelectedIndex=0;
            cmbVWait.SelectedIndex=0;
            cmbRefreshRate.SelectedIndex=0;
            cmbZoom.SelectedIndex=0;
            cbFog.Checked=true;
            cbFpsCounter.Checked=false;
            cbSkipMovie.Checked=false;
            cbConsole.Checked=false;
            cbDisplayMessages.Checked=true;
            cmbScreenshotFormat.SelectedIndex=0;
            cbHWShader.Checked=false;
            cbAAFix.Checked=false;
            cbIdle.Checked=false;
            tbLOD.Text="0";
            cbSetEffectVars.Checked=true;
            tbMinView.Text="25";
            tbMaxView.Text="72";
        }

        private void bQuality_Click(object sender,EventArgs e) {
            int MaxAA=cbWindowed.Checked?DXMain.mCaps.MaxWindowedAA:DXMain.mCaps.MaxFullscreenAA;
            if(MaxAA==0) cmbAlias.SelectedIndex=0; else cmbAlias.SelectedIndex=MaxAA-1;
            cmbMipFilter.SelectedIndex=2;
            cmbScaleFilter.SelectedIndex=2;
            switch(DXMain.mCaps.MaxAF) {
                case 2:
                    cmbAnsiLevel.SelectedIndex=1;
                    break;
                case 4:
                    cmbAnsiLevel.SelectedIndex=2;
                    break;
                case 8:
                    cmbAnsiLevel.SelectedIndex=3;
                    break;
                case 16:
                    cmbAnsiLevel.SelectedIndex=4;
                    break;
                default:
                    cmbMipFilter.SelectedIndex=1;
                    cmbScaleFilter.SelectedIndex=1;
                    cmbAnsiLevel.SelectedIndex=0;
                    break;
            }
            cmbZBuffer.SelectedIndex=0;
            if(MaxAA>0) cbAAFix.Checked=true; else cbAAFix.Checked=false;
            tbLOD.Text="-1";
        }

        private void bHelp_Click(object sender,EventArgs e) {
            try {
                System.Diagnostics.Process p=System.Diagnostics.Process.Start("MGE3/MGE Readme.chm");
                p.Close();
            } catch { }
        }

        private void bShaderEd_Click(object sender,EventArgs e) {
            ShaderEditorForm ed=new ShaderEditorForm();
            ed.ShowDialog();
        }

        private void cbWindowed_CheckedChanged(object sender, EventArgs e) {
            cmbVWait.Enabled=!cbWindowed.Checked;
            cmbRefreshRate.Enabled=!cbWindowed.Checked;
            RegistryKey key=Registry.LocalMachine.OpenSubKey(@"Software\Bethesda Softworks\Morrowind",true);
            key.SetValue("Fullscreen", new byte[] { Convert.ToByte(!cbWindowed.Checked) });
            key.Close();
        }

        private void cbFPSConst_CheckedChanged(object sender, EventArgs e) {
            tbFPSDes.Enabled=cbFPSConst.Checked;
            tbFPSTol.Enabled=cbFPSConst.Checked;
            tbMaxView.Enabled=cbFPSConst.Checked;
            tbMinView.Enabled=cbFPSConst.Checked;
        }

        private void bExport_Click(object sender,EventArgs e) {
            if(SaveFileDialog.ShowDialog()==DialogResult.OK) {
                BinaryWriter bw=new BinaryWriter(File.Open(SaveFileDialog.FileName,FileMode.Create));
                SaveFile();
                byte[] b;
                b=File.ReadAllBytes("MGE3/DInput.data");
                bw.Write(b.Length);
                bw.Write(b);
                b=File.ReadAllBytes("MGE3/DX8.data");
                bw.Write(b.Length);
                bw.Write(b);
                b=File.ReadAllBytes("MGE3/Macros.sav");
                bw.Write(b.Length);
                bw.Write(b);
                b=File.ReadAllBytes("MGE3/Triggers.sav");
                bw.Write(b.Length);
                bw.Write(b);
                b=File.ReadAllBytes("Morrowind.ini");
                bw.Write(b.Length);
                bw.Write(b);
                bw.Close();
            }
        }

        private void bImport_Click(object sender,EventArgs e) {
            if(OpenFileDialog.ShowDialog()==DialogResult.OK) {
                BinaryReader br=new BinaryReader(File.OpenRead(OpenFileDialog.FileName));
                FileStream fs;
                byte[] b;
                int len;
                len=br.ReadInt32();
                b=br.ReadBytes(len);
                fs=File.Open("MGE3/DInput.data",FileMode.Create);
                fs.Write(b,0,len);
                fs.Close();
                len=br.ReadInt32();
                b=br.ReadBytes(len);
                fs=File.Open("MGE3/DX8.data",FileMode.Create);
                fs.Write(b,0,len);
                fs.Close();
                len=br.ReadInt32();
                b=br.ReadBytes(len);
                fs=File.Open("MGE3/Macros.sav",FileMode.Create);
                fs.Write(b,0,len);
                fs.Close();
                len=br.ReadInt32();
                b=br.ReadBytes(len);
                fs=File.Open("MGE3/Triggers.sav",FileMode.Create);
                fs.Write(b,0,len);
                fs.Close();
                if(cbImportIni.Checked) {
                    if(br.PeekChar()==-1) {
                        MessageBox.Show("Settings file does not contain ini data");
                    } else {
                        len=br.ReadInt32();
                        b=br.ReadBytes(len);
                        fs=File.Open("Morrowind.ini",FileMode.Create);
                        fs.Write(b,0,len);
                        fs.Close();
                    }
                }
                br.Close();
                LoadFile();
            }
        }

        private void bViewCaps_Click(object sender,EventArgs e) {
            MessageBox.Show("Max fullscreen antialiasing level: "+DXMain.mCaps.MaxFullscreenAA.ToString()+"x\n"+
                "Max windowed antialiasing level: "+DXMain.mCaps.MaxWindowedAA.ToString()+"x\n"+
                "Max anisotropic filtering level: "+DXMain.mCaps.MaxAF.ToString()+"x\n\n"+
                "Graphics drivers often report their cababilities incorrectly.\n"+
                "If you think your graphics card only supports lower levels of AA then you will need to reduce it by hand if you use best quality mode",
                "Message");
        }

        private void cbDisplayMessages_CheckedChanged(object sender,EventArgs e) {
            cbFpsCounter.Checked&=cbDisplayMessages.Checked;
            cbFpsCounter.Enabled=cbDisplayMessages.Checked;
        }

        private void bCalcRefresh_Click(object sender,EventArgs e) {
            int width,height;
            RegistryKey key=Registry.LocalMachine.OpenSubKey(@"Software\Bethesda Softworks\Morrowind");
            width=(int)key.GetValue("Screen Width");
            height=(int)key.GetValue("Screen Height");
            key.Close();
            cmbRefreshRate.Items.Clear();
            cmbRefreshRate.Items.Add("Default");
            cmbRefreshRate.Text="Default";
            foreach(int i in DirectX.DXMain.GetRefreshRates(width, height)) {
                cmbRefreshRate.Items.Add(i.ToString());
            }
        }

        private void bCalcResolution_Click(object sender, EventArgs e) {
            System.Drawing.Point p;
            if(ResolutionForm.ShowDialog(out p,cbWindowed.Checked)) {
                tbResolution.Text=p.X.ToString()+"*"+p.Y.ToString();
            }
        }

        private void tbLOD_Leave(object sender, EventArgs e) {
            try {
                sbyte s=Convert.ToSByte(tbLOD.Text);
                if(s>12) tbLOD.Text="12";
                if(s<-12) tbLOD.Text="-12";
            } catch {
                tbLOD.Text="0";
            }
        }

        private void tbLOD_KeyPress(object sender, KeyPressEventArgs e) {
            if(e.KeyChar!='\b'&&e.KeyChar!='-'&&!char.IsDigit(e.KeyChar)) e.Handled=true;
        }

        private void tbIdle_KeyPress(object sender, KeyPressEventArgs e) {
            if(e.KeyChar!='\b'&&!char.IsDigit(e.KeyChar)) e.Handled=true;
        }

        private void tbIdle_Leave(object sender, EventArgs e) {
            TextBox tb=(TextBox)sender;
            try {
                Convert.ToByte(tb.Text);
            } catch {
                if(tb==tbIdle) {
                    tb.Text="0";
                } else if(tb==tbFPSDes) {
                    tb.Text="30";
                } else if(tb==tbFPSTol) {
                    tb.Text="5";
                } else if(tb==tbMaxView) {
                    tb.Text="72";
                } else if(tb==tbMinView) {
                    tb.Text="25";
                } else if(tb==tbHDR) {
                    tb.Text="5";
                } else if(tb==tbDLDist) {
                    tb.Text="2";
                } else if(tb==tbDLFogDistance) {
                    tb.Text="15";
                } else {
                    tb.Text="0";
                }
            }
        }

        private void cbIdle_CheckedChanged(object sender, EventArgs e) {
            tbIdle.Enabled=cbIdle.Checked;
        }

        private void bAbout_Click(object sender, EventArgs e) {
            string s="Morrowind Graphics Extender "+Statics.versionString+" by Timeslip (www.timeslipped.co.uk)";
            MessageBox.Show(s, "About");
        }

        private void bRemapper_Click(object sender, EventArgs e) {
            (new KeyRemapper()).ShowDialog();
        }

        private void bHookGen_Click(object sender, EventArgs e) {
            (new HookCreator()).ShowDialog();
        }

        private void cbDistantLand_CheckedChanged(object sender, EventArgs e) {
            tbDLDist.Enabled=cbDistantLand.Checked;
            tbDLFogDistance.Enabled=cbDistantLand.Checked;
            if(cmbFogMode.SelectedIndex!=0) {
                MessageBox.Show("Distand land is not compatable with vertex fog when using some graphics cards.\n"+
                    "If the distant land does not appear in game, or appears untextured, try switching back to pixel fog.", "Warning");
            }
        }

        private void cmbFogMode_SelectedIndexChanged(object sender, EventArgs e) {
            if(cmbFogMode.SelectedIndex!=0&&cbDistantLand.Checked) {
                MessageBox.Show("Distand land is not compatable with vertex fog when using some graphics cards.\n"+
                    "If the distant land does not appear in game, or appears untextured, try switching back to pixel fog.", "Warning");
            }
        }

        private void tbDLDist_Leave(object sender, EventArgs e) {
            tbIdle_Leave(sender, e);
            byte b=byte.Parse(tbDLDist.Text);
            if(b==0) {
                MessageBox.Show("Land distance cannot be 0", "Error");
                tbDLDist.Text="1";
            }
            if(b>5) {
                MessageBox.Show("Land distances of greater than 5 may cause excessive loading times and/or a large fps hit","Warning");
            }
        }

        private void cmbAlias_SelectedIndexChanged(object sender, EventArgs e) {
            if(cmbAlias.SelectedIndex==0) return;
            if(!DXMain.CheckAALevel(cmbAlias.SelectedIndex+1, cbWindowed.Checked)) {
                MessageBox.Show("You're selected antialiasing level is not supported by your graphics card.", "Error");
                cmbAlias.SelectedIndex=0;
            }
        }

        private void cmbAnsiLevel_SelectedIndexChanged(object sender, EventArgs e) {
            if(cmbAnsiLevel.SelectedIndex==0) return;
            if(1<<cmbAnsiLevel.SelectedIndex > DXMain.mCaps.MaxAF) {
                MessageBox.Show("You're selected anisotropic filtering level is not supported by your graphics card.", "Error");
                cmbAnsiLevel.SelectedIndex=0;
            }
        }

        private void bDistantLand_Click(object sender, EventArgs e) {
            if(!DXMain.mCaps.SupportsSM1) {
                MessageBox.Show("Your graphics card lacks the required shader model 1.4 support required to use distant land.", "Error");
                return;
            }
            bool exists=false;
            if(Directory.Exists("data files\\distantland")) {
                if(!File.Exists(@"data files\distantland\version")) {
                    if(MessageBox.Show("Distant land files already exist, but are either incomplete, corrupt, or from an older version of MGE and must be deleted to continue.\n"+
                        "Do you wish to continue?", "Warning", MessageBoxButtons.YesNo)!=DialogResult.Yes) return;
                    Directory.Delete("data files\\distantland", true);
                } else {
                    byte[] b=File.ReadAllBytes(@"data files\distantland\version");
                    if(b.Length!=1||b[0]!=Statics.DistantLandVersion) {
                        if(MessageBox.Show("Distant land files already exist, but are from a different version of MGE and must be deleted to continue.\n"+
                        "Do you wish to continue?", "Warning", MessageBoxButtons.YesNo)!=DialogResult.Yes) return;
                        Directory.Delete("data files\\distantland", true);
                    } else exists=true;
                } 
            }
            (new DistantLand.DistantLandForm(exists)).ShowDialog();
        }

        private void cbDistantLand_CheckedChanged_1(object sender, EventArgs e) {
            if(cbDistantLand.Checked) {
                if(!DXMain.mCaps.SupportsSM1) {
                    cbDistantLand.Checked=false;
                    MessageBox.Show("Your graphics card lacks the required shader model 1.4 support required to use distant land.", "Error");
                    return;
                }
                if(!File.Exists(@"data files\distantland\version")) {
                    cbDistantLand.Checked=false;
                    MessageBox.Show("Distant land files have not been created or are from an older version of MGE.\n"+
                        "Use the 'distant land file creation wizard' on the tools tab to create them", "Error");
                    return;
                }
                byte[] bytes=File.ReadAllBytes(@"data files\distantland\version");
                if(bytes.Length!=1||bytes[0]!=Statics.DistantLandVersion) {
                    cbDistantLand.Checked=false;
                    MessageBox.Show("Distant land files have not been created or are from a different version of MGE.\n"+
                        "Use the 'distant land file creation wizard' on the tools tab to create them", "Error");
                    return;
                }
            }
            tbDLDist.Enabled=cbDistantLand.Checked;
            tbDLFogDistance.Enabled=cbDistantLand.Checked;
            cbDistantStatics.Enabled=cbDistantLand.Checked;
            cbDLReflectiveWater.Enabled=cbDistantLand.Checked;
            cbDLReflectNStatics.Enabled=cbDistantLand.Checked&&cbDLReflectiveWater.Checked&&cbDistantStatics.Checked;
            cbDLReflectFStatics.Enabled=cbDistantLand.Checked&&cbDLReflectiveWater.Checked&&cbDistantStatics.Checked&&cbDLReflectNStatics.Checked;
            cbDLsm3Water.Enabled=cbDistantLand.Checked&&cbDLReflectiveWater.Checked;
        }

        private void cbDLReflectiveWater_CheckedChanged(object sender, EventArgs e) {
            if(cbDLReflectiveWater.Checked&&!DXMain.mCaps.SupportsSM2) {
                cbDLReflectiveWater.Checked=false;
                MessageBox.Show("You graphics card lacks the required shader model 2.0 support required to use reflective water.", "Error");
                return;
            }

            cbDLReflectNStatics.Enabled=cbDistantLand.Checked&&cbDLReflectiveWater.Checked&&cbDistantStatics.Checked;
            cbDLReflectFStatics.Enabled=cbDistantLand.Checked&&cbDLReflectiveWater.Checked&&cbDistantStatics.Checked&&cbDLReflectNStatics.Checked;
            cbDLsm3Water.Enabled=cbDistantLand.Checked&&cbDLReflectiveWater.Checked;
        }

        private void cbDLReflectNStatics_CheckedChanged(object sender, EventArgs e) {
            cbDLReflectFStatics.Enabled=cbDistantLand.Checked&&cbDLReflectiveWater.Checked&&cbDistantStatics.Checked&&cbDLReflectNStatics.Checked;
        }

        private void cbDLsm3Water_CheckedChanged(object sender, EventArgs e) {
            if(cbDLsm3Water.Checked&&!DXMain.mCaps.SupportsSM3) {
                cbDLsm3Water.Checked=false;
                MessageBox.Show("Your graphics card does not support sm3", "Error");
                return;
            }
        }

        private void cbDistantStatics_CheckedChanged(object sender, EventArgs e) {
            if(cbDistantStatics.Checked&&!File.Exists(@"data files\distantland\statics\usage.data")) {
                cbDistantStatics.Checked=false;
                MessageBox.Show("The distant statics files are missing.\n"+
                    "Use the 'distant land file creation wizard' on the tools tab to create them.", "Error");
                return;
            }
            cbDLReflectNStatics.Enabled=cbDistantLand.Checked&&cbDLReflectiveWater.Checked&&cbDistantStatics.Checked;
            cbDLReflectFStatics.Enabled=cbDistantLand.Checked&&cbDLReflectiveWater.Checked&&cbDistantStatics.Checked&&cbDLReflectNStatics.Checked;
        }

        private void cbDisableMWSE_CheckedChanged(object sender, EventArgs e) {
            if(cbDisableMWSE.Checked) {
                File.Create(@"MGE3\mwse").Close();
            } else {
                File.Delete(@"MGE3\mwse");
            }
        }


    }
}