#define _DistantLand
#include "d3d8header.h"
#include "MorrowindBSA.h"

#define aDistantLand(x,y) DistantLand[(x)+((Options.DL_Distance*2+1)*(y))]
#define DL_VERTEXFORMAT D3DFVF_XYZ|D3DFVF_TEX1|D3DFVF_NORMAL
#define DL_WATERFORMAT D3DFVF_XYZ
#define SIZEOFVERT 32
#define SIZEOFWVERT 12

#define PASS_SETUP 9
#define PASS_REFLECTLAND 0
#define PASS_REFLECTSTATICS 8
#define PASS_RENDERLAND 1
#define PASS_SM2WATER 2
#define PASS_SM3WATER 4
#define PASS_RENDERSTATICS 5
#define PASS_UNDERWATER 6
#define PASS_NONREFLECTIVEWATER 7

struct WaterVertex { float x,y,z; };

struct sDistantLand {
    IDirect3DTexture* tex;
    D3DXMATRIX world;
	IDirect3DVertexBuffer* vb;

	void RelPointers() {
		if(tex) tex->Release();
		tex=0;
		if(vb) vb->Release();
		vb=0;
	}
};

static sDistantLand* DistantLand;
static sDistantLand* DistantWorld;
static sDistantLand* DistantDefault;

static short OldX,OldY,TempX,TempY;
static bool SetPos;
static bool Initialized=false;
static bool Zoomed=false;
static D3DXMATRIX WorldProj;
static D3DXMATRIX CellProj;
static D3DXMATRIX CellProj2; //Used for water reflections. Closer near-z plane.
static IDirect3DIndexBuffer9* WorldIB;
static IDirect3DIndexBuffer9* CellIB;
static DWORD WorldVerts;
static DWORD WorldFaces;
extern bool DistantLandActive;

static ID3DXEffect* DL_Effect;
static D3DXHANDLE ehWorldViewProj;
static D3DXHANDLE ehWorld;
static D3DXHANDLE ehWorld2;
static D3DXHANDLE ehTex0;
static D3DXHANDLE ehTex1;
static D3DXHANDLE ehTex2;
static D3DXHANDLE ehTex3;
static D3DXHANDLE ehSunCol;
static D3DXHANDLE ehSunAmb;
static D3DXHANDLE ehSunVec;
static D3DXHANDLE ehFogCol;
static D3DXHANDLE ehFogCol2;
static D3DXHANDLE ehFogRange;
static D3DXHANDLE ehEyePos;
static D3DXHANDLE ehTicks;
static D3DXHANDLE ehOffset;
static D3DXHANDLE ehTexBlend;

static D3DXMATRIX WaterWorld;
static IDirect3DIndexBuffer9* WaterIB;
static IDirect3DVertexBuffer9* WaterVB1;
static IDirect3DVertexBuffer9* WaterVB2;

static IDirect3DTexture9* RealWaterTextures[32];
static IDirect3DTexture9** MorrowWaterPtrs[32];
static IDirect3DTexture9* FakeWaterTexture;
static BYTE WaterLoadedCount=0;

static IDirect3DSurface9* ReflectionZBuffer;

struct ViewFrustum {
	D3DXPLANE frustum[6];

	ViewFrustum(D3DXMATRIX* viewProj) {
			D3DXMATRIX viewProjection = *viewProj;

            // Left plane 
            frustum[0].a = viewProjection._14 + viewProjection._11;
            frustum[0].b = viewProjection._24 + viewProjection._21;
            frustum[0].c = viewProjection._34 + viewProjection._31;
            frustum[0].d = viewProjection._44 + viewProjection._41;

            // Right plane 
            frustum[1].a = viewProjection._14 - viewProjection._11;
            frustum[1].b = viewProjection._24 - viewProjection._21;
            frustum[1].c = viewProjection._34 - viewProjection._31;
            frustum[1].d = viewProjection._44 - viewProjection._41;

            // Top plane 
            frustum[2].a = viewProjection._14 - viewProjection._12;
            frustum[2].b = viewProjection._24 - viewProjection._22;
            frustum[2].c = viewProjection._34 - viewProjection._32;
            frustum[2].d = viewProjection._44 - viewProjection._42;

            // Bottom plane 
            frustum[3].a = viewProjection._14 + viewProjection._12;
            frustum[3].b = viewProjection._24 + viewProjection._22;
            frustum[3].c = viewProjection._34 + viewProjection._32;
            frustum[3].d = viewProjection._44 + viewProjection._42;

            // Near plane 
            frustum[4].a = viewProjection._13;
            frustum[4].b = viewProjection._23;
            frustum[4].c = viewProjection._33;
            frustum[4].d = viewProjection._43;

            // Far plane 
            frustum[5].a = viewProjection._14 - viewProjection._13;
            frustum[5].b = viewProjection._24 - viewProjection._23;
            frustum[5].c = viewProjection._34 - viewProjection._33;
            frustum[5].d = viewProjection._44 - viewProjection._43;

            // Normalize planes 
            for(int i = 0;i < 6;i++) D3DXPlaneNormalize(&frustum[i], &D3DXPLANE(frustum[i]));
	}
};
struct DistantSubset {
	IDirect3DTexture9* tex;
	IDirect3DVertexBuffer9* vBuffer;
	IDirect3DIndexBuffer9* iBuffer;
	int verts;
	int faces;
	D3DXMATRIX transform;
};
struct DistantStatic {
	int NumSubsets;
	float radius;	//Maximum radius squared
	DistantSubset* Subsets;
	bool Far;

	void Render(D3DXMATRIX* transform, D3DXMATRIX* viewproj) {
		D3DXMATRIX worldviewproj;
		D3DXMATRIX world;
		D3DXMATRIX tempmat;
		for(int i=0;i<NumSubsets;i++) {
			//world=(*transform)*Subsets[i].transform;
			world=Subsets[i].transform*(*transform);
			worldviewproj=world*(*viewproj);
			DL_Effect->SetMatrix(ehWorld2, &world);
			D3DXMatrixInverse(&tempmat, 0, &world);
			D3DXMatrixTranspose(&world, &tempmat);
			DL_Effect->SetTexture(ehTex3, Subsets[i].tex);
			DL_Effect->SetMatrix(ehWorldViewProj, &worldviewproj);
			DL_Effect->SetMatrix(ehWorld, &world);
			DL_Effect->CommitChanges();
			RealDevice->SetStreamSource(0, Subsets[i].vBuffer, 0, SIZEOFVERT);
			RealDevice->SetIndices(Subsets[i].iBuffer);
			RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, Subsets[i].verts, 0, Subsets[i].faces);
		}
	}
};
static DistantStatic* DistantStatics;
struct UsedDistantStatic {
	DWORD StaticRef;
	//DWORD ID;
	D3DXMATRIX transform;
	float radius;
	float radiussq;
	D3DXVECTOR3 pos;

	float distance() {
		return (EyePos.x-pos.x)*(EyePos.x-pos.x) + (EyePos.y-pos.y)*(EyePos.y-pos.y);
	}

	bool CheckVisable(ViewFrustum* frustum) {
		for(int i=0;i<6;i++) {
			if(D3DXPlaneDotCoord(&frustum->frustum[i], &pos) + radius < 0 )
				return false;
		}
        return true;
	}
	void Render(D3DXMATRIX* viewproj) {
		DistantStatics[StaticRef].Render(&transform, viewproj);
	}

};
static UsedDistantStatic* UsedDistantStatics;
static int DistantStaticCount=0;
static int UsedDistantStaticCount=0;

static void RenderDistantStatics(D3DXMATRIX* view, D3DXMATRIX* proj, bool Far) {
	D3DXMATRIX viewproj=(*view)*(*proj);
	ViewFrustum frustum(&viewproj);
	if(Far) {
		for(int i=0;i<UsedDistantStaticCount;i++) {
			if(!DistantStatics[UsedDistantStatics[i].StaticRef].Far||!UsedDistantStatics[i].CheckVisable(&frustum)) continue;
			UsedDistantStatics[i].Render(&viewproj);
		}
	} else {
		for(int i=0;i<UsedDistantStaticCount;i++) {
			if(!UsedDistantStatics[i].CheckVisable(&frustum)) continue;
			UsedDistantStatics[i].Render(&viewproj);
		}
	}
}
static void LoadDistantStatics() {
	if(!Options.UseDistantStatics) return;
	DEBUGSTRING("Loading distant statics");
	DWORD unused;
	InitBSAs();
	HANDLE h = CreateFile("data files\\distantland\\statics\\usage.data", GENERIC_READ, 0, 0, OPEN_EXISTING, 0, 0);
	if(h==INVALID_HANDLE_VALUE) {
		ERRORMSG("Required distant statics data is missing or corrupted");
		return;
	}
	ReadFile(h, &DistantStaticCount, 4, &unused, 0);
	ReadFile(h, &UsedDistantStaticCount, 4, &unused, 0);
	DistantStatics=new DistantStatic[DistantStaticCount];
	UsedDistantStatics=new UsedDistantStatic[UsedDistantStaticCount];
	for(int i=0;i<DistantStaticCount;i++) {
		char buf[MAX_PATH];
		char num[9];
		strcpy(buf, "data files\\distantland\\statics\\");
		_itoa(i, num, 10);
		strcat(buf, num);
		HANDLE h2 = CreateFile(buf, GENERIC_READ, 0, 0, OPEN_EXISTING, 0, 0);
		if(h2==INVALID_HANDLE_VALUE) ERRORMSG("Required distant statics data is missing or corrupted");
		ReadFile(h2, &DistantStatics[i].NumSubsets, 4, &unused, 0);
		ReadFile(h2, &DistantStatics[i].radius, 4, &unused, 0);
		DistantStatics[i].radius*=DistantStatics[i].radius;
		ReadFile(h2, &DistantStatics[i].Far, 1, &unused, 0);
		DistantStatics[i].Subsets=new DistantSubset[DistantStatics[i].NumSubsets];
		for(int j=0;j<DistantStatics[i].NumSubsets;j++) {
			UINT verts, faces;
			void* data;
			IDirect3DVertexBuffer9* vb;
			IDirect3DIndexBuffer9* ib;
			D3DXMATRIX tempmat;

			ReadFile(h2, &tempmat, sizeof(D3DXMATRIX), &unused, 0);
			ReadFile(h2, &verts, 4, &unused, 0);
			ReadFile(h2, &faces, 4, &unused, 0);

			RealDevice->CreateVertexBuffer(verts*SIZEOFVERT,D3DUSAGE_WRITEONLY,DL_VERTEXFORMAT,D3DPOOL_DEFAULT,&vb, 0);
			vb->Lock(0,0,&data,0);
			ReadFile(h2,data,verts*SIZEOFVERT,&unused,0);
			vb->Unlock();
			RealDevice->CreateIndexBuffer(faces*6,D3DUSAGE_WRITEONLY,D3DFMT_INDEX16,D3DPOOL_DEFAULT,&ib, 0);
			ib->Lock(0,0,&data,0);
			ReadFile(h2,data,faces*6,&unused,0); //Morrowind nifs don't support 32 bit indicies?
			ib->Unlock();

			short texsize;
			ReadFile(h2, &texsize,2,&unused,0);
			char* texname=new char[texsize];
			ReadFile(h2, texname, texsize, &unused, 0);

			IDirect3DTexture9* tex=BSALoadTexture(texname);
			if(!tex) tex=DistantDefault->tex;

			DistantStatics[i].Subsets[j].transform=tempmat;
			DistantStatics[i].Subsets[j].verts=verts;
			DistantStatics[i].Subsets[j].faces=faces;
			DistantStatics[i].Subsets[j].vBuffer=vb;
			DistantStatics[i].Subsets[j].iBuffer=ib;
			DistantStatics[i].Subsets[j].tex=tex;

			delete[] texname;
		}
		CloseHandle(h2);
	}
	for(int i=0;i<UsedDistantStaticCount;i++) {
		ReadFile(h, &UsedDistantStatics[i].StaticRef, 4, &unused, 0);
		ReadFile(h, &UsedDistantStatics[i].pos, 12, &unused, 0);

		float yaw, pitch, roll, scale;
		ReadFile(h, &yaw, 4, &unused, 0);
		ReadFile(h, &pitch, 4, &unused, 0);
		ReadFile(h, &roll, 4, &unused, 0);
		ReadFile(h, &scale, 4, &unused, 0);
		if(scale==0) scale=1;
		
		UsedDistantStatics[i].radiussq=DistantStatics[UsedDistantStatics[i].StaticRef].radius*scale*scale;
		UsedDistantStatics[i].radius=sqrt(UsedDistantStatics[i].radiussq);

		D3DXMATRIX transmat, rotmatx, rotmaty, rotmatz, scalemat;
		//D3DXMATRIX transmat, rotmat, scalemat;
		D3DXMatrixTranslation(&transmat, UsedDistantStatics[i].pos.x, UsedDistantStatics[i].pos.y, UsedDistantStatics[i].pos.z);
		//D3DXMatrixRotationYawPitchRoll(&rotmat, -yaw, pitch, -roll);
		D3DXMatrixRotationX(&rotmatx, -yaw);
		D3DXMatrixRotationY(&rotmaty, -pitch);
		D3DXMatrixRotationZ(&rotmatz, -roll);
		//D3DXMatrixRotationYawPitchRoll(&rotmat, yaw, pitch, -roll);
		D3DXMatrixScaling(&scalemat, scale, scale, scale);
		UsedDistantStatics[i].transform=scalemat*rotmatz*rotmaty*rotmatx*transmat;
		//UsedDistantStatics[i].transform=scalemat*rotmat*transmat;
	}
	CloseHandle(h);
	DEBUGSTRING("Finished loading distant statics");
}

void HookWaterLoad(IDirect3DTexture9** ptrAddr) {
	if(WaterLoadedCount==32) return;
	//LOG::logline("MGE: Water texture %d hooked", (DWORD)WaterLoadedCount);
	MorrowWaterPtrs[WaterLoadedCount]=ptrAddr;
	RealWaterTextures[WaterLoadedCount]=*ptrAddr;
	WaterLoadedCount++;
}
#ifndef RELEASE
void ReloadShader() {
    if(DL_Effect) DL_Effect->Release();
	DL_Effect=0;
	HRESULT hr;
	hr=D3DXCreateEffectFromFile(RealDevice,"Data files\\shaders\\distantland\\InGame.fx",0,0,0,0,&DL_Effect,NULL);
	if(hr!=D3D_OK) {
		int iii=1;
	}
	DL_Effect->SetFloatArray("rcpres",(float*)&RcpRes,2);
	ehWorldViewProj=DL_Effect->GetParameterByName(0, "worldviewproj");
	ehWorld=DL_Effect->GetParameterByName(0, "world");
	ehWorld2=DL_Effect->GetParameterByName(0, "world2");
	ehTex0=DL_Effect->GetParameterByName(0, "tex0");
	ehTex1=DL_Effect->GetParameterByName(0, "tex1");
	ehTex2=DL_Effect->GetParameterByName(0, "tex2");
	ehTex3=DL_Effect->GetParameterByName(0, "tex3");
	ehSunCol=DL_Effect->GetParameterByName(0, "SunCol");
	ehSunAmb=DL_Effect->GetParameterByName(0, "SunAmb");
	ehSunVec=DL_Effect->GetParameterByName(0, "SunVec");
	ehFogCol=DL_Effect->GetParameterByName(0, "FogCol");
	ehFogCol2=DL_Effect->GetParameterByName(0, "FogCol2");
	ehFogRange=DL_Effect->GetParameterByName(0, "FogRange");
	ehEyePos=DL_Effect->GetParameterByName(0, "EyePos");
	ehTicks=DL_Effect->GetParameterByName(0, "ticks");
	ehOffset=DL_Effect->GetParameterByName(0, "offset");
	ehTexBlend=DL_Effect->GetParameterByName(0, "TexBlend");
}
#endif
void FreeCell(short x, short y) {
    if(aDistantLand(x,y).tex!=0) aDistantLand(x,y).RelPointers();
}
HRESULT LoadWorldMesh() {
	HANDLE file=CreateFile("data files\\distantland\\world",GENERIC_READ,0,0,OPEN_EXISTING,0,0);
	if(file==INVALID_HANDLE_VALUE) return E_FAIL;
	DWORD verts, faces, unused;
	bool large;
	ReadFile(file,&verts,4,&unused,0);
	ReadFile(file,&faces,4,&unused,0);
	byte* data;
	if(verts>0xFFFF||faces>0xFFFF) large=true; else large=false;
	RealDevice->CreateVertexBuffer(verts*SIZEOFVERT,D3DUSAGE_WRITEONLY,DL_VERTEXFORMAT,D3DPOOL_DEFAULT,&DistantWorld->vb, 0);
	DistantWorld->vb->Lock(0,0,(void**)&data,0);
	ReadFile(file,data,verts*SIZEOFVERT,&unused,0);
	DistantWorld->vb->Unlock();
	RealDevice->CreateIndexBuffer(faces*(large?4:2)*3,D3DUSAGE_WRITEONLY,large?D3DFMT_INDEX32:D3DFMT_INDEX16,D3DPOOL_DEFAULT,&WorldIB, 0);
	WorldIB->Lock(0,0,(void**)&data,0);
	ReadFile(file,data,faces*(large?4:2)*3,&unused,0);
	WorldIB->Unlock();
	CloseHandle(file);
	WorldFaces=faces;
	WorldVerts=verts;
	return S_OK;
}
HRESULT LoadMesh(char* path,sDistantLand* land) {
	HANDLE file=CreateFile(path,GENERIC_READ,0,0,OPEN_EXISTING,0,0);
	if(file==INVALID_HANDLE_VALUE) return E_FAIL;
	DWORD unused;
	byte* data;
	RealDevice->CreateVertexBuffer(0x1081*SIZEOFVERT,D3DUSAGE_WRITEONLY,DL_VERTEXFORMAT,D3DPOOL_DEFAULT,&land->vb, 0);
	land->vb->Lock(0,0,(void**)&data,0);
	ReadFile(file,data,0x1081*SIZEOFVERT,&unused,0);
	land->vb->Unlock();
	CloseHandle(file);
	return S_OK;
}
void LoadSingleCell(short cellx,short celly,short posx, short posy) {
    D3DXMatrixTranslation(&aDistantLand(posx,posy).world,(float)(cellx*8192),(float)(celly*8192),0);//+128
    char path[260];
    char temp[10];
    strcpy(path,"data files\\distantland\\__");
    _itoa(cellx,temp,10);
    strcat(path,temp);
    strcat(path,"_");
    _itoa(celly,temp,10);
    strcat(path,temp);

	if(FAILED(LoadMesh(path,&aDistantLand(posx,posy)))) {
		aDistantLand(posx,posy).tex=0;
		return;
    } else {
		strcat(path,".dds");
		if(FAILED(D3DXCreateTextureFromFileEx(RealDevice,path,0,0,0,0,D3DFMT_UNKNOWN,D3DPOOL_DEFAULT,
			D3DX_DEFAULT,D3DX_DEFAULT,0,0,0,&aDistantLand(posx,posy).tex))) {
			aDistantLand(posx,posy).vb->Release();
			aDistantLand(posx,posy).vb=0;
            aDistantLand(posx,posy).tex=0;
			return;
        }
		
    }
}

void ChangeCell(short cellx, short celly) {
    //Stick the projection stuff here, since there's nowhere better to put it
    float zf=(float)(((float)Options.DL_Distance*2)*8192.0f);
    float zn=2000; //2000
    CellProj=TransProj;
    CellProj._33=zf/(zf-zn);
    CellProj._43=-zn*zf/(zf-zn);
	zn=10;
	CellProj2=TransProj;
    CellProj2._33=zf/(zf-zn);
    CellProj2._43=-zn*zf/(zf-zn);
    zn=zf/5;
	zf=(float)((Options.DL_Fog+1)*8192);
    WorldProj=TransProj;
    WorldProj._33=zf/(zf-zn);
    WorldProj._43=-zn*zf/(zf-zn);
    if(abs(cellx-OldX)+abs(celly-OldY)==1) {
        if(cellx!=OldX) {
            if(cellx>OldX) {
                for(int y=0;y<=Options.DL_Distance*2;y++) {
                    for(int x=0;x<=Options.DL_Distance*2;x++) {
                        if(x==0)FreeCell(x,y);
                        if(x==Options.DL_Distance*2) LoadSingleCell(cellx+x-Options.DL_Distance,celly+y-Options.DL_Distance,x,y);
						else aDistantLand(x,y)=aDistantLand(x+1,y);
                    }
                }
            } else {
                for(int y=0;y<=Options.DL_Distance*2;y++) {
                    for(int x=Options.DL_Distance*2;x>=0;x--) {
                        if(x==Options.DL_Distance*2) FreeCell(x,y);
                        if(x==0) LoadSingleCell(cellx+x-Options.DL_Distance,celly+y-Options.DL_Distance,x,y);
                        else aDistantLand(x,y)=aDistantLand(x-1,y);
                    }
                }
            }
        } else {
            if(celly>OldY) {
                for(int y=0;y<=Options.DL_Distance*2;y++) {
                    for(int x=0;x<=Options.DL_Distance*2;x++) {
                        if(y==0) FreeCell(x,y);
                        if(y==Options.DL_Distance*2) LoadSingleCell(cellx+x-Options.DL_Distance,celly+y-Options.DL_Distance,x,y);
                        else aDistantLand(x,y)=aDistantLand(x,y+1);
                    }
                }
            } else {
                for(int y=Options.DL_Distance*2;y>=0;y--) {
                    for(int x=0;x<=Options.DL_Distance*2;x++) {
                        if(y==Options.DL_Distance*2) FreeCell(x,y);
                        if(y==0) LoadSingleCell(cellx+x-Options.DL_Distance,celly+y-Options.DL_Distance,x,y);
                        else aDistantLand(x,y)=aDistantLand(x,y-1);
                    }
                }
            }
        }
    } else {
        for(int y=-Options.DL_Distance;y<=Options.DL_Distance;y++) {
            for(int x=-Options.DL_Distance;x<=Options.DL_Distance;x++) {
                FreeCell(x+Options.DL_Distance,y+Options.DL_Distance);
                LoadSingleCell(x+cellx,y+celly,x+Options.DL_Distance,y+Options.DL_Distance);
            }
        }
    }
    OldX=cellx;
    OldY=celly;
}


void Init() {
	Initialized=true;
	HANDLE h = CreateFile("data files\\distantland\\version", GENERIC_READ, 0, 0, OPEN_EXISTING, 0, 0);
	if(h==INVALID_HANDLE_VALUE) {
		ERRORMSG("Distant land was enabled, but distant land version file doesn't exist.\n"
			"You need to run the distant land setup wizard");
	} else {
		byte b=0xff;
		DWORD unused;
		ReadFile(h, &b, 1, &unused, 0);
		if(b!=MGE_DL_VERSION) {
			ERRORMSG("Distant land was enabled, but distant land files are from an old version of MGE.\n"
				"You need to rerun the distant land setup wizard");
		}
		CloseHandle(h);
	}
    DEBUGSTRING("DistantLand init");
    DistantLand=new sDistantLand[(Options.DL_Distance*2+1)*(Options.DL_Distance*2+1)];
	DistantWorld=new sDistantLand();
	DistantDefault=new sDistantLand();
    for(int x=0;x<=Options.DL_Distance*2;x++) {
        for(int y=0;y<=Options.DL_Distance*2;y++) {
            aDistantLand(x,y).vb=0;
            aDistantLand(x,y).tex=0;
        }
    }
    OldX=OldY=TempX=TempY=-999;
    SetPos=false;
	if(Options.DLSM1Mode) {
		if(FAILED(D3DXCreateEffectFromFile(RealDevice,"Data files\\shaders\\distantland\\InGame1.fx",0,0,0,0,&DL_Effect,NULL))) {
			ERRORMSG("Could not load distant land effect file");
		}
	} else {
		if(FAILED(D3DXCreateEffectFromFile(RealDevice,"Data files\\shaders\\distantland\\InGame.fx",0,0,0,0,&DL_Effect,NULL))) {
			ERRORMSG("Could not load distant land effect file");
		}
	}
	DL_Effect->SetFloatArray("rcpres",(float*)&RcpRes,2);
	ehWorldViewProj=DL_Effect->GetParameterByName(0, "worldviewproj");
	ehWorld=DL_Effect->GetParameterByName(0, "world");
	ehWorld2=DL_Effect->GetParameterByName(0, "world2");
	ehTex0=DL_Effect->GetParameterByName(0, "tex0");
	ehTex1=DL_Effect->GetParameterByName(0, "tex1");
	ehTex2=DL_Effect->GetParameterByName(0, "tex2");
	ehTex3=DL_Effect->GetParameterByName(0, "tex3");
	ehSunCol=DL_Effect->GetParameterByName(0, "SunCol");
	ehSunAmb=DL_Effect->GetParameterByName(0, "SunAmb");
	ehSunVec=DL_Effect->GetParameterByName(0, "SunVec");
	ehFogCol=DL_Effect->GetParameterByName(0, "FogCol");
	ehFogCol2=DL_Effect->GetParameterByName(0, "FogCol2");
	ehFogRange=DL_Effect->GetParameterByName(0, "FogRange");
	ehEyePos=DL_Effect->GetParameterByName(0, "EyePos");
	ehTicks=DL_Effect->GetParameterByName(0, "ticks");
	ehOffset=DL_Effect->GetParameterByName(0, "offset");
	ehTexBlend=DL_Effect->GetParameterByName(0, "TexBlend");
	

	if(FAILED(LoadWorldMesh())) ERRORMSG("Could not load world mesh for distant land");
	if(FAILED(LoadMesh("Data files\\distantland\\default",DistantDefault))) ERRORMSG("Could not load default distant land mesh");
	if(FAILED(D3DXCreateTextureFromFileEx(RealDevice,"Data files\\distantland\\world.dds",0,0,0,0,D3DFMT_UNKNOWN,D3DPOOL_DEFAULT,
		D3DX_DEFAULT,D3DX_DEFAULT,0,0,0,&DistantWorld->tex))) {
        ERRORMSG("Could not load world texture for distant land");
    }
	if(FAILED(D3DXCreateTextureFromFileEx(RealDevice,"Data files\\distantland\\default.dds",0,0,0,0,D3DFMT_UNKNOWN,D3DPOOL_DEFAULT,
		D3DX_DEFAULT,D3DX_DEFAULT,0,0,0,&DistantDefault->tex))) {
        ERRORMSG("Count not load default distant land texture");
    }
	D3DXMatrixTranslation(&DistantWorld->world,0,0,0);

	if(FAILED(RealDevice->CreateDepthStencilSurface(JointInfo.GraphicsWidth,JointInfo.GraphicsHeight,D3DFMT_D24X8,(D3DMULTISAMPLE_TYPE)0,0,true,&ReflectionZBuffer,0)))
		ERRORMSG("Failed to create distant land z-buffer");

	if(FAILED(RealDevice->CreateIndexBuffer(0xc000,D3DUSAGE_WRITEONLY,D3DFMT_INDEX16,D3DPOOL_DEFAULT,&CellIB, 0)))
		ERRORMSG("Failed to create cell index buffer");
	WORD* i;
	CellIB->Lock(0,0,(void**)&i,0);
	for(int y=0;y<64;y++) {
		for(int x=0;x<64;x++) {
			i[(x+(y*64))*6+0]=(y*65+x);
			i[(x+(y*64))*6+2]=(y*65+x+1);
			i[(x+(y*64))*6+1]=(y*65+x+65);
			i[(x+(y*64))*6+3]=(y*65+x+1);
			i[(x+(y*64))*6+5]=(y*65+x+65+1);
			i[(x+(y*64))*6+4]=(y*65+x+65);
		}
	}
	CellIB->Unlock();

	if(FAILED(D3DXCreateTextureFromFile(RealDevice,"data files\\textures\\mge\\default\\fakewater.dds",&FakeWaterTexture)))
		ERRORMSG("Failed to load distant water texture");

	if(FAILED(RealDevice->CreateIndexBuffer(12,D3DUSAGE_WRITEONLY,D3DFMT_INDEX16,D3DPOOL_DEFAULT,&WaterIB,0)))
		ERRORMSG("Failed to create distant water index buffer");
	WaterIB->Lock(0,0,(void**)&i,0);
	i[0]=0;
	i[1]=1;
	i[2]=3;
	i[3]=1;
	i[4]=2;
	i[5]=3;
	WaterIB->Unlock();

	WaterVertex* data;
	if(FAILED(RealDevice->CreateVertexBuffer(4*SIZEOFWVERT,D3DUSAGE_WRITEONLY,DL_WATERFORMAT,D3DPOOL_DEFAULT,&WaterVB1,0))
		||FAILED(RealDevice->CreateVertexBuffer(4*SIZEOFWVERT,D3DUSAGE_WRITEONLY,DL_WATERFORMAT,D3DPOOL_DEFAULT,&WaterVB2,0)))
		ERRORMSG("Failed to create distant water vertex buffer");
	WaterVB1->Lock(0,0,(void**)&data,0);
	data[0].x=500000;
	data[0].y=500000;
	data[0].z=0;
	data[1].x=-500000;
	data[1].y=500000;
	data[1].z=0;
	data[2].x=-500000;
	data[2].y=-500000;
	data[2].z=0;
	data[3].x=500000;
	data[3].y=-500000;
	data[3].z=0;
	WaterVB1->Unlock();
	WaterVB2->Lock(0,0,(void**)&data,0);
	data[0].x=(float)(Options.DL_Distance*8192 + 8192);
	data[0].y=(float)(Options.DL_Distance*8192 + 8192);
	data[0].z=0;
	data[1].x=-(float)(Options.DL_Distance*8192);
	data[1].y=(float)(Options.DL_Distance*8192 + 8192);
	data[1].z=0;
	data[2].x=-(float)(Options.DL_Distance*8192);
	data[2].y=-(float)(Options.DL_Distance*8192);
	data[2].z=0;
	data[3].x=(float)(Options.DL_Distance*8192 + 8192);
	data[3].y=-(float)(Options.DL_Distance*8192);
	data[3].z=0;
	WaterVB2->Unlock();

	LoadDistantStatics();

	DEBUGSTRING("Distant land init complete");
}

void DL_RenderLand() {
	if(!Initialized) Init();
	if(!MyBackbuffer) {
		//DEBUGSTRING("Distant land render failed because the wrong backbuffer was set");
		return;
	}
	if(!DistantLandActive&&!DL_Activate()) return;
	D3DXMATRIX tempmat;
	D3DXMatrixTranslation(&WaterWorld,EyePos.x,EyePos.y,0);
	if(Zoom||Zoomed) {
		float zf=(float)(((float)Options.DL_Distance*3)*8192.0f);
		float zn=2000;	//2000;
		CellProj=TransProj;
		CellProj._33=zf/(zf-zn);
		CellProj._43=-zn*zf/(zf-zn);
		zn=zf/10;
		zf=500000;
		WorldProj=TransProj;
		WorldProj._33=zf/(zf-zn);
		WorldProj._43=-zn*zf/(zf-zn);
		Zoomed=Zoom;
	}
    SetPos=false;
#ifndef RELEASE
    bool bbb=false;
    if(bbb) {
		LoadDistantStatics();
        //ReloadShader();
    }
#endif
    UINT passes;
    //Set up shader
    
	DL_Effect->SetFloatArray(ehSunVec,(float*)&SunVec,3);
    DL_Effect->SetFloatArray(ehSunCol,(float*)&SunCol,3);
	D3DXVECTOR4 tAmbCol(SunAmb + AmbCol);
	DL_Effect->SetFloatArray(ehSunAmb,(float*)&tAmbCol,3);
	DL_Effect->SetInt(ehFogCol,FogColour);
	float FogColTmp[3] = { ((FogColour & 0xFF0000)>>16)/255.0f, ((FogColour & 0xFF00)>>8)/255.0f, ((FogColour & 0xFF)>>0)/255.0f };
	DL_Effect->SetFloatArray(ehFogCol2,FogColTmp, 3);
	DL_Effect->SetFloat(ehFogRange, (float)(Options.DL_Fog*8192));
	DWORD ticks=GetTickCount();
	DL_Effect->SetFloat(ehTicks,((float)(ticks%2000)/2000.0f)*6.28318530717f);
	DL_Effect->SetFloat(ehTexBlend,(float)(ticks%100)/100.0f);
	DL_Effect->SetTexture(ehTex1,RealWaterTextures[(ticks%3200 - ticks%100)/100]);
	DL_Effect->SetTexture(ehTex2,RealWaterTextures[(((ticks%3200 - ticks%100)/100) + 1)%32]);
	DL_Effect->SetTexture(ehTex0,ShaderStruct.lastpass);
	float offset[2] = { -EyePos.x,-EyePos.y };
	DL_Effect->SetFloatArray(ehEyePos,(float*)&EyePos,3);
	DL_Effect->SetFloatArray(ehOffset,(float*)&offset,2);
    
	//Render land
	DL_Effect->Begin(&passes,0);
	DL_Effect->BeginPass(PASS_SETUP);
	DL_Effect->EndPass();

	//Store the current FVF
	DWORD FVF;
	RealDevice->GetFVF(&FVF);
	RealDevice->SetFVF(DL_VERTEXFORMAT);

	//Create relection texture
	if(Options.ReflectiveWater&&EyePos.z>0) {
		D3DXMATRIX reflectionview;
		if(Options.UseDistantStatics&&Options.NearStaticsReflect) {
			D3DXVECTOR3 refeye(EyePos.x, EyePos.y, -EyePos.z);
			D3DXVECTOR3 lookat(EyeVec.x, EyeVec.y, -EyeVec.z);
			D3DXVECTOR3 up(0,0,-1);
			lookat*=100;
			lookat+=refeye;
			D3DXMatrixLookAtLH(&reflectionview, &refeye, &lookat, &up);
		}

		tempmat=DistantWorld->world*TransView*WorldProj;
		DL_Effect->SetMatrix(ehWorldViewProj,&tempmat);
		DL_Effect->SetTexture(ehTex3,0);

		DL_Effect->BeginPass(PASS_REFLECTLAND);

		RealDevice->SetRenderTarget(0,ShaderStruct.lastpassSurface);
		RealDevice->SetDepthStencilSurface(ReflectionZBuffer);
		RealDevice->Clear(0,0,D3DCLEAR_ZBUFFER|D3DCLEAR_TARGET,SunAmbD,1,0);

		RealDevice->SetIndices(WorldIB);
		RealDevice->SetStreamSource(0,DistantWorld->vb,0,SIZEOFVERT);
		DL_Effect->CommitChanges();
		RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,WorldVerts,0,WorldFaces);

		if(Options.UseDistantStatics&&Options.NearStaticsReflect&&Options.FarStaticsReflect) {
			DL_Effect->EndPass();
			DL_Effect->BeginPass(PASS_REFLECTSTATICS);
			RenderDistantStatics(&reflectionview, &WorldProj, true);
			DL_Effect->EndPass();
			DL_Effect->BeginPass(PASS_REFLECTLAND);
		}

		RealDevice->SetIndices(CellIB);
		RealDevice->Clear(0,0,D3DCLEAR_ZBUFFER,0,1,0);
		for(int x=0;x<=Options.DL_Distance*2;x++) {
			for(int y=0;y<=Options.DL_Distance*2;y++) {
				if(aDistantLand(x,y).tex!=0) {
					tempmat=aDistantLand(x,y).world*TransView*CellProj2;
					DL_Effect->SetMatrix(ehWorldViewProj,&tempmat);
					DL_Effect->SetTexture(ehTex3,aDistantLand(x,y).tex);
					DL_Effect->CommitChanges();
					RealDevice->SetStreamSource(0,aDistantLand(x,y).vb,0,SIZEOFVERT);
					RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,0x1081,0,0x2000);
				}
			}
		}

		if(Options.UseDistantStatics&&Options.NearStaticsReflect) {
			DL_Effect->EndPass();
			DL_Effect->BeginPass(PASS_REFLECTSTATICS);
			RenderDistantStatics(&reflectionview, &CellProj2, false);
			DL_Effect->EndPass();
			DL_Effect->BeginPass(PASS_REFLECTLAND);
		}

		DL_Effect->EndPass();
		RealDevice->SetRenderTarget(0,BackBuffer);
		RealDevice->SetDepthStencilSurface(DepthBuffer);

		DL_Effect->SetTexture(ehTex3,ShaderStruct.lastpass);
	}

	//Draw the world mesh
	RealDevice->Clear(0,0,D3DCLEAR_ZBUFFER,0,1,0);

	tempmat=DistantWorld->world*TransView*WorldProj;
    DL_Effect->SetMatrix(ehWorldViewProj,&tempmat);
	DL_Effect->SetTexture(ehTex3, DistantWorld->tex);
	DL_Effect->BeginPass(1);
	if(Options.Fog) {
		RealDevice->SetRenderState(D3DRS_FOGENABLE,Options.Fog);
		RealDevice->SetRenderState(D3DRS_FOGCOLOR,FogColour);
	}
	RealDevice->SetIndices(WorldIB);
	RealDevice->SetStreamSource(0,DistantWorld->vb,0,SIZEOFVERT);
	RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,WorldVerts,0,WorldFaces);

	DL_Effect->EndPass();

	if(Options.UseDistantStatics) {
		DL_Effect->BeginPass(PASS_RENDERSTATICS);
		RenderDistantStatics(&TransView, &WorldProj, true);
		DL_Effect->EndPass();
	}

	//Draw water around the world mesh
	tempmat=WaterWorld*TransView*WorldProj;
	DL_Effect->SetMatrix(ehWorldViewProj,&tempmat);
	RealDevice->SetIndices(WaterIB);
	RealDevice->SetStreamSource(0,WaterVB1,0,SIZEOFWVERT);
	RealDevice->SetFVF(DL_WATERFORMAT);
	if(EyePos.z>0) {
		if(Options.ReflectiveWater) {
			if(Options.UseSM3Water) {
				DL_Effect->BeginPass(PASS_SM3WATER);
				RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,4,0,2);
				DL_Effect->EndPass();
			} else {
				DL_Effect->BeginPass(PASS_SM2WATER);
				RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,4,0,2);
				DL_Effect->EndPass();
				DL_Effect->BeginPass(3);
				RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,4,0,2);
				DL_Effect->EndPass();
			}
		} else {
			DL_Effect->BeginPass(PASS_NONREFLECTIVEWATER);
			RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,4,0,2);
			DL_Effect->EndPass();
		}
	} else {
		DL_Effect->BeginPass(PASS_UNDERWATER);
		RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,4,0,2);
		DL_Effect->EndPass();
	}

	//Draw the cell meshes
	RealDevice->SetIndices(CellIB);
	RealDevice->SetFVF(DL_VERTEXFORMAT);
	
	DL_Effect->BeginPass(1);

    RealDevice->Clear(0,0,D3DCLEAR_ZBUFFER,0,1,0);
    for(int x=0;x<=Options.DL_Distance*2;x++) {
        for(int y=0;y<=Options.DL_Distance*2;y++) {
            tempmat=aDistantLand(x,y).world*TransView*CellProj;
            DL_Effect->SetMatrix(ehWorldViewProj,&tempmat);
            if(aDistantLand(x,y).tex==0) {
				DL_Effect->SetTexture(ehTex3,DistantDefault->tex);
				DL_Effect->CommitChanges();
				RealDevice->SetStreamSource(0,DistantDefault->vb,0,SIZEOFVERT);
            } else {
				DL_Effect->SetTexture(ehTex3,aDistantLand(x,y).tex);
				DL_Effect->CommitChanges();
				RealDevice->SetStreamSource(0,aDistantLand(x,y).vb,0,SIZEOFVERT);
            }
			RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,0x1081,0,0x2000);
        }
    }
	DL_Effect->EndPass();

	if(Options.UseDistantStatics) {
		DL_Effect->BeginPass(PASS_RENDERSTATICS);
		RenderDistantStatics(&TransView, &CellProj, false);
		DL_Effect->EndPass();
	}

	//Draw the water around the cell meshes
	tempmat=aDistantLand(Options.DL_Distance,Options.DL_Distance).world*TransView*CellProj;
	offset[0] = -(float)((OldX)*8192);
	offset[1] = -(float)((OldY)*8192);
	DL_Effect->SetMatrix(ehWorldViewProj,&tempmat);
	DL_Effect->SetFloatArray(ehOffset,offset,2);
	RealDevice->SetIndices(WaterIB);
	RealDevice->SetStreamSource(0,WaterVB2,0,SIZEOFWVERT);
	RealDevice->SetFVF(DL_WATERFORMAT);
	if(EyePos.z>0) {
		if(Options.ReflectiveWater) {
			if(Options.UseSM3Water) {
				DL_Effect->BeginPass(PASS_SM3WATER);
				RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,4,0,2);
				DL_Effect->EndPass();
			} else {
				DL_Effect->BeginPass(PASS_SM2WATER);
				RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,4,0,2);
				DL_Effect->EndPass();
				DL_Effect->BeginPass(3);
				RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,4,0,2);
				DL_Effect->EndPass();
			}
		} else {
			DL_Effect->BeginPass(PASS_NONREFLECTIVEWATER);
			RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,4,0,2);
			DL_Effect->EndPass();
		}
	} else {
		DL_Effect->BeginPass(PASS_UNDERWATER);
		RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,4,0,2);
		DL_Effect->EndPass();
	}

	//Restore state
	DL_Effect->End();
	RealDevice->SetFVF(FVF);
	RealDevice->Clear(0,0,D3DCLEAR_ZBUFFER,0,1,0);
}
void DL_RenderWater() {
	if(!DistantLandActive) return;
	DWORD FVF;
	RealDevice->GetFVF(&FVF);
	RealDevice->SetFVF(DL_WATERFORMAT);
	UINT passes;
	D3DXMATRIX tempmat=aDistantLand(Options.DL_Distance,Options.DL_Distance).world*TransView*TransProj;
	DL_Effect->SetMatrix(ehWorldViewProj,&tempmat);
	DL_Effect->Begin(&passes,0);
	DL_Effect->BeginPass(0);
	DL_Effect->EndPass();
	RealDevice->SetIndices(WaterIB);
	RealDevice->SetStreamSource(0,WaterVB2,0,SIZEOFWVERT);
	RealDevice->SetFVF(DL_WATERFORMAT);
	if(EyePos.z>0) {
		if(Options.ReflectiveWater) {
			if(Options.UseSM3Water) {
				DL_Effect->BeginPass(PASS_SM3WATER);
				RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,4,0,2);
				DL_Effect->EndPass();
			} else {
				DL_Effect->BeginPass(PASS_SM2WATER);
				RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,4,0,2);
				DL_Effect->EndPass();
				DL_Effect->BeginPass(3);
				RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,4,0,2);
				DL_Effect->EndPass();
			}
		} else {
			DL_Effect->BeginPass(PASS_NONREFLECTIVEWATER);
			RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,4,0,2);
			DL_Effect->EndPass();
		}
	} else {
		DL_Effect->BeginPass(PASS_UNDERWATER);
		RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,4,0,2);
		DL_Effect->EndPass();
	}
	DL_Effect->End();
	RealDevice->SetFVF(FVF);
}
void DL_PlayerPos(float x,float y) {
	if(!Initialized) Init();
    if(SetPos) return;
    SetPos=true;
    short xpos=(short)(floor(x/8192.0f));
    short ypos=(short)(floor(y/8192.0f));
    if(xpos!=OldX||ypos!=OldY) {
        if(TempX==xpos&&TempY==ypos) {
            ChangeCell(xpos,ypos);
        } else {
            TempX=xpos;
            TempY=ypos;
        }
    }
}
bool DL_Activate() {
	if(!Options.UseDistantLand||!UsingDistantLand) return false;
	if(DistantLandActive) return false;
	if(!IsExterior()||IsMenu()) return false;
	if(WaterLoadedCount<32) {
		DEBUGSTRING("Distantland activation failed because water textures are not yet loaded");
		return false;
	}
	if(!LoadedMemMap) {
		DEBUGSTRING("Distantland activation failed because the morrowind memmap has not yet been loaded");
		return false;
	}
	DistantLandActive=true;
	RealDevice->SetRenderState(D3DRS_FOGSTART,F2DW(Options.DL_Distance*8192*FogRange));
	RealDevice->SetRenderState(D3DRS_FOGEND,F2DW(Options.DL_Fog*8192*FogRange));
	for(int i=0;i<32;i++) *MorrowWaterPtrs[i]=FakeWaterTexture;
	return true;
}
void DL_Deactivate() {
	if(!DistantLandActive) return;
	DistantLandActive=false;
	RealDevice->SetRenderState(D3DRS_FOGSTART,F2DW(FogStart*FogRange));
	RealDevice->SetRenderState(D3DRS_FOGEND,F2DW(FogEnd*FogRange));
	for(int i=0;i<32;i++) *MorrowWaterPtrs[i]=RealWaterTextures[i];
}
