#define _MWSEfunctions
#include "d3d8header.h"
#include "MME function defs.h"

#define bSize 4096

static BYTE PipeBuffer[bSize];  //Just allocate it now to avoid having to repeatedly allocate/deallocate it each frame
static BYTE PipeOutBuffer[bSize];
static char ActiveModel[256];
static bool UsingModel=false;

static char ActiveHUDbit[33];
static bool UsingHUDbit=false;

extern HANDLE MWSEpipe2;
void MWSE_Update2(HANDLE);

int MWSE_HUDIndex(char* p,char* file) {
    if(strlen(p)>32) return -1;
    BYTE FirstFree=0xFF;
    for(int i=0;i<MAXMWSEOBJECTS;i++) {
        if(HUDbits[i].texture) { if(!strcmp(p,HUDbits[i].Tag)) return i; }
        else if(FirstFree==0xFF) FirstFree=i;
    }
    if(file!=NULL) {
        if(FirstFree!=0xFF) {
            int index;
            if((index=MWSE_TextureIndex(file,true))==-1) {
                DEBUGSTRING("MGE: Could not create HUD element because the texture could not be loaded");
                return -1;
            }
            HUDbits[FirstFree].texture=LoadedTextures[index].texture;
            HUDbits[FirstFree].texture->AddRef();
            strcpy_s(HUDbits[FirstFree].Tag,33,p);
            HUDbits[FirstFree].scale.x=1.0f;
            HUDbits[FirstFree].scale.y=1.0f;
            HUDbits[FirstFree].position.x=0;
            HUDbits[FirstFree].position.y=0;
            DEBUGSTRING("MGE: HUD element loaded successfully");
            return FirstFree;
        } else {
            DEBUGSTRING("MGE: Failed to load HUD element because there is no more room");
        }
    }
    return -1;
}
void MWSE_FreeHUD(int index) {
    if(HUDbits[index].texture!=NULL) HUDbits[index].texture->Release();
    else return;
    if(HUDbits[index].effect!=NULL) HUDbits[index].effect->Release();
    HUDbits[index].Enabled=false;
    HUDbits[index].texture=NULL;
    HUDbits[index].effect=NULL;
    HUDbits[index].Tag[0]='\0';
    DEBUGSTRING("MGE: HUD element release successful");
}
int MWSE_ObjectIndex(char* p,char* file) {
    if(strlen(p)>32) return -1;
    BYTE FirstFree=0xFF;
    for(int i=0;i<MAXMWSEOBJECTS;i++) {
        if(Objects[i].Created) {
            if(!strcmp(p,Objects[0].Tag)) return i;
        } else if(FirstFree==0xFF) FirstFree=i;
    }
    if(file!=NULL&&FirstFree!=0xFF&&MWSE_IsSafeString(file)) {
        char buffer[257]; buffer[0]='\0';
        strcat_s(buffer,256,"Data Files\\MGE meshes\\");
        strcat_s(buffer,256,file);
        DEBUGSTRING("MGE: Trying to load mesh");
        DEBUGSTRING(buffer);
        ID3DXMesh* TempMesh=NULL;
        ID3DXBuffer* Materials=NULL;
        DWORD subsets;
        ID3DXBuffer* Effects=NULL;
        HRESULT hr=D3DXLoadMeshFromXA(buffer,D3DXMESH_MANAGED,RealDevice,NULL,&Materials,
            &Effects,&subsets,&TempMesh);
        if(Effects) Effects->Release();
        if(hr!=D3D_OK) {
            DEBUGSTRING("MGE: Failed to load mesh");
        } else if(subsets>MAXMWSESUBSETS) {
            DEBUGSTRING("MGE: Could not load mesh because it contained to many subsets");
            if(TempMesh!=NULL) TempMesh->Release();
            if(Materials!=NULL) Materials->Release();
        } else {
            ZeroMemory(&Objects[FirstFree],sizeof(MWSEobject));
            strcpy_s(Objects[FirstFree].Tag,33,p);
            Objects[FirstFree].Created=true;
            Objects[FirstFree].IsMesh=true;
            D3DXMATERIAL* Mat=(D3DXMATERIAL*)Materials->GetBufferPointer();
            if(Materials!=NULL) {
                for(UINT i=0;i<subsets;i++) {
                    Objects[FirstFree].materials[i]=Mat[i].MatD3D;
                    if(Mat[i].pTextureFilename!=NULL&&Mat[i].pTextureFilename[0]!='\0') {
                        int index=MWSE_TextureIndex(Mat[i].pTextureFilename,true);
                        if(index!=-1) {
                            Objects[FirstFree].textures[i]=LoadedTextures[index].texture;
                            LoadedTextures[index].texture->AddRef();
                        }
                    }
                }
                Materials->Release();
            }
            Objects[FirstFree].mesh=TempMesh;
            Objects[FirstFree].subsets=subsets;
            D3DXMatrixRotationYawPitchRoll(&Objects[FirstFree].RotMatrix,0,0,0);
            D3DXMatrixScaling(&Objects[FirstFree].ScaleMatrix,1,1,1);
            DEBUGSTRING("MGE: Mesh loaded successfully");
            UsedMWSEobjects++;
            return FirstFree;
        }
    }
    return -1;
}
void MWSE_FreeModel(int index) {
    if(Objects[index].Created) UsedMWSEobjects--;
    Objects[index].Enabled=false;
    Objects[index].Created=false;
    if(Objects[index].vBuffer!=NULL) Objects[index].vBuffer->Release();
    if(Objects[index].iBuffer!=NULL) Objects[index].iBuffer->Release();
    for(BYTE i=0;i<MAXMWSESUBSETS;i++)
        if(Objects[index].textures[i]!=NULL) Objects[index].textures[i]->Release();
    if(Objects[index].effect!=NULL) Objects[index].effect->Release();
    Objects[index].Tag[0]='\0';
    Objects[index].mesh->Release();
    DEBUGSTRING("MGE: Mesh release successful");
}
void MWSE_UpdateObjectMatrix(int index,BYTE type) { //0 - pos, 1 - rot, 2 - scale
    switch(type) {
        case 0:
            D3DXMatrixTranslation(&Objects[index].PosMatrix,Objects[index].pos.x,
                Objects[index].pos.y,Objects[index].pos.z);
            break;
        case 1:
            D3DXMatrixRotationYawPitchRoll(&Objects[index].RotMatrix,Objects[index].rot.x,
                Objects[index].rot.y,Objects[index].rot.z);
            break;
        case 2:
            D3DXMatrixScaling(&Objects[index].ScaleMatrix,Objects[index].scale.x,
                Objects[index].scale.y,Objects[index].scale.z);
            break;
    }  
    Objects[index].WorldMatrix=Objects[index].ScaleMatrix*Objects[index].RotMatrix*Objects[index].PosMatrix;
}
int MWSE_ShaderIndex(char* p,bool LoadIfNeeded) {
    BYTE FirstFree=0xFF;
    for(int i=0;i<MAXMWSESHADERS;i++) {
        if(ShaderStruct.MWSEeffects[i].effect!=NULL) {
            if(!strcmp(p,ShaderStruct.MWSEeffects[i].FilePath)) return i;
        } else if(FirstFree==0xFF) FirstFree=i;
    }
    if(LoadIfNeeded&&FirstFree!=0xFF&&MWSE_IsSafeString(p)) {
        char buffer[257]; buffer[0]='\0';
        strcat_s(buffer,256,"Data Files\\shaders\\");
        strcat_s(buffer,256,p);
        DEBUGSTRING("MGE: Trying to load shader");
        DEBUGSTRING(buffer);
        HRESULT hr=D3DXCreateEffectFromFileA(RealDevice, buffer,0,0,0,0,
            &ShaderStruct.MWSEeffects[FirstFree].effect, NULL);
        if(hr==D3D_OK) {
            DEBUGSTRING("MGE: Shader loaded");
            strcpy_s(ShaderStruct.MWSEeffects[FirstFree].FilePath,MAXFILEPATHLENGTH+1,p);
            LoadShaderTextures(ShaderStruct.MWSEeffects[FirstFree].effect);
            return FirstFree;
        } else {
            DEBUGSTRING("MGE: Error loading shader");
        }
    }
    return -1;
}
void MWSE_FreeEffect(int index) {
    if(ShaderStruct.MWSEeffects[index].effect==NULL) return;
    ShaderStruct.MWSEeffects[index].Active=false;
    ShaderStruct.MWSEeffects[index].effect->Release();
    ShaderStruct.MWSEeffects[index].effect=NULL;
    ShaderStruct.MWSEeffects[index].FilePath[0]='\0';
    DEBUGSTRING("MGE: Effect freed");
}
int MWSE_TextureIndex(char* p,bool LoadIfNeeded) {
    BYTE FirstFree=0xFF;
    for(int i=0;i<MAXMWSESHADERS;i++) {
        if(LoadedTextures[i].texture!=NULL) {
            if(!strcmp(p,LoadedTextures[i].FilePath)) return i;
        } else if(FirstFree==0xFF) FirstFree=i;
    }
    if(LoadIfNeeded&&FirstFree!=0xFF&&MWSE_IsSafeString(p)) {
        char buffer[256]; buffer[0]='\0';
        HRESULT hr;
        if(p[0]=='#') {
            if(!MWSE_IsSafeString(p+1)) return -1;
            strcat_s(buffer,256,"Data Files\\MGE videos\\");
            strcat_s(buffer,256,(p+1));
            DEBUGSTRING("MGE: Trying to load directshow texture");
            DEBUGSTRING(buffer);
            LoadedTextures[FirstFree].dsInfo=CreateDSGraph(buffer,&LoadedTextures[FirstFree].texture);
            if(LoadedTextures[FirstFree].texture) {
                DEBUGSTRING("MGE: Directshow texture loaded");
                strcpy_s(LoadedTextures[FirstFree].FilePath,MAXFILEPATHLENGTH+1,p);
                return FirstFree;
            } else {
                DEBUGSTRING("MGE: Error loading Directshow texture");
            }
        } else {
            strcat_s(buffer,256,"Data Files\\Textures\\");
            strcat_s(buffer,256,p);
            DEBUGSTRING("MGE: Trying to load texture");
            DEBUGSTRING(buffer);
            hr=D3DXCreateTextureFromFileA(RealDevice, buffer,&LoadedTextures[FirstFree].texture);
            if(hr==D3D_OK) {
                DEBUGSTRING("MGE: Texture loaded");
                strcpy_s(LoadedTextures[FirstFree].FilePath,MAXFILEPATHLENGTH+1,p);
                return FirstFree;
            } else {
                DEBUGSTRING("MGE: Error loading texture");
            }
        }
    }
    return -1;
}
void MWSE_FreeTexture(int index) {
    if(LoadedTextures[index].texture) LoadedTextures[index].texture->Release();
    LoadedTextures[index].texture=NULL;
    if(LoadedTextures[index].dsInfo) {
        FreeMovie(LoadedTextures[index].dsInfo);
        delete LoadedTextures[index].dsInfo;
        LoadedTextures[index].dsInfo=NULL;
    }
    LoadedTextures[index].FilePath[0]='\0';
    DEBUGSTRING("MGE: texture freed");
}
bool MWSE_IsSafeString(char* str) {
#define NOTSAFESTRING { DEBUGSTRING("MGE: file path failed security validation check"); return false; }
    int pointer=0;
    char c;
    if(str[0]=='.'||str[0]=='\\') NOTSAFESTRING
    bool LastWasDot=true;
    while((c=str[pointer++])!='\0') {
        if(!((c>='a'&&c<='z')||(c>='0'||c<='9')||c=='.'||c=='_'||c=='\\'||c==' '||c=='#')) NOTSAFESTRING
        if(c=='.') {
            if(LastWasDot) NOTSAFESTRING
            else LastWasDot=true;
        } else LastWasDot=false;
    }
    if(pointer>MAXFILEPATHLENGTH) NOTSAFESTRING
    else return true;
#undef NOTSAFESTRING
}
void MWSE_GetString(char* buffer,DWORD* pPointer) {
    BYTE pos=0;
    DWORD Pointer=*pPointer;
    while(PipeBuffer[Pointer]!='\0'&&pos<255) {
        buffer[pos++]=PipeBuffer[Pointer++];
    }
    buffer[pos]='\0';
    *pPointer=Pointer+1;
    _strlwr_s(buffer,256);
}
int MWSE_TextureHookIndex(char* str) {
   int index=-1;
   for(int i=0;i<MAXHOOKEDTEXTURES;i++) {
       if(!strcmp(str,HookedTextures[i].path)) {
            index=i;
            break;
       }
   }
   return index;
}

void DoMWSEFunction() { MWSE_Update2(MWSEpipe2); }
void MWSE_Update() { MWSE_Update2(MWSEpipe); }
void MWSE_Update2(HANDLE MWSEpipe) {
    DWORD size;
    DWORD pointer=0;
    DWORD BytesLeft;
    DWORD BytesToSend=0;
    PeekNamedPipe(MWSEpipe,NULL,0,NULL,&BytesLeft,NULL);
    if(BytesLeft&&ReadFile(MWSEpipe,PipeBuffer,BytesLeft,&size,NULL)&&size>0) {
        DEBUGSTRING("MGE: Read MWSE command");
        while(pointer<size) {
            DEBUGSTRING("MGE: Parsed command");
            switch(PipeBuffer[pointer++]) {
                case 'd':
                case 0: //special debuging commands
#pragma region Debug
                {
                    DEBUGSTRING("MGE: Recieved special command");
                    switch(PipeBuffer[pointer++]) {
                        case 'w':   //Complete wipe
                            if(size>4080) {
                                DEBUGSTRING("MGE: Recieved too many commands in one frame to use '\0w'");
                            } else {
                                PipeBuffer[size]='m';
                                PipeBuffer[size+1]='w';
                                PipeBuffer[size+2]='t';
                                PipeBuffer[size+3]='w';
                                PipeBuffer[size+4]='S';
                                PipeBuffer[size+5]='w';
                                PipeBuffer[size+7]='H';
                                PipeBuffer[size+8]='w';
                                size+=8;
                            }
                            break;
                        case 'p':   //DebugString
                            {
                                DEBUGSTRING("MGE: Script is trying to print a debugging comment");
                                char buffer[256];
                                MWSE_GetString(buffer,&pointer);
                                DEBUGSTRING(buffer);
                                break;
                            }
                        case 'v':
                            DEBUGSTRING(VERSION);
                            break;
                        case 'V':
                            DEBUGSTRING("MGE: A script has requested MGE's version number");
                            *((DWORD*)&PipeOutBuffer[BytesToSend])=MGE_EX_VERSION;
                            BytesToSend+=4;
                            break;
                        case 'x':
                            DEBUGSTRING("MGE: A script has requested the screen width");
                            *((WORD*)&PipeOutBuffer[BytesToSend])=JointInfo.GraphicsWidth;
                            BytesToSend+=2;
                            break;
                        case 'y':
                            DEBUGSTRING("MGE: A script has requested the screen height");
                            *((WORD*)&PipeOutBuffer[BytesToSend])=JointInfo.GraphicsHeight;
                            BytesToSend+=2;
                            break;
                        case 'f':
                            DEBUGSTRING("MGE: A script has requested the current fps");
                            *((float*)&PipeOutBuffer[BytesToSend])=fps;
                            BytesToSend+=4;
                            break;
						case 'P':
							DEBUGSTRING("MGE: A script has requested the players viewing direction");
							*((float*)&PipeOutBuffer[BytesToSend+0])=EyeVec.x;
							*((float*)&PipeOutBuffer[BytesToSend+4])=EyeVec.y;
							*((float*)&PipeOutBuffer[BytesToSend+8])=EyeVec.z;
							BytesToSend+=12;
                        case '9':
                            DEBUGSTRING("MGE: A script has requested if this is a DX9 build of MGE");
                            *((WORD*)&PipeOutBuffer[BytesToSend])=1;
                            BytesToSend+=2;
                            break;
                        case 'm':
                            DEBUGSTRING("MGE: A script has requested if this is a NoMem build of MGE");
                            *((float*)&PipeOutBuffer[BytesToSend])=0;
                            BytesToSend+=2;
                            break;
                        default:
                            DEBUGSTRING("MGE: Special command not recognized");
                            break;
                    }
                    break;
                }
#pragma endregion
				case 'D': //Distant land
#pragma region DistantLand
				{
					DEBUGSTRING("MGE: Recieved a distant land command");
					char Command=PipeBuffer[pointer++];
					switch(Command) {
						case 'd':   //Toggle distant land
							DEBUGSTRING("MGE: A script is trying to disable distant land");
							UsingDistantLand=false;
							DL_Deactivate();
							break;
						case 'e':
							DEBUGSTRING("MGE: A script is trying to enable distant land");
							UsingDistantLand=true;
							DL_Activate();
							break;
					}
				}
#pragma endregion
                case 'T':   //Texture hook
#pragma region TextureHook
                {
                    char CharPointer[256];
                    char Command=PipeBuffer[pointer++];
                    MWSE_GetString(CharPointer,&pointer);
                    DEBUGSTRING("MGE: Recieved a texture hook command");
                    DEBUGSTRING(CharPointer);
                    int index = MWSE_TextureHookIndex(CharPointer);
                    if(index==-1) {
                        DEBUGSTRING("MGE: texture hook command failed because video does not exist");
                        break;
                    }
                    switch(Command) {
                        case 'a':
                            if(HookedTextures[index].AutoClean) {
                                DEBUGSTRING("MGE: Texture hook is already set to auto-free");
                            } else {
                                DEBUGSTRING("MGE: Texture hook will freed automatically");
                                HookedTextures[index].AutoClean=true;
                                HookedTextures[index].tHook->Release();
                            }
                            break;
                        case 'd':
                            if(!HookedTextures[index].AutoClean) {
                                DEBUGSTRING("MGE: Texture hook is already set to don't-free");
                            } else {
                                DEBUGSTRING("MGE: Texture hook will no longer be freed automatically");
                                HookedTextures[index].AutoClean=false;
                                HookedTextures[index].tHook->AddRef();
                            }
                            break;
                    }
                }
                break;
#pragma endregion
                case 'v':   //Video
#pragma region video
                {
                    char CharPointer[256];
                    char Command=PipeBuffer[pointer++];
                    MWSE_GetString(CharPointer,&pointer);
                    DEBUGSTRING("MGE: Recieved a video command");
                    DEBUGSTRING(CharPointer);
                    int index = MWSE_TextureIndex(CharPointer, true);
                    if(index==-1) {
                        DEBUGSTRING("MGE: video command failed because video does not exist");
                        if(PipeBuffer[pointer]=='S') pointer+=4;
                    } else if(LoadedTextures[index].dsInfo==NULL) {
                        DEBUGSTRING("MGE: video command failed because the specified file is not a video");
                        if(PipeBuffer[pointer]=='S') pointer+=4;
                    } else {
                        switch(Command) {
                            case 'r':
                                DEBUGSTRING("MGE: Resuming video");
                                ResumeMovie(LoadedTextures[index].dsInfo);
                                break;
                            case 'p':
                                DEBUGSTRING("MGE: Pausing video");
                                PauseMovie(LoadedTextures[index].dsInfo);
                                break;
                            case 's':
                                DEBUGSTRING("MGE: Stopping video");
                                StopMovie(LoadedTextures[index].dsInfo);
                                break;
                            case 'R':
                                DEBUGSTRING("MGE: Rewinding video");
                                RewindMovie(LoadedTextures[index].dsInfo);
                                break;
                            case 'S':
                                DEBUGSTRING("MGE: Seeking video");
                                DWORD d = *(DWORD*)(&PipeBuffer[pointer]);
                                SeekMovie(LoadedTextures[index].dsInfo,d);
                                pointer+=4;
                                break;
                        }
                    }
                    break;
                }
#pragma endregion
                case 'H':   //HUD
#pragma region HUD
                {
                    char CharPointer[256];
                    int index;
                    DEBUGSTRING("MGE: Recieved a HUD element command");
                    switch(PipeBuffer[pointer++]) {
                        case 'w':  // wipe all loaded HUD elements
                            DEBUGSTRING("MGE: Unloading all loaded HUD elements");
                            for(UINT i=0;i<MAXMWSEOBJECTS;i++)
                                if(HUDbits[i].texture) MWSE_FreeHUD(i);
                            break;
                        case 'W':  //with HUD element do
                            MWSE_GetString(CharPointer,&pointer);
                            DEBUGSTRING("MGE: All following HUD commands will refer to object");
                            DEBUGSTRING(CharPointer);
                            if(strlen(CharPointer)>33) {
                                DEBUGSTRING("MGE: 'HW' failed because object id was more than 32 characters");
                            } else {
                                strcpy_s(ActiveHUDbit,33,CharPointer);
                                UsingHUDbit=true;
                            }
                            break;
                        case 'c': //continue
                            DEBUGSTRING("MGE: All following model commands must specify their own object");
                            UsingHUDbit=false;
                            break;
                        case 'l':  //Assign and load HUD
                        {
                            char FileName[256];
                            if(!UsingHUDbit) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveHUDbit);
                            MWSE_GetString(FileName,&pointer);
                            DEBUGSTRING("MGE: Trying to create a new HUD element");
                            DEBUGSTRING(CharPointer);
                            DEBUGSTRING(FileName);
                            MWSE_HUDIndex(CharPointer,FileName);
                            break;
                        }
                        case 'p': //Position
                            if(!UsingHUDbit) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveHUDbit);
                            DEBUGSTRING("MGE: Trying to position HUD element");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_HUDIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: Position failed because HUD id does not exist");
                            } else {
                                HUDbits[index].position.x=*(float*)(&PipeBuffer[pointer]);
                                HUDbits[index].position.y=*(float*)(&PipeBuffer[pointer+4]);
                                DEBUGSTRING("MGE: Position successful");
                            }
                            pointer+=8;
                            break;
                        case 's': //Scale
                            if(!UsingHUDbit) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveHUDbit);
                            DEBUGSTRING("MGE: Trying to scale HUD element");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_HUDIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: Scale failed because HUD id does not exist");
                            } else {
                                HUDbits[index].scale.x=*(float*)(&PipeBuffer[pointer]);
                                HUDbits[index].scale.y=*(float*)(&PipeBuffer[pointer+4]);
                                DEBUGSTRING("MGE: Scale successful");
                            }
                            pointer+=8;
                            break;
                        case 'F': //Make fullscreen
                            if(!UsingHUDbit) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveHUDbit);
                            DEBUGSTRING("MGE: Trying to make HUD element full screen");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_HUDIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: Scale failed because HUD id does not exist");
                            } else {
                                D3DSURFACE_DESC desc;
                                HUDbits[index].texture->GetLevelDesc(0,&desc);
                                HUDbits[index].scale.x=(float)JointInfo.GraphicsWidth/(float)desc.Width;
                                HUDbits[index].scale.y=(float)JointInfo.GraphicsHeight/(float)desc.Height;
                                HUDbits[index].position.x=0;
                                HUDbits[index].position.y=0;
                            }
                            break;
                        case 'e': //enable
                            if(!UsingHUDbit) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveHUDbit);
                            DEBUGSTRING("MGE: Trying to enable HUD element");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_HUDIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: Enable failed because HUD id does not exist");
                            } else {
                                if(HUDbits[index].Enabled) { 
                                    DEBUGSTRING("MGE: Object already enabled");
                                } else {
                                    UsedHUDbits++;
                                }
                                HUDbits[index].Enabled=true;
                                DEBUGSTRING("MGE: Enable successful");
                            }
                            break;
                        case 'd': //disable
                            if(!UsingHUDbit) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveHUDbit);
                            DEBUGSTRING("MGE: Trying to disable HUD element");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_HUDIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: Disable failed because HUD id does not exist");
                            } else {
                                if(!HUDbits[index].Enabled) {
                                    DEBUGSTRING("MGE: Object already disabled");
                                } else {
                                    UsedHUDbits--;
                                }
                                HUDbits[index].Enabled=false;
                                DEBUGSTRING("MGE: Disable successful");
                            }
                            break;
                        case 't': //SetTexture
                        {
                            if(!UsingHUDbit) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveHUDbit);
                            DEBUGSTRING("MGE: Trying to texture HUD element");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_HUDIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: Set texture failed because HUD id does not exist");
                            }
                            MWSE_GetString(CharPointer,&pointer);
                            DEBUGSTRING("MGE: Texture name");
                            DEBUGSTRING(CharPointer);
                            if(index!=-1) {
                                int index2;
                                if(CharPointer[0]=='\0') {
                                    DEBUGSTRING("MGE: Set texture failed. HUD elements must have a texture applied at all times.");
                                } else if((index2=MWSE_TextureIndex(CharPointer,true))==-1) {
                                    DEBUGSTRING("MGE: Set texture failed because texture could not be loaded");
                                } else {
                                    if(HUDbits[index].texture!=NULL) HUDbits[index].texture->Release();
                                    HUDbits[index].texture=LoadedTextures[index2].texture;
                                    HUDbits[index].texture->AddRef();
                                    DEBUGSTRING("MGE: Set texture successful");
                                }
                            }
                            break;
                        }
                        case 'E': //SetEffect
                        {
                            if(!UsingHUDbit) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveHUDbit);
                            DEBUGSTRING("MGE: Trying to bind effect to object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_HUDIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: Set effect failed because HUD id does not exist");
                            }
                            MWSE_GetString(CharPointer,&pointer);
                            DEBUGSTRING("MGE: Effect name");
                            DEBUGSTRING(CharPointer);
                            if(index!=-1) {
                                int index2;
                                if(CharPointer[0]!='\0'&&(index2=MWSE_ShaderIndex(CharPointer,true))==-1) {
                                    DEBUGSTRING("MGE: Set effect failed because effect could not be loaded");
                                } else {
                                    if(HUDbits[index].effect!=NULL) HUDbits[index].effect->Release();
                                    if(CharPointer[0]=='\0') {
                                        HUDbits[index].effect=NULL;
                                    } else {
                                        HUDbits[index].effect=ShaderStruct.MWSEeffects[index2].effect;
                                        HUDbits[index].effect->AddRef();
                                    }
                                    DEBUGSTRING("MGE: Set effect successful");
                                }
                            }
                            break;
                        }
                        case 'f': //Free sprite
                            if(!UsingHUDbit) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveHUDbit);
                            DEBUGSTRING("MGE: Trying to free HUD element");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_HUDIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("Free object failed because object id does not exist");
                            } else {
                                MWSE_FreeHUD(index);
                            }
                            break;
                        default:
                            DEBUGSTRING("MGE: HUD command not recognised");
                            break;
                    }
                    break;
                }
#pragma endregion
                case 'w':   //Weather
#pragma region Weather
                {
                    DEBUGSTRING("MGE: Recieved a weather control command");
                    switch(PipeBuffer[pointer++]) {
                        case 'm': SunManual=true; break;
                        case 'M': SunManual=false; break;
                        case 'c':
                            DEBUGSTRING("MGE: A script is manually setting the suns colour");
                            SunManColour=true;
                            SunColour.r=*(float*)(&PipeBuffer[pointer]);
                            SunColour.g=*(float*)(&PipeBuffer[pointer+4]);
                            SunColour.b=*(float*)(&PipeBuffer[pointer+8]);
                            SunColour.a=1;
                            pointer+=12;
                            break;
                        case 'C': SunManColour=false; break;
                        case 'g':
                            DEBUGSTRING("MGE: A script has requested the suns colour");
                            *((float*)&PipeOutBuffer[BytesToSend])=SunColour.r;
                            *((float*)&PipeOutBuffer[BytesToSend+4])=SunColour.g;
                            *((float*)&PipeOutBuffer[BytesToSend+8])=SunColour.b;
                            BytesToSend+=12;
                            break;
                        case 'd':
                            DEBUGSTRING("MGE: A script is manually setting the suns position");
                            SunManDir=true;
                            SunDirection.x=*(float*)(&PipeBuffer[pointer]);
                            SunDirection.y=*(float*)(&PipeBuffer[pointer+4]);
                            SunDirection.z=*(float*)(&PipeBuffer[pointer+8]);
                            pointer+=12;
                            break;
                        case 'D': SunManDir=false; break;
                        case 'G':
                            DEBUGSTRING("MGE: A script has requested the suns direction");
                            *((float*)&PipeOutBuffer[BytesToSend])=SunDirection.x;
                            *((float*)&PipeOutBuffer[BytesToSend+4])=SunDirection.y;
                            *((float*)&PipeOutBuffer[BytesToSend+8])=SunDirection.z;
                            BytesToSend+=12;
                            break;
                        case 't':
                        {
                            char FilePointer[256]; FilePointer[0]='\0';
                            MWSE_GetString(FilePointer,&pointer);
                            DEBUGSTRING("MGE: Assigning a new texture to the sun");
                            DEBUGSTRING(FilePointer);
                            int index;
                            if((index=MWSE_TextureIndex(FilePointer,true))!=-1) {
                                if(SunTexture) SunTexture->Release();
                                SunTexture=LoadedTextures[index].texture;
                                SunTexture->AddRef();
                            } else {
                                DEBUGSTRING("MGE: Could not use texture because it could not be loaded");
                            }
                            break;
                        }
                        case 'T':
                            SunTexture->Release();
                            SunTexture=NULL;
                            break;
                        case 'a':
                        {
                            DEBUGSTRING("MGE: A script is manually setting the ambient colour");
                            AmbientManual=true;
                            D3DXCOLOR NewColour;
                            NewColour.r=*(float*)(&PipeBuffer[pointer]);
                            NewColour.g=*(float*)(&PipeBuffer[pointer+4]);
                            NewColour.b=*(float*)(&PipeBuffer[pointer+8]);
                            NewColour.a=1;
                            AmbientColour=NewColour;
                            pointer+=12;
                            RealDevice->SetRenderState(D3DRS_AMBIENT,AmbientColour);
                            break;
                        }
                        case 'A': AmbientManual=false; break;
                        case 'l':
                        {
                            DEBUGSTRING("MGE: A script has requested the ambient colour");
                            D3DXCOLOR NewColour=AmbientColour;
                            *((float*)&PipeOutBuffer[BytesToSend])=NewColour.r;
                            *((float*)&PipeOutBuffer[BytesToSend+4])=NewColour.g;
                            *((float*)&PipeOutBuffer[BytesToSend+8])=NewColour.b;
                            BytesToSend+=12;
                            break;
                        }
                        case 'b':
                        {
                            DEBUGSTRING("MGE: A script is manually setting the cls colour");
                            ClsManual=true;
                            D3DXCOLOR NewColour;
                            NewColour.r=*(float*)(&PipeBuffer[pointer]);
                            NewColour.g=*(float*)(&PipeBuffer[pointer+4]);
                            NewColour.b=*(float*)(&PipeBuffer[pointer+8]);
                            NewColour.a=1;
                            ClsColour=NewColour;
                            pointer+=12;
                            break;
                        }
                        case 'B': ClsManual=false; break;
                        case 'L':
                        {
                            DEBUGSTRING("MGE: A script has requested the cls colour");
                            D3DXCOLOR NewColour=ClsColour;
                            *((float*)&PipeOutBuffer[BytesToSend])=NewColour.r;
                            *((float*)&PipeOutBuffer[BytesToSend+4])=NewColour.g;
                            *((float*)&PipeOutBuffer[BytesToSend+8])=NewColour.b;
                            BytesToSend+=12;
                            break;
                        }
                        case 's':
                        {
                            DEBUGSTRING("MGE: A script is manually setting the sky colour");
                            SkyManual=true;
                            D3DXCOLOR NewColour;
                            NewColour.r=*(float*)(&PipeBuffer[pointer]);
                            NewColour.g=*(float*)(&PipeBuffer[pointer+4]);
                            NewColour.b=*(float*)(&PipeBuffer[pointer+8]);
                            NewColour.a=1;
                            SkyColour=NewColour;
                            pointer+=12;
                            break;
                        }
                        case 'S': SkyManual=false; break;
                        case 'h':
                        {
                            DEBUGSTRING("MGE: A script is manually setting the HUD colour");
                            HUDManual=true;
                            D3DXCOLOR NewColour;
                            NewColour.r=*(float*)(&PipeBuffer[pointer]);
                            NewColour.g=*(float*)(&PipeBuffer[pointer+4]);
                            NewColour.b=*(float*)(&PipeBuffer[pointer+8]);
                            NewColour.a=1;
                            HUDColour=NewColour;
                            pointer+=12;
                            break;
                        }
                        case 'H': HUDManual=false; break;
                        default:
                            DEBUGSTRING("MGE: Unrecognised weather command");
                            break;
                    }
                    break;
                }
#pragma endregion
                case 'k': //Keyboard key control
#pragma region keyboard
                {
                    WORD code;
                    DEBUGSTRING("MGE: Recieved a direct input control command");
                    float f;
                    switch(PipeBuffer[pointer++]) {
                        case 't': code=*(WORD*)(&PipeBuffer[pointer]); pointer+=2; MWSE_TapKey(code);      break;
                        case 'p': code=*(WORD*)(&PipeBuffer[pointer]); pointer+=2; MWSE_HoldKey(code);     break;
                        case 'r': code=*(WORD*)(&PipeBuffer[pointer]); pointer+=2; MWSE_UnholdKey(code);   break;
                        case 'h': code=*(WORD*)(&PipeBuffer[pointer]); pointer+=2; MWSE_HammerKey(code);   break;
                        case 'u': code=*(WORD*)(&PipeBuffer[pointer]); pointer+=2; MWSE_UnhammerKey(code); break;
                        case 'd': code=*(WORD*)(&PipeBuffer[pointer]); pointer+=2; MWSE_DisallowKey(code); break;
                        case 'a': code=*(WORD*)(&PipeBuffer[pointer]); pointer+=2; MWSE_AllowKey(code);    break;
                        case 'H': code=*(WORD*)(&PipeBuffer[pointer]); pointer+=2; MWSE_AHammerKey(code);  break;
                        case 'U': code=*(WORD*)(&PipeBuffer[pointer]); pointer+=2; MWSE_AUnhammerKey(code);break;
                        case 'x': code=*(WORD*)(&PipeBuffer[pointer]); pointer+=2; MWSE_MoveMouseX(code);  break;
                        case 'X': code=*(WORD*)(&PipeBuffer[pointer]); pointer+=2; MWSE_MoveMouseX(-code); break;
                        case 'y': code=*(WORD*)(&PipeBuffer[pointer]); pointer+=2; MWSE_MoveMouseY(code);  break;
                        case 'Y': code=*(WORD*)(&PipeBuffer[pointer]); pointer+=2; MWSE_MoveMouseY(-code); break;
                        case 'b': code=*(WORD*)(&PipeBuffer[pointer]); pointer+=2; MWSE_BufferedDown(code);break;
                        case 'B': code=*(WORD*)(&PipeBuffer[pointer]); pointer+=2; MWSE_BufferedUp(code);  break;
                        case 'v': 
                            f=*(float*)(&PipeBuffer[pointer]);
                            pointer+=4;
                            MWSE_MouseSpeedX(f);
                            break;
                        case 'V': 
                            f=*(float*)(&PipeBuffer[pointer]);
                            pointer+=4;
                            MWSE_MouseSpeedX(-f);
                            break;
                        case 'w': 
                            f=*(float*)(&PipeBuffer[pointer]);
                            pointer+=4;
                            MWSE_MouseSpeedY(f);
                            break;
                        case 'W': 
                            f=*(float*)(&PipeBuffer[pointer]);
                            pointer+=4;
                            MWSE_MouseSpeedY(-f);
                            break;
                        case 'm': MWSE_DisableMouse();  break;
                        case 'M': MWSE_EnableMouse();   break;
                        default:
                            DEBUGSTRING("MGE: recieved unknown input control command");
                            break;
                    }
                    break;
                }
#pragma endregion
                case 'c': //Camera Shake
#pragma region CameraShake
                    DEBUGSTRING("MGE: Recieved a camera shake command");
                    switch(PipeBuffer[pointer++]) {
                        case 'e':
                            DEBUGSTRING("MGE: Camera shake enabled");
                            CameraShake=true;
                            break;
                        case 'd':
                            DEBUGSTRING("MGE: Camera shake disabled");
                            CameraShake=false;
                            break;
                        case 's':
                            DEBUGSTRING("MGE: Setting camera shake amplitude");
                            CameraShakeAmp=CLAMP(*(float*)(&PipeBuffer[pointer]),MINSHAKE,MAXSHAKE);
                            pointer+=4;
                            break;
                        case 'z':
                            DEBUGSTRING("MGE: Setting camera shake rate of change of amplitude");
                            CameraShakeChange=true;
                            CameraShakeChangeRate=*(float*)(&PipeBuffer[pointer]);
                            pointer+=4;
                            break;
                        case 'Z':
                            CameraShakeChange=false;
                            break;
                        default:
                            DEBUGSTRING("MGE: recieved unknown camera shake command");
                            break;
                    }
                    break;
#pragma endregion
                case 'm': //Model command
#pragma region model
                    DEBUGSTRING("MGE: Recieved MGE Object command");
                    char CharPointer[256]; CharPointer[0]='\0';
                    switch(PipeBuffer[pointer++]) {
                        int index;
                        case 'w':  // wipe all loaded models
                            DEBUGSTRING("MGE: Unloading all loaded models");
                            for(UINT i=0;i<MAXMWSEOBJECTS;i++)
                                if(Objects[i].Created) MWSE_FreeModel(i);
                            break;
                        case 'W':  //with model do
                            MWSE_GetString(CharPointer,&pointer);
                            DEBUGSTRING("MGE: All following model commands will refer to object");
                            DEBUGSTRING(CharPointer);
                            if(strlen(CharPointer)>32) {
                                DEBUGSTRING("MGE: 'mW' failed because object id was longer than 32 characters");
                            } else {
                                strcpy_s(ActiveModel,33,CharPointer);
                                UsingModel=true;
                            }
                            break;
                        case 'c': //continue
                            DEBUGSTRING("MGE: All following model commands must specify their own object");
                            UsingModel=false;
                            break;
                        case 'l':  //Assign and load model
                        {
                            char FileName[256];
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            MWSE_GetString(FileName,&pointer);
                            DEBUGSTRING("MGE: Trying to assign id to file");
                            DEBUGSTRING(CharPointer);
                            DEBUGSTRING(FileName);
                            MWSE_ObjectIndex(CharPointer,FileName);
                            break;
                        }
                        case 'e': //enable
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to enable object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: Enable failed because object id does not exist");
                            } else {
                                if(Objects[index].Enabled) { DEBUGSTRING("MGE: Object already enabled"); }
                                Objects[index].Enabled=true;
                                DEBUGSTRING("MGE: Enable successful");
                            }
                            break;
                        case 'd': //disable
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to disable object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: Disable failed because object id does not exist");
                            } else {
                                if(!Objects[index].Enabled) { DEBUGSTRING("MGE: Object already disabled"); }
                                Objects[index].Enabled=false;
                                DEBUGSTRING("MGE: Disable successful");
                            }
                            break;
                        case 'p': //Set position
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to set position of object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: SetPos failed because object id does not exist");
                            } else {
                                Objects[index].pos.x=*(float*)(&PipeBuffer[pointer]);
                                Objects[index].pos.y=*(float*)(&PipeBuffer[pointer+4]);
                                Objects[index].pos.z=*(float*)(&PipeBuffer[pointer+8]);
                                MWSE_UpdateObjectMatrix(index,0);
                                DEBUGSTRING("MGE: SetPos successful");
                            }
                            pointer+=12;
                            break;
                        case 'r': //Set yaw pitch roll
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to set rotation of object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: SetRot failed because object id does not exist");
                            } else {
                                Objects[index].rot.x=*(float*)(&PipeBuffer[pointer]);
                                Objects[index].rot.y=*(float*)(&PipeBuffer[pointer+4]);
                                Objects[index].rot.z=*(float*)(&PipeBuffer[pointer+8]);
                                MWSE_UpdateObjectMatrix(index,1);
                                DEBUGSTRING("MGE: SetRot successful");
                            }
                            pointer+=12;
                            break;
                        case 's': //Scale
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to scale object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: SetScale failed because object id does not exist");
                            } else {
                                Objects[index].scale.x=*(float*)(&PipeBuffer[pointer]);
                                Objects[index].scale.y=*(float*)(&PipeBuffer[pointer+4]);
                                Objects[index].scale.z=*(float*)(&PipeBuffer[pointer+8]);
                                MWSE_UpdateObjectMatrix(index,2);
                                DEBUGSTRING("MGE: SetScale successful");
                            }
                            pointer+=12;
                            break;
                        case 'M': //Move
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to move object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: Move failed because object id does not exist");
                            } else {
                                Objects[index].pos.x+=*(float*)(&PipeBuffer[pointer]);
                                Objects[index].pos.y+=*(float*)(&PipeBuffer[pointer+4]);
                                Objects[index].pos.z+=*(float*)(&PipeBuffer[pointer+8]);
                                MWSE_UpdateObjectMatrix(index,0);
                                DEBUGSTRING("MGE: Move successful");
                            }
                            pointer+=12;
                            break;
                        case 'R': //rotate
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to rotate object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: Rotate failed because object id does not exist");
                            } else {
                                Objects[index].rot.x+=*(float*)(&PipeBuffer[pointer]);
                                Objects[index].rot.y+=*(float*)(&PipeBuffer[pointer+4]);
                                Objects[index].rot.z+=*(float*)(&PipeBuffer[pointer+8]);
                                MWSE_UpdateObjectMatrix(index,1);
                                DEBUGSTRING("MGE: Rotate successful");
                            }
                            pointer+=12;
                            break;
                        case 'S': //scale
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to scale object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: Scale failed because object id does not exist");
                            } else {
                                Objects[index].scale.x+=*(float*)(&PipeBuffer[pointer]);
                                Objects[index].scale.y+=*(float*)(&PipeBuffer[pointer+4]);
                                Objects[index].scale.z+=*(float*)(&PipeBuffer[pointer+8]);
                                MWSE_UpdateObjectMatrix(index,2);
                                DEBUGSTRING("MGE: Scale successful");
                            }
                            pointer+=12;
                            break;
                        case 'v':
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to set velocity of object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: SetVelocity failed because object id does not exist");
                            } else {
                                Objects[index].speed.x=*(float*)(&PipeBuffer[pointer]);
                                Objects[index].speed.y=*(float*)(&PipeBuffer[pointer+4]);
                                Objects[index].speed.z=*(float*)(&PipeBuffer[pointer+8]);
                                Objects[index].SpeedFlags|=0x01;
                                DEBUGSTRING("MGE: SetVelocity successful");
                            }
                            pointer+=12;
                            break;
                        case 'a':
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to set angular velocity of object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: SetAngularVelocity failed because object id does not exist");
                            } else {
                                Objects[index].AngSpeed.x=*(float*)(&PipeBuffer[pointer]);
                                Objects[index].AngSpeed.y=*(float*)(&PipeBuffer[pointer+4]);
                                Objects[index].AngSpeed.z=*(float*)(&PipeBuffer[pointer+8]);
                                Objects[index].SpeedFlags|=0x02;
                                DEBUGSTRING("MGE: SetAngularVelocity successful");
                            }
                            pointer+=12;
                            break;
                        case 'g':
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to set growth of object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: SetGrowth failed because object id does not exist");
                            } else {
                                Objects[index].growth.x=*(float*)(&PipeBuffer[pointer]);
                                Objects[index].growth.y=*(float*)(&PipeBuffer[pointer+4]);
                                Objects[index].growth.z=*(float*)(&PipeBuffer[pointer+8]);
                                Objects[index].SpeedFlags|=0x04;
                                DEBUGSTRING("MGE: SetGrowth successful");
                            }
                            pointer+=12;
                            break;
                        case 'V':
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to stop movement of object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: StopMovement failed because object id does not exist");
                            } else {
                                Objects[index].speed.x=0;
                                Objects[index].speed.y=0;
                                Objects[index].speed.z=0;
                                Objects[index].SpeedFlags&=0xFFFFFFFE;
                                DEBUGSTRING("MGE: StopMovement successful");
                            }
                            break;
                        case 'A':
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to stop rotation of object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: StopRotation failed because object id does not exist");
                            } else {
                                Objects[index].AngSpeed.x=0;
                                Objects[index].AngSpeed.y=0;
                                Objects[index].AngSpeed.z=0;
                                Objects[index].SpeedFlags&=0xFFFFFFFD;
                                DEBUGSTRING("MGE: StopRotation successful");
                            }
                            break;
                        case 'G':
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to stop growth of object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: StopGrowth failed because object id does not exist");
                            } else {
                                Objects[index].growth.x=0;
                                Objects[index].growth.y=0;
                                Objects[index].growth.z=0;
                                Objects[index].SpeedFlags&=0xFFFFFFFB;
                                DEBUGSTRING("MGE: StopGrowth successful");
                            }
                            break;
                        case 't': //SetTexture
                        {
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to texture object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: Set texture failed because object id does not exist");
                            }
                            MWSE_GetString(CharPointer,&pointer);
                            DEBUGSTRING("MGE: Texture name");
                            DEBUGSTRING(CharPointer);
                            if(index!=-1) {
                                int index2;
                                if(CharPointer[0]!='\0'&&(index2=MWSE_TextureIndex(CharPointer,true))==-1) {
                                    DEBUGSTRING("MGE: Set texture failed because texture could not be loaded");
                                } else {
                                    WORD d=*(WORD*)(&PipeBuffer[pointer]);
                                    if(d>=MAXMWSESUBSETS) {
                                        DEBUGSTRING("MGE: Set texture failed because texture stage was out of range");
                                    } else {
                                        if(Objects[index].textures[d]!=NULL) Objects[index].textures[d]->Release();
                                        if(CharPointer[0]=='\0') {
                                            Objects[index].textures[d]=NULL;
                                        } else {
                                            Objects[index].textures[d]=LoadedTextures[index2].texture;
                                            Objects[index].textures[d]->AddRef();
                                        }
                                        DEBUGSTRING("MGE: Set texture successful");
                                    }
                                }
                            }
                            pointer+=2;
                            break;
                        }
                        case 'E': //SetEffect
                        {
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to bind effect to object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: Set effect failed because object id does not exist");
                            }
                            MWSE_GetString(CharPointer,&pointer);
                            DEBUGSTRING("MGE: Effect name");
                            DEBUGSTRING(CharPointer);
                            if(index!=-1) {
                                int index2;
                                if(CharPointer[0]!='\0'&&(index2=MWSE_ShaderIndex(CharPointer,true))==-1) {
                                    DEBUGSTRING("MGE: Set effect failed because effect could not be loaded");
                                } else {
                                    if(Objects[index].effect!=NULL) Objects[index].effect->Release();
                                    if(CharPointer[0]=='\0') {
                                        Objects[index].effect=NULL;
                                    } else {
                                        Objects[index].effect=ShaderStruct.MWSEeffects[index2].effect;
                                        Objects[index].effect->AddRef();
                                    }
                                    DEBUGSTRING("MGE: Set effect successful");
                                }
                            }
                            break;
                        }
                        case 'D': //Destroy mesh
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to destroy mesh of object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("MGE: Destroy mesh failed because object id does not exist");
                            } else if(!Objects[index].IsMesh) {
                                DEBUGSTRING("MGE: Destroy mesh failed because object is not a mesh");
                            } else {
                                Objects[index].FVF=Objects[index].mesh->GetFVF();
                                Objects[index].faces=Objects[index].mesh->GetNumFaces();
                                Objects[index].stride=D3DXGetFVFVertexSize(Objects[index].FVF);
                                Objects[index].vertices=Objects[index].mesh->GetNumVertices();
                                Objects[index].mesh->GetVertexBuffer(&Objects[index].vBuffer);
                                Objects[index].mesh->GetIndexBuffer(&Objects[index].iBuffer);
                                Objects[index].IsMesh=false;
                                DEBUGSTRING("MGE: Destroy mesh successful");
                            }
                            break;
                        case 'f': //Free mesh
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to free object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("Free object failed because object id does not exist");
                            } else {
                                MWSE_FreeModel(index);
                            }
                            break;
                        case 'u': //Update lighting
                            if(!UsingModel) MWSE_GetString(CharPointer,&pointer);
                            else strcpy_s(CharPointer,256,ActiveModel);
                            DEBUGSTRING("MGE: Trying to update lighting on object");
                            DEBUGSTRING(CharPointer);
                            if((index=MWSE_ObjectIndex(CharPointer,NULL))==-1) {
                                DEBUGSTRING("Update lights failed because object id does not exist");
                            } else {
                                UpdateLighting(index);
                                DEBUGSTRING("MGE: Update lighting sucessful");
                            }
                            break;
                        default:
                            DEBUGSTRING("MGE: Object command not recognised");
                            break;
                        }
                        break;
#pragma endregion
                case 'f':   //Fogging
#pragma region fogging
                    DEBUGSTRING("MGE: Recieved fogging command");
                    switch(PipeBuffer[pointer++]) {
                        case 'd': if(Options.Fog) ToggleFog(); break;
                        case 'e': if(!Options.Fog) ToggleFog(); break;
                        case 't': ToggleFog(); break;
                        case 'i': IncreaseFog(); break;
                        case 'o': DecreaseFog(); break;
                        case 'I': 
                            //SatDown(&FogRange,MINFOG,*(float*)(&PipeBuffer[pointer]));
                            FogRange=CLAMP(FogRange-*(float*)(&PipeBuffer[pointer]),MINFOG,MAXFOG);
                            UpdateFogRange();
                            pointer+=4;
                            break;
                        case 'O':
                            //SatUp(&FogRange,MAXFOG,*(float*)(&PipeBuffer[pointer]));
                            FogRange=CLAMP(FogRange+*(float*)(&PipeBuffer[pointer]),MINFOG,MAXFOG);
                            UpdateFogRange();
                            pointer+=4;
                            break;
                        case 's':
                            FogRange=CLAMP(*(float*)(&PipeBuffer[pointer]),MINFOG,MAXFOG);
                            UpdateFogRange();
                            pointer+=4;
                            break;
                        case 'z':
                            DEBUGSTRING("MGE: Begining fog zoom");
                            FogZooming=true;
                            FogZoomRate=*(float*)(&PipeBuffer[pointer]);
                            pointer+=4;
                            break;
                        case 'Z':
                            DEBUGSTRING("MGE: Ending fog zoom");
                            FogZooming=false;
                            break;
                        case 'g':
                            DEBUGSTRING("MGE: A script has requested the current fog");
                            *((float*)&PipeOutBuffer[BytesToSend])=FogRange;
                            BytesToSend+=4;
                            break;
                        case 'c':
                        {
                            DEBUGSTRING("MGE: A script is manually setting the fog colour");
                            FogManual=true;
                            D3DXCOLOR NewColour;
                            NewColour.r=*(float*)(&PipeBuffer[pointer]);
                            NewColour.g=*(float*)(&PipeBuffer[pointer+4]);
                            NewColour.b=*(float*)(&PipeBuffer[pointer+8]);
                            NewColour.a=1;
                            FogColour=NewColour;
                            pointer+=12;
                            RealDevice->SetRenderState(D3DRS_FOGCOLOR,FogColour);
                            break;
                        }
                        case 'C': FogManual=false; break;
                        case 'G':
                        {
                            DEBUGSTRING("MGE: A script has requested the current fog colour");
                            D3DXCOLOR NewColour=FogColour;
                            *((float*)&PipeOutBuffer[BytesToSend])=NewColour.r;
                            *((float*)&PipeOutBuffer[BytesToSend+4])=NewColour.g;
                            *((float*)&PipeOutBuffer[BytesToSend+8])=NewColour.b;
                            BytesToSend+=12;
                            break;
                        }
                        default:
                            DEBUGSTRING("MGE: Unknown fogging command");
                            break;
                    }
                    break;
#pragma endregion
                case 's':   //Screenshot
#pragma region screenshots
                    DEBUGSTRING("MGE: Recieved screenshot command");
                    switch(PipeBuffer[pointer++]) {
                        BYTE b;
                        case 0:
						case 's':
							TakeScreenshot();
							break;
                        case '0':
                            b=Options.SaveFormat;
                            Options.SaveFormat=0;
                            TakeScreenshot();
                            Options.SaveFormat=b;
                            break;
                        case '1':
                            b=Options.SaveFormat;
                            Options.SaveFormat=1;
                            TakeScreenshot();
                            Options.SaveFormat=b;
                            break;
                        case '4':
                            b=Options.SaveFormat;
                            Options.SaveFormat=4;
                            TakeScreenshot();
                            Options.SaveFormat=b;
                            break;
                        case '5':
                            b=Options.SaveFormat;
                            Options.SaveFormat=5;
                            TakeScreenshot();
                            Options.SaveFormat=b;
                            break;
                        default:
                            DEBUGSTRING("MGE: Unknown screenshot command");
                            break;
                    }
                    break;
#pragma endregion
                case 'F':   //fps counter
#pragma region FpsCounter
                    DEBUGSTRING("MGE: Recived fps counter command");
                    switch(PipeBuffer[pointer++]) {
                        case 'd': if(Options.DisplayFPS) ToggleFpsCounter(); break;
                        case 'e': if(!Options.DisplayFPS) ToggleFpsCounter(); break;
                        case 't': ToggleFpsCounter(); break;
                        case 'g':
                            DEBUGSTRING("MGE: A script has requested the current fps");
                            *((float*)&PipeOutBuffer[BytesToSend])=fps;
                            BytesToSend+=4;
                            break;
                        default: DEBUGSTRING("MGE: Unknown fps counter command"); break;
                    }
                    break;
#pragma endregion
                case 'r':
#pragma region rotation
                    DEBUGSTRING("MGE: Recived rotation command");
                    switch(PipeBuffer[pointer++]) {
                        case 0:
						case '9':
                            RotateScreen();
                            break;
                        case '0': 
                            Rotation=0;
                            UpdateRotation();
                            break;
                        case '1': 
                            Rotation=PI/2;
                            UpdateRotation();
                            break;
                        case '2': 
                            Rotation=PI;
                            UpdateRotation();
                            break;
                        case '3': 
                            Rotation=(PI*2.0f/3.0f);
                            UpdateRotation();
                            break;
                        case 'r':
                            Rotation+=*(float*)(&PipeBuffer[pointer]);
                            pointer+=4;
                            UpdateRotation();
                            break;
                        case 'R':
                            Rotation=*(float*)(&PipeBuffer[pointer]);
                            pointer+=4;
                            UpdateRotation();
                            break;
                        case 's':
                            DEBUGSTRING("MGE: Spinning screen");
                            RotSpinning=true;
                            RotSpinRate=*(float*)(&PipeBuffer[pointer]);
                            pointer+=4;
                            break;
                        case 'S':
                            DEBUGSTRING("MGE: Finished spinning screen");
                            RotSpinning=false;
                            break;
                        case 'g':
                            DEBUGSTRING("MGE: A script has requested the current rotation");
                            *((float*)&PipeOutBuffer[BytesToSend])=Rotation;
                            BytesToSend+=4;
                            break;
                        default: DEBUGSTRING("MGE: Unknown rotation command"); break;
                    }
                    break;
#pragma endregion
                case 'z':   //zoom
#pragma region Zoom
                    DEBUGSTRING("MGE: Recived zoom command");
                    switch(PipeBuffer[pointer++]) {
                        case 'd': Zoom=false;break;
                        case 'e': Zoom=true;break;
                        case 't': Zoom=!Zoom;break;
                        case 'i': IncreaseZoom();break;
                        case 'o': DecreaseZoom();break;
                        case 'I': 
                            FOV=CLAMP(FOV+*(float*)(&PipeBuffer[pointer]),MINZOOM,MAXZOOM);
                            pointer+=4;
                            break;
                        case 'O':
                            FOV=CLAMP(FOV-*(float*)(&PipeBuffer[pointer]),MINZOOM,MAXZOOM);
                            pointer+=4;
                            break;
                        case 's':
                            FOV=CLAMP(*(float*)(&PipeBuffer[pointer]),MINZOOM,MAXZOOM);
                            pointer+=4;
                            break;
                        case 'z':
                            DEBUGSTRING("MGE: Begining FOV zoom");
                            FOVZooming=true;
                            FOVZoomRate=*(float*)(&PipeBuffer[pointer]);
                            pointer+=4;
                            break;
                        case 'Z':
                            DEBUGSTRING("MGE: Ending FOV zoom");
                            FOVZooming=false;
                            break;
                        case 'g':
                            DEBUGSTRING("MGE: A script has requested the current zoom");
                            *((float*)&PipeOutBuffer[BytesToSend])=FOV;
                            BytesToSend+=4;
                            break;
                        default:
                            DEBUGSTRING("MGE: Unknown zoom command");
                            break;
                    }
                    break;
#pragma endregion
                case 'S':   //Hardware shaders
#pragma region Shaders
                {
                    char FilePointer[256]; FilePointer[0]='\0';
                    int index;
                    DEBUGSTRING("MGE: Recieved hardware shader command");
                    switch(PipeBuffer[pointer++]) {
                        case 'w':
                            DEBUGSTRING("MGE: Wiping all loaded effects");
                            for(UINT i=0;i<MAXMWSESHADERS;i++)
                                if(ShaderStruct.MWSEeffects[i].effect!=NULL) MWSE_FreeEffect(i);
                            break;
                        case 'd': if(Options.HWShader) ToggleHardwareShader(); break;
                        case 'e': if(!Options.HWShader) ToggleHardwareShader(); break;
                        case 't': ToggleHardwareShader(); break;
                        case 'E':
                            MWSE_GetString(FilePointer,&pointer);
                            if((index=MWSE_ShaderIndex(FilePointer,true))==-1) {
                                DEBUGSTRING("MGE: Could not enable shader because it could not be loaded");
                            } else {
                                if(ShaderStruct.MWSEeffects[index].Active) {
                                    DEBUGSTRING("MGE: Shader was already enabled");
                                } else {
                                    ShaderStruct.MWSEeffects[index].Active=true;
                                    UsedMWSEShaders++;
                                }
                            }
                            break;
                        case 'D':
                            MWSE_GetString(FilePointer,&pointer);
                            if((index=MWSE_ShaderIndex(FilePointer,false))==-1) {
                                DEBUGSTRING("MGE: Could not disable shader because it was not loaded");
                            } else {
                                if(!ShaderStruct.MWSEeffects[index].Active) {
                                    DEBUGSTRING("MGE: Shader was already disabled");
                                } else {
                                    ShaderStruct.MWSEeffects[index].Active=false;
                                    UsedMWSEShaders--;
                                }
                            }
                            break;
                        case 'l': 
                            MWSE_GetString(FilePointer,&pointer);
                            MWSE_ShaderIndex(FilePointer,true);
                            break;
                        case 's':
                        {
                            MWSE_GetString(FilePointer,&pointer);
                            DEBUGSTRING("MGE: Trying to set technique");
                            DEBUGSTRING(FilePointer);
                            if((index=MWSE_ShaderIndex(FilePointer,true))==-1) {
                                DEBUGSTRING("MGE: Couldn't set technique because effect could not be loaded");
                            }
                            MWSE_GetString(FilePointer,&pointer);   //Need to pop this string even if index==-1
                            DEBUGSTRING("MGE: Trying to set technique");
                            DEBUGSTRING(FilePointer);
                            if(index!=-1) {
                                ShaderStruct.MWSEeffects[index].effect->SetTechnique(FilePointer);
                                DEBUGSTRING("MGE: Technique set");
                            }
                            break;
                        }
                        case 'f':   //Free effect
                        {
                            MWSE_GetString(FilePointer,&pointer);
                            DEBUGSTRING("MGE: Trying to free effect");
                            DEBUGSTRING(FilePointer);
                            if((index=MWSE_ShaderIndex(FilePointer,false))!=-1) {
                                MWSE_FreeEffect(index);
                            } else {
                                DEBUGSTRING("MGE: Could not free effect because it was not loaded");
                            }
                            break;
                        }
                        case 'L':   //Set DWORD
                        {
                            MWSE_GetString(FilePointer,&pointer);
                            DEBUGSTRING("MGE: Setting effect DWORD");
                            DEBUGSTRING(FilePointer);
                            if(FilePointer[0]=='\0') index=-2;
                            else if((index=MWSE_ShaderIndex(FilePointer,true))==-1) {
                                DEBUGSTRING("MGE: Could not set DWORD because effect could not be loaded");
                            } 
                            MWSE_GetString(FilePointer,&pointer);
                            DEBUGSTRING("MGE: Variable name");
                            DEBUGSTRING(FilePointer);
                            if(index>=0) {
                                DWORD d=*(DWORD*)&PipeBuffer[pointer];
                                ShaderStruct.MWSEeffects[index].effect->SetInt(FilePointer,d);
                            } else if(index==-2&&ShaderStruct.DefEffectCount) {
                                DWORD d=*(DWORD*)&PipeBuffer[pointer];
                                ShaderStruct.DefaultEffects[0]->SetInt(FilePointer,d);
                            }
                            pointer+=4;
                            break;
                        }
                        case 'F':   //Set float
                        {
                            MWSE_GetString(FilePointer,&pointer);
                            DEBUGSTRING("MGE: Setting effect float");
                            DEBUGSTRING(FilePointer);
                            if(FilePointer[0]=='\0') index=-2;
                            else if((index=MWSE_ShaderIndex(FilePointer,true))==-1) {
                                DEBUGSTRING("MGE: Could not set float because effect could not be loaded");
                            }
                            MWSE_GetString(FilePointer,&pointer);
                            DEBUGSTRING("MGE: Variable name");
                            DEBUGSTRING(FilePointer);
                            if(index>=0) {
                                float d=*(float*)&PipeBuffer[pointer];
                                ShaderStruct.MWSEeffects[index].effect->SetFloat(FilePointer,d);
                            } else if(index==-2&&ShaderStruct.DefEffectCount) {
                                float d=*(float*)&PipeBuffer[pointer];
                                ShaderStruct.DefaultEffects[0]->SetFloat(FilePointer,d);
                            }
                            pointer+=4;
                            break;
                        }
                        case 'V':   //Set vector
                        {
                            MWSE_GetString(FilePointer,&pointer);
                            D3DXVECTOR4 vec;
                            DEBUGSTRING("MGE: Setting effect vector");
                            DEBUGSTRING(FilePointer);
                            if(FilePointer[0]=='\0') index=-2;
                            else if((index=MWSE_ShaderIndex(FilePointer,true))==-1) {
                                DEBUGSTRING("MGE: Could not set vector because effect could not be loaded");
                            }
                            MWSE_GetString(FilePointer,&pointer);
                            DEBUGSTRING("MGE: Variable name");
                            DEBUGSTRING(FilePointer);
                            if(index>=0) {
                                vec.x=*(float*)&PipeBuffer[pointer];
                                vec.y=*(float*)&PipeBuffer[pointer+4];
                                vec.z=*(float*)&PipeBuffer[pointer+8];
                                vec.w=*(float*)&PipeBuffer[pointer+12];
                                ShaderStruct.MWSEeffects[index].effect->SetVector(FilePointer,&vec);
                            } else if(index==-2&&ShaderStruct.DefEffectCount) {
                                vec.x=*(float*)&PipeBuffer[pointer];
                                vec.y=*(float*)&PipeBuffer[pointer+4];
                                vec.z=*(float*)&PipeBuffer[pointer+8];
                                vec.w=*(float*)&PipeBuffer[pointer+12];
                                ShaderStruct.DefaultEffects[0]->SetVector(FilePointer,&vec);
                            }
                            pointer+=16;
                            break;
                        }
                        case 'T':   //Set texture
                        {
                            MWSE_GetString(FilePointer,&pointer);
                            char texture[256];
                            DEBUGSTRING("MGE: Setting effect texture");
                            DEBUGSTRING(FilePointer);
                            if(FilePointer[0]=='\0') index=-2;
                            else if((index=MWSE_ShaderIndex(FilePointer,true))==-1) {
                                DEBUGSTRING("MGE: Could not set texture because effect could not be loaded");
                            }
                            MWSE_GetString(FilePointer,&pointer);
                            MWSE_GetString(texture,&pointer);
                            DEBUGSTRING("MGE: Variable name");
                            DEBUGSTRING(FilePointer);
                            DEBUGSTRING("MGE: Texture path");
                            DEBUGSTRING(texture);
                            if(index>=0) {
                                DWORD index2;
                                if((index2=MWSE_TextureIndex(texture,true))==-1) {
                                    DEBUGSTRING("MGE: Set texture failed because texture could not be loaded");
                                } else {
                                    ShaderStruct.MWSEeffects[index].effect->SetTexture(FilePointer,LoadedTextures[index2].texture);
                                }
                            } else if(index==-2&&ShaderStruct.DefEffectCount) {
                                DWORD index2;
                                if((index2=MWSE_TextureIndex(texture,true))==-1) {
                                    DEBUGSTRING("MGE: Set texture failed because texture could not be loaded");
                                } else {
                                    ShaderStruct.DefaultEffects[0]->SetTexture(FilePointer,LoadedTextures[index2].texture);
                                }
                            }
                            break;
                        }
                        default:
                            DEBUGSTRING("MGE: Unknown hardware shader command");
                            break;
                    }
                    break;
                }
#pragma endregion
                case 't':   //Textures
#pragma region textures
                {
                    char FilePointer[256]; FilePointer[0]='\0';
                    int index;
                    DEBUGSTRING("MGE: Recieved texture command");
                    switch(PipeBuffer[pointer++]) {
                        case 'w':
                            DEBUGSTRING("MGE: Wiping all loaded textures");
                            for(UINT i=0;i<MAXMWSESHADERS;i++)
                                if(LoadedTextures[i].texture!=NULL) MWSE_FreeTexture(i);
                            break;
                        case 'l':
                            DEBUGSTRING("MGE: Trying to load texture");
                            MWSE_GetString(FilePointer,&pointer);
                            MWSE_TextureIndex(FilePointer,true);
                            break;
                        case 'f':
                            MWSE_GetString(FilePointer,&pointer);
                            DEBUGSTRING("MGE: Trying to free texture");
                            DEBUGSTRING(FilePointer);
                            if((index=MWSE_TextureIndex(FilePointer,false))!=-1) {
                                MWSE_FreeTexture(index);
                            } else {
                                DEBUGSTRING("MGE: Could not free texture because it was not loaded");
                            }
                            break;
                        case 'b': //BeginWrite
                        {
                            D3DSURFACE_DESC desc;
                            MWSE_GetString(FilePointer,&pointer);
                            DEBUGSTRING("MGE: Trying to begin writing to texture");
                            DEBUGSTRING(FilePointer);
                            if(TexRenderTarget!=NULL) {
                                DEBUGSTRING("MGE: Write to texture failed because another texture is already being written to.");
                                break;
                            } 
                            if((index=MWSE_TextureIndex(FilePointer,true))==-1) {
                                DEBUGSTRING("MGE: Could not write to texture because it could not be loaded");
                                break;
                            }
                            if(LoadedTextures[index].texture->GetLevelDesc(0,&desc)!=D3D_OK) {
                                DEBUGSTRING("MGE: Could not get description of texture");
                                break;
                            }
                            if(desc.Format!=D3DFMT_X8R8G8B8) {
                                DEBUGSTRING("MGE: Texture is not in the correct format");
                                break;
                            }
                            if(RealDevice->CreateTexture(desc.Width,desc.Height,1,D3DUSAGE_RENDERTARGET,
                                D3DFMT_X8R8G8B8,D3DPOOL_DEFAULT,&TexRenderTarget,NULL)!=D3D_OK) {
                                    DEBUGSTRING("MGE: Could not create render target");
                                    break;
                            }
                            TexRenderTarget->GetSurfaceLevel(0,&SufRenderTarget);
                            RealDevice->SetRenderTarget(0,SufRenderTarget);
                            TexCopyTo=LoadedTextures[index].texture;
                            TexCopyTo->AddRef();
                            TexCopyTo->GetSurfaceLevel(0,&SufCopyTo);
                            RealDevice->BeginScene();
                            ZeroStruct(TexDrawOrigin);
                            TexDrawScale.x=1;
                            TexDrawScale.y=1;
                            TexDrawColour=0xFFFFFFFF;
                            RealDevice->StretchRect(SufCopyTo,0,SufRenderTarget,NULL,D3DTEXF_NONE);
                            DEBUGSTRING("MGE: Successfully began writing to texture");       
                            break;
                        }
                        case 'e': //end write
                            DEBUGSTRING("MGE: Stopping writing to texture");
                            if(TexRenderTarget==NULL) {
                                DEBUGSTRING("MGE: EndWrite failed because no texture was being written to");
                                break;
                            }
                            RealDevice->EndScene();
                            RealDevice->SetRenderTarget(0,SufRenderTarget);
                            RealDevice->StretchRect(SufRenderTarget,NULL,SufCopyTo,NULL,D3DTEXF_NONE);
                            SufRenderTarget->Release();
                            TexRenderTarget->Release();
                            SufCopyTo->Release();
                            TexCopyTo->Release();
                            TexRenderTarget=TexCopyTo=NULL;
                            SufRenderTarget=SufCopyTo=NULL;
                            break;
                        case 'o': //Origin
                            DEBUGSTRING("MGE: Setting render origin");
                            TexDrawOrigin.x=*(float*)&PipeBuffer[pointer];
                            TexDrawOrigin.y=*(float*)&PipeBuffer[pointer+4];
                            pointer+=8;
                            break;
                        case 's': //Scale
                            DEBUGSTRING("MGE: Setting render scale");
                            TexDrawScale.x=*(float*)&PipeBuffer[pointer];
                            TexDrawScale.y=*(float*)&PipeBuffer[pointer+4];
                            pointer+=8;
                            break;
                        case 'c': //Colour
                            DEBUGSTRING("MGE: Setting texture render colour");
                            TexDrawColour=*(DWORD*)&PipeBuffer[pointer];
                            pointer+=4;
                            break;
                        case 'T': //Text
                        {
                            MWSE_GetString(FilePointer,&pointer);
                            DEBUGSTRING("MGE: Drawing text to texture");
                            DEBUGSTRING(FilePointer);
                            RECT r;
                            r.left=(DWORD)TexDrawOrigin.x;
                            r.top=(DWORD)TexDrawOrigin.y;
                            r.right=r.left+(DWORD)TexDrawScale.x;
                            r.bottom=r.top+(DWORD)TexDrawScale.y;
                            Font->DrawText(NULL,FilePointer,-1,&r,0,TexDrawColour);
                            break;
                        }
                        case 't': //Draw texture
                            MWSE_GetString(FilePointer,&pointer);
                            DEBUGSTRING("MGE: Drawing texture to texture");
                            DEBUGSTRING(FilePointer);
                            if((index=MWSE_TextureIndex(FilePointer,true))==-1) {
                                DEBUGSTRING("MGE: Could load texture to be written");
                                break;
                            }
                            D3DXMATRIX m;
                            D3DXVECTOR3 v;
                            v.x=TexDrawOrigin.x;
                            v.y=TexDrawOrigin.y;
                            v.z=0;
                            HUDsprite->SetTransform(D3DXMatrixScaling(&m,TexDrawScale.x,TexDrawScale.y,1));
                            HUDsprite->Draw(LoadedTextures[index].texture,0,0,&v,TexDrawColour);
                            break;
                    }
                    break;
                }
#pragma endregion
                default:
                    DEBUGSTRING("MGE: Unknown command recieved from MWSE");
                    break;
            }
        }
		if(BytesToSend) {
			WriteFile(MWSEpipe,PipeOutBuffer,BytesToSend,&size,NULL);
		}
    }
}
