#define _WrapperClasses
#include "WrapperClasses.h"

static bool RenderedLand=false;
static bool LoadingScreen=false;
bool DistantLandActive=false;

static IDirect3D8* fakeDirect3D;
static IDirect3DDevice8* fakeDevice;
static D3DCAPS8 d3d8Caps;

class FakeSurface : public IDirect3DSurface8
{
public:
	IDirect3DSurface9* RealObject;
	/*** Constructor ***/
	FakeSurface(IDirect3DSurface9* Real) {
		RealObject=Real;
		void* This = this;
		Real->SetPrivateData(guid,(void*)&This,4,0);
	}
	/*** IUnknown methods ***/
	HRESULT _stdcall QueryInterface(REFIID riid, void** ppvObj) { return RealObject->QueryInterface(riid, ppvObj); }
	ULONG _stdcall AddRef() { return RealObject->AddRef(); }
	ULONG _stdcall Release() { 
		ULONG Refs=RealObject->Release();
		if(!Refs) {
			delete this;
			return 0;
		}
		return Refs;
	}

	/*** IDirect3DSurface8 methods ***/
	HRESULT _stdcall GetDevice(IDirect3DDevice8** ppDevice) {
		*ppDevice=fakeDevice;
		fakeDevice->AddRef();
		return D3D_OK; }
	HRESULT _stdcall SetPrivateData(REFGUID refguid, CONST void* pData, DWORD SizeOfData, DWORD Flags) { return RealObject->SetPrivateData(refguid, pData, SizeOfData, Flags); }
	HRESULT _stdcall GetPrivateData(REFGUID refguid, void* pData, DWORD* pSizeOfData) { return RealObject->GetPrivateData(refguid, pData, pSizeOfData); }
	HRESULT _stdcall FreePrivateData(REFGUID refguid) { return RealObject->FreePrivateData(refguid); }
	HRESULT _stdcall GetContainer(REFIID riid, void** ppContainer) { return RealObject->GetContainer(riid, ppContainer); }
	HRESULT _stdcall GetDesc(D3DSURFACE_DESC8 *pDesc) { 
		D3DSURFACE_DESC a2;
		HRESULT hr=RealObject->GetDesc(&a2);
		if(hr==D3D_OK) {
			pDesc->Format=a2.Format;
			pDesc->Height=a2.Height;
			pDesc->MultiSampleType=a2.MultiSampleType;
			pDesc->Pool=a2.Pool;
			pDesc->Size=0; //TODO: Fix;
			pDesc->Type=a2.Type;
			pDesc->Usage=a2.Usage;
			pDesc->Width=a2.Width;
		}
		return hr;
	}
	HRESULT _stdcall LockRect(D3DLOCKED_RECT* pLockedRect, CONST RECT* pRect, DWORD Flags) { return RealObject->LockRect(pLockedRect, pRect, Flags); }
	HRESULT _stdcall UnlockRect() { return RealObject->UnlockRect(); }
};

class FakeTexture : public IDirect3DTexture8
{
public:
	IDirect3DTexture9* RealObject;
	/*** Constructor ***/
	FakeTexture(IDirect3DTexture9* Real) {
		RealObject=Real;
		void* This = this;
		Real->SetPrivateData(guid,(void*)&This,4,0);
	}
	/*** IUnknown methods ***/
	HRESULT _stdcall QueryInterface(REFIID riid, void** ppvObj) { return RealObject->QueryInterface(riid, ppvObj); }
	ULONG _stdcall AddRef() { return RealObject->AddRef(); }
	ULONG _stdcall Release() {
		ULONG Refs=RealObject->Release();
		if(!Refs) {
			delete this;
			return 0;
		}
		return Refs;
	}

	/*** IDirect3DBaseTexture8 methods ***/
	HRESULT _stdcall GetDevice(IDirect3DDevice8** ppDevice) {
		*ppDevice=fakeDevice;
		fakeDevice->AddRef();
		return D3D_OK;
	}
	HRESULT _stdcall SetPrivateData(REFGUID refguid, CONST void* pData, DWORD SizeOfData, DWORD Flags) { return RealObject->SetPrivateData(refguid, pData, SizeOfData, Flags); }
	HRESULT _stdcall GetPrivateData(REFGUID refguid, void* pData, DWORD* pSizeOfData) { return RealObject->GetPrivateData(refguid, pData, pSizeOfData); }
	HRESULT _stdcall FreePrivateData(REFGUID refguid) { return RealObject->FreePrivateData(refguid); }
	DWORD _stdcall SetPriority(DWORD PriorityNew) { return RealObject->SetPriority(PriorityNew); }
	DWORD _stdcall GetPriority() { return RealObject->GetPriority(); }
	void _stdcall PreLoad() { return RealObject->PreLoad(); }
	D3DRESOURCETYPE _stdcall GetType() { return RealObject->GetType(); }
	DWORD _stdcall SetLOD(DWORD LODNew) { return RealObject->SetLOD(LODNew); }
	DWORD _stdcall GetLOD() { return RealObject->GetLOD(); }
	DWORD _stdcall GetLevelCount() { return RealObject->GetLevelCount(); }
	HRESULT _stdcall GetLevelDesc(UINT Level, D3DSURFACE_DESC8 *pDesc) {
		D3DSURFACE_DESC b2;
		HRESULT hr=RealObject->GetLevelDesc(Level,&b2);
		if(hr==D3D_OK) {
			pDesc->Format=b2.Format;
			pDesc->Height=b2.Height;
			pDesc->MultiSampleType=b2.MultiSampleType;
			pDesc->Pool=b2.Pool;
			pDesc->Size=0; //TODO: Fix;
			pDesc->Type=b2.Type;
			pDesc->Usage=b2.Usage;
			pDesc->Width=b2.Width;
		}
		return hr;
	}
	HRESULT _stdcall GetSurfaceLevel(UINT Level, IDirect3DSurface8** ppSurfaceLevel) {
		IDirect3DSurface9* b2;
		*ppSurfaceLevel=NULL;
		DWORD unused=4;
		HRESULT hr=RealObject->GetSurfaceLevel(Level,&b2);
		if(hr!=D3D_OK||b2==NULL) return hr;
		hr=b2->GetPrivateData(guid,(void*)ppSurfaceLevel,&unused);
		if(hr!=D3D_OK) *ppSurfaceLevel=new FakeSurface(b2);
		return D3D_OK;
	}
	HRESULT _stdcall LockRect(UINT Level, D3DLOCKED_RECT* pLockedRect, CONST RECT* pRect, DWORD Flags) { return RealObject->LockRect(Level, pLockedRect, pRect, Flags); }
	HRESULT _stdcall UnlockRect(UINT Level) { return RealObject->UnlockRect(Level); }
	HRESULT _stdcall AddDirtyRect(CONST RECT* pDirtyRect) { return RealObject->AddDirtyRect(pDirtyRect); }
};

class TextureHook : public IDirect3DTexture8 {
public:
	DWORD refs;
	BYTE HookIndex;
	DWORD w;
	DWORD h;
	BYTE* locked;
	/*** Misc methods ***/
	TextureHook(BYTE index,DWORD W,DWORD H) { HookIndex=index; w=W; h=H; refs=2; locked=NULL; }
	/*** IUnknown methods ***/
	HRESULT _stdcall QueryInterface (REFIID a, LPVOID * b) { return E_NOINTERFACE; }
	ULONG _stdcall AddRef(void) { return ++refs; }
	ULONG _stdcall Release(void) {
		if(--refs==0) {
			if(HookedTextures[HookIndex].AutoClean) {
				MWSE_FreeTexture(HookedTextures[HookIndex].TextureIndex);
				MWSE_FreeEffect(HookedTextures[HookIndex].ShaderIndex);
				ZeroStruct(HookedTextures[HookIndex]);
			}
			HookedTextures[HookIndex].tHook=0;
			delete this;
			return 0;
		}
		return refs;
	}
	/*** IDirect3DResource8 methods ***/
	HRESULT _stdcall GetDevice(IDirect3DDevice8** a) { return D3DERR_INVALIDCALL; }
	HRESULT _stdcall SetPrivateData(REFGUID a,const void* b,DWORD c,DWORD d) { return D3DERR_INVALIDCALL; }
	HRESULT _stdcall GetPrivateData(REFGUID a,void* b,DWORD* c) { return D3DERR_INVALIDCALL; }
	HRESULT _stdcall FreePrivateData(REFGUID a) { return D3DERR_INVALIDCALL; }
	DWORD _stdcall SetPriority(DWORD a) { return a; }
	DWORD _stdcall GetPriority(void) { return HookIndex; }
	void _stdcall PreLoad() { }
	D3DRESOURCETYPE _stdcall GetType() { return D3DRTYPE_TEXTURE; }
	/*** IDirect3DBaseTexture8 methods ***/
	DWORD _stdcall SetLOD(DWORD a) { return a; }
	DWORD _stdcall GetLOD() { return HOOKEDTEXTUREID; }
	DWORD _stdcall GetLevelCount() { return 1; }
	/*** IDirect3DTexture8 methods ***/
	HRESULT _stdcall GetLevelDesc(UINT a,D3DSURFACE_DESC8 *b) { return D3DERR_INVALIDCALL; }
	HRESULT _stdcall GetSurfaceLevel(UINT a,IDirect3DSurface8** b) { return D3DERR_INVALIDCALL; }
	HRESULT _stdcall LockRect(UINT a,D3DLOCKED_RECT* b,CONST RECT* c,DWORD d) { 
		if(locked) return D3DERR_INVALIDCALL;
		b->Pitch=w*4;
		b->pBits=new BYTE[w*h*4];
		return D3D_OK;
	}
	HRESULT _stdcall UnlockRect(UINT a) { 
		if(locked) {
			delete[w*h*4] locked;
			locked=NULL;
		}
		return D3D_OK;
	}
	HRESULT _stdcall AddDirtyRect(const RECT* a) { return D3D_OK; }
};

class FakeDevice : public IDirect3DDevice8 {
public:
	/*** Fields ***/
	IDirect3D8* Object;
	DWORD Refs;
	UINT BaseVertexIndex;
	DWORD CurrentPShader;
	DWORD CurrentVShader;
	IDirect3DVertexShader9* NewVShaders[4];
	IDirect3DPixelShader9* NewPShaders[9];
	DWORD pShaderCount;
	DWORD vShaderCount;
	IDirect3DVertexDeclaration9* decl[2];
	/*** Misc methods ***/
	FakeDevice(IDirect3D8* Ob) {
		Object=Ob;
		Refs=1;
		//TempBool=false;
		BaseVertexIndex=0;
		CurrentPShader=0;
		CurrentVShader=0;
		pShaderCount=0;
		vShaderCount=1;
		ZeroStruct(NewVShaders);
		ZeroStruct(NewPShaders);
		D3DVERTEXELEMENT9 elements[6] = {
			{0,0,D3DDECLTYPE_FLOAT3,D3DDECLMETHOD_DEFAULT,D3DDECLUSAGE_POSITION},
			{0,12,D3DDECLTYPE_FLOAT2,D3DDECLMETHOD_DEFAULT,D3DDECLUSAGE_TEXCOORD},
			D3DDECL_END(),
			{0,0,D3DDECLTYPE_FLOAT3,D3DDECLMETHOD_DEFAULT,D3DDECLUSAGE_POSITION},
			{0,12,D3DDECLTYPE_D3DCOLOR,D3DDECLMETHOD_DEFAULT,D3DDECLUSAGE_COLOR},
			D3DDECL_END()
		};
		RealDevice->CreateVertexDeclaration(elements,&decl[0]);
		RealDevice->CreateVertexDeclaration(&elements[3],&decl[1]);
	}
	/*** IUnknown methods ***/
	HRESULT _stdcall QueryInterface (REFIID a, LPVOID * b) {return RealDevice->QueryInterface(a,b);}
	ULONG _stdcall AddRef(void) {
		return ++Refs;
	}
	ULONG _stdcall Release(void) {
		if(--Refs==0) {
			DEBUGSTRING("MGE: Graphics device lost. Wiping all loaded MWSE objects.");
			for(UINT i=0;i<MAXMWSEOBJECTS;i++)
				if(Objects[i].Created) MWSE_FreeModel(i);
			for(UINT i=0;i<MAXMWSESHADERS;i++)
				if(ShaderStruct.MWSEeffects[i].effect!=NULL) MWSE_FreeEffect(i);
			for(UINT i=0;i<MAXMWSESHADERS;i++)
				if(LoadedTextures[i].texture!=NULL) MWSE_FreeTexture(i);
			for(UINT i=0;i<MAXMWSEOBJECTS;i++)
				if(HUDbits[i].texture) MWSE_FreeHUD(i);
			for(UINT i=0;i<MAXHOOKEDTEXTURES;i++) {
				if(HookedTextures[i].tHook) {
					delete HookedTextures[i].tHook;
					HookedTextures[i].tHook=0;
				}
			}
			ZeroStruct(HookedTextures);
			if(SunTexture!=NULL) SunTexture->Release();
			if(HUDsprite!=NULL) HUDsprite->Release();
			DEBUGSTRING("MGE: Graphics device lost. Releasing all created resources.");
			if(ShaderStruct.thisframeSurface!=NULL) ShaderStruct.thisframeSurface->Release();
			if(ShaderStruct.lastframeSurface!=NULL) ShaderStruct.lastframeSurface->Release();
			if(ShaderStruct.lastpassSurface!=NULL) ShaderStruct.lastpassSurface->Release();
			if(ShaderStruct.lastshaderSurface!=NULL) ShaderStruct.lastshaderSurface->Release();
			if(ShaderStruct.thisframe!=NULL) ShaderStruct.thisframe->Release();
			if(ShaderStruct.lastframe!=NULL) ShaderStruct.lastframe->Release();
			if(ShaderStruct.lastpass!=NULL) ShaderStruct.lastpass->Release();
			if(ShaderStruct.lastshader!=NULL) ShaderStruct.lastshader->Release();
			if(ShaderStruct.SoftwareSurface!=NULL) ShaderStruct.SoftwareSurface->Release();
			if(ShaderStruct.buffer!=NULL) ShaderStruct.buffer->Release();
			for(int i=0;i<ShaderStruct.DefEffectCount;i++) {
				ShaderStruct.DefaultEffects[i]->Release();
			}
			if(BackBuffer!=NULL) { while(BackBuffer->Release()>0) { } }
			if(DepthBuffer!=NULL) { while(DepthBuffer->Release()>0) {} }
			if(Font!=NULL) Font->Release();
			delete this;
			return 0;
		}
		return Refs;
	}
	/*** IDirect3DDevice8 methods ***/
	HRESULT _stdcall TestCooperativeLevel(void) {
		return RealDevice->TestCooperativeLevel();
	}
	UINT _stdcall GetAvailableTextureMem(void) {
		return RealDevice->GetAvailableTextureMem();
	}
	HRESULT _stdcall ResourceManagerDiscardBytes(DWORD a) {
		return RealDevice->EvictManagedResources();
	}
	HRESULT _stdcall GetDirect3D(IDirect3D8** a) {
		*a=Object;
		Object->AddRef();
		return D3D_OK;
	}
	HRESULT _stdcall GetDeviceCaps(D3DCAPS8* a) {
		*a=d3d8Caps;
		return D3D_OK;
	}
	HRESULT _stdcall GetDisplayMode(D3DDISPLAYMODE* a) {
		return RealDevice->GetDisplayMode(0,a);
	}
	HRESULT _stdcall GetCreationParameters(D3DDEVICE_CREATION_PARAMETERS *a) {
		return RealDevice->GetCreationParameters(a);
	}
	HRESULT _stdcall SetCursorProperties(UINT a,UINT b,IDirect3DSurface8* c) {
		return RealDevice->SetCursorProperties(a,b,((FakeSurface*)c)->RealObject);
	}
	void _stdcall SetCursorPosition(int X,int Y,DWORD c) {
		RealDevice->SetCursorPosition(X,Y,c);
	}
	BOOL _stdcall ShowCursor(BOOL a) {
		return RealDevice->ShowCursor(a);
	}
	HRESULT _stdcall CreateAdditionalSwapChain(D3DPRESENT_PARAMETERS8* a,IDirect3DSwapChain8** b) {
		return UnusedFunction();
	}
	HRESULT _stdcall Reset(D3DPRESENT_PARAMETERS8* a) {
		return UnusedFunction();
	}
	HRESULT _stdcall Present(const RECT* a,const RECT* b,HWND c,const RGNDATA* d) {
		HRESULT hr;
		DWORD time=GetTickCount();
		if(Options.UseHDR) HDR_Update(time);
		RenderedLand=false;
		//Update SinVar and LinVar variables
		if(Options.SetShaderVars||UsingDistantLand) {
			float TimePassed=((float)(time%1000))/1000.0f;
			LinVar.x=TimePassed;
			LinVar.y=TimePassed;
			LinVar.z=TimePassed;
			LinVar.w=1-TimePassed;
			//need to mod to stop it loosing resolution after a few hours. Makes it jerk a bit every minute or so.
			TimePassed=sin((float)((time%60001))*(0.00025f*PI));
			SinVar.x=TimePassed;
			SinVar.y=TimePassed;
			SinVar.z=TimePassed;
			SinVar.w=cos((float)((time%60001))*(0.00025f*PI));
			for(int i=0;i<MAXMWSESHADERS;i++) {
				if(ShaderStruct.MWSEeffects[i].effect) {
					SetShaderVars(ShaderStruct.MWSEeffects[i].effect,time);
				}
			}
		}
		//HUD
		if(UsedHUDbits) {
			RealDevice->BeginScene();
			HUDsprite->Begin(D3DXSPRITE_ALPHABLEND);
			for(DWORD d=0;d<MAXMWSEOBJECTS;d++) {
				if(HUDbits[d].texture&&HUDbits[d].Enabled) {
					if(HUDbits[d].effect) {
						UINT passes;
						HUDbits[d].effect->Begin(&passes,0);
						for(UINT j=0;j<passes;j++) {
							HUDbits[d].effect->BeginPass(j);
							D3DXMATRIX m;
							D3DXVECTOR3 v;
							v.x=HUDbits[d].position.x;
							v.y=HUDbits[d].position.y;
							v.z=0;
							HUDsprite->SetTransform(::D3DXMatrixScaling(&m,HUDbits[d].scale.x,HUDbits[d].scale.y,1));
							HUDsprite->Draw(HUDbits[d].texture,0,0,&v,0xFFFFFFFF);
							HUDbits[d].effect->EndPass();


						}
						HUDbits[d].effect->End();
					} else {
						D3DXMATRIX m;
						D3DXVECTOR3 v;
						v.x=HUDbits[d].position.x;
						v.y=HUDbits[d].position.y;
						v.z=0;
						HUDsprite->SetTransform(::D3DXMatrixScaling(&m,HUDbits[d].scale.x,HUDbits[d].scale.y,1));
						HRESULT hrr=HUDsprite->Draw(HUDbits[d].texture,0,0,&v,0xFFFFFFFF);
					}
				}
			}
			HUDsprite->End();
			RealDevice->EndScene();
		}
		//Calculate fps
		fpsFrames++;
		if(time-fpsLastTime>=1000) {
			fps=(float)fpsFrames/((float)(time-fpsLastTime)/1000.0f);
			fpsFrames=0;
			fpsLastTime=time;
			if(Options.DisplayFPS) {
				if(fps>0.1) {
					_gcvt_s(fpsString,10,fps,4);
				}
			}
		}
		if(Options.FPSHold) {
			if(LoadedMemMap) {
				if(!fpsAdjusting&&(fps>((float)Options.FPSDesired)+((float)Options.FPSTolerence))||
					fpsAdjusting&&(fps>((float)Options.FPSDesired)+((float)Options.FPSTolerence)/2.0f)) {
						fpsAdjusting=true;
						float f=read_float(MemMap[eView0]);
						SatUp(&f,MaxViewRange,fps-(float)Options.FPSDesired);
						write_float(MemMap[eView0],f);
						write_float(MemMap[12],f);
						write_float(MemMap[15],f);
						write_float(MemMap[20],f);
						write_float(MemMap[23],f);
						if(Options.BindAIToView) {
							float f2=f/STDMAXVIEWDISTANCE;
							write_float(MemMap[eAI],f2);
						}
				} else if(!fpsAdjusting&&(fps<((float)Options.FPSDesired)-((float)Options.FPSTolerence))||
					fpsAdjusting&&(fps<((float)Options.FPSDesired)-((float)Options.FPSTolerence)/2.0f)) {
						fpsAdjusting=true;
						float f=read_float(MemMap[eView0]);
						SatDown(&f,MinViewRange,(float)Options.FPSDesired-fps);
						write_float(MemMap[eView0],f);
						write_float(MemMap[12],f);
						write_float(MemMap[15],f);
						write_float(MemMap[20],f);
						write_float(MemMap[23],f);
						if(Options.BindAIToView) {
							float f2=f/STDMAXVIEWDISTANCE;
							write_float(MemMap[eAI],f2);
						}
				} else if(fpsAdjusting) fpsAdjusting=false;
			}
		}
		if(Options.HWShader||Options.UseAAFix||UsedMWSEShaders) {
			RealDevice->StretchRect(BackBuffer,0,ShaderStruct.thisframeSurface,0,D3DTEXF_NONE);
		}
		//Hardware shaders
		if(UsedMWSEShaders||Options.HWShader) {
			DWORD FVF;
			DWORD CullMode;
			UINT passes=0;
			IDirect3DBaseTexture9* textures[6];
			RealDevice->BeginScene();
			for(int j=0;j<6;j++) {
				RealDevice->GetTexture(j,&textures[j]);
				RealDevice->SetTexture(0,ShaderStruct.thisframe);
			}
			RealDevice->SetStreamSource(0,ShaderStruct.buffer,0,sizeof(MyVertex));
			RealDevice->GetFVF(&FVF);
			RealDevice->SetFVF(MYVERTEXFORMAT);
			RealDevice->GetRenderState(D3DRS_CULLMODE,&CullMode);
			RealDevice->SetRenderState(D3DRS_CULLMODE,2);

			int i=0;
			if(Options.HWShader) {
				for(i=0;i<ShaderStruct.DefEffectCount;i++) {
					if(i>0) {
						RealDevice->StretchRect(BackBuffer,0,ShaderStruct.lastshaderSurface,0,D3DTEXF_NONE);
						if(i==1) {
							for(int j=0;j<6;j++) RealDevice->SetTexture(j,ShaderStruct.lastshader);
						}
					}
					SetShaderVars(ShaderStruct.DefaultEffects[i],time);
					ShaderStruct.DefaultEffects[i]->Begin(&passes,0);
					for(UINT pass = 0; pass < passes; pass++) {
						ShaderStruct.DefaultEffects[i]->BeginPass(pass);
						RealDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 );
						ShaderStruct.DefaultEffects[i]->EndPass();
						RealDevice->StretchRect(BackBuffer,0,ShaderStruct.lastpassSurface,0,D3DTEXF_NONE);
					}
					ShaderStruct.DefaultEffects[i]->End();
				}
			}
			if(UsedMWSEShaders) {
				for(int j=0;j<MAXMWSESHADERS;j++) {
					if(!ShaderStruct.MWSEeffects[j].Active) continue;
					if(i>0) {
						RealDevice->StretchRect(BackBuffer,0,ShaderStruct.lastshaderSurface,0,D3DTEXF_NONE);
						if(i==1) {
							for(int k=0;k<6;k++) RealDevice->SetTexture(k,ShaderStruct.lastshader);
						}
					}
					ShaderStruct.MWSEeffects[j].effect->Begin(&passes,0);
					for(UINT pass = 0; pass < passes; pass++) {
						ShaderStruct.MWSEeffects[j].effect->BeginPass(pass);
						RealDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 );
						ShaderStruct.MWSEeffects[j].effect->EndPass();
						RealDevice->StretchRect(BackBuffer,0,ShaderStruct.lastpassSurface,0,D3DTEXF_NONE);
					}
					ShaderStruct.MWSEeffects[j].effect->End();
					i++;
				}
			}
			//clean up
			RealDevice->SetRenderState(D3DRS_CULLMODE,CullMode);
			RealDevice->SetFVF(FVF);
			RealDevice->StretchRect(ShaderStruct.thisframeSurface,0,ShaderStruct.lastframeSurface,0,D3DTEXF_NONE);
			for(int i=0;i<6;i++) {
				RealDevice->SetTexture(i,textures[i]);
				if(textures[i]!=NULL) textures[i]->Release();
			}
			RealDevice->EndScene();
		}

		//fps counter
		if(Options.DisplayMessages&&(Options.DisplayFPS||StatusStringTime)) {
			RealDevice->BeginScene();
			if(Options.DisplayFPS) {
				Font->DrawTextA(0,fpsString,-1,&fpsRect,0,fpsColor);
			}
			if(StatusStringTime) {
				if(StatusStringTime>time) {
					Font->DrawTextA(0,StatusString,-1,&StatusRect,0,fpsColor);
				} else {
					StatusStringTime=0;
				}
			}
			RealDevice->EndScene();
		}
		hr=RealDevice->Present(a,b,c,d);
		//cpu idle
		if(Options.CPUIdle) Sleep(Options.IdleTime);
		//Antialiasing colour fix
		if(Options.UseAAFix) {
			RealDevice->StretchRect(ShaderStruct.thisframeSurface,0,BackBuffer,0,D3DTEXF_NONE);
		}
		//Handles model movement, fog zooming and zoom zooming
		if(TimerFrequency) {
			__int64 TimerNow;
			if(QueryPerformanceCounter((LARGE_INTEGER*)&TimerNow)) {
				if(TimerNow>LastTimer) {
					JointInfo.FrameTime=( ((float)(TimerNow-LastTimer))/TimerFrequency );
					LastTimer=TimerNow;
					for(int i=0;i<MAXMWSEOBJECTS;i++) {
						if((!Objects[i].Enabled)||(!Objects[i].SpeedFlags)) continue;
						if(Objects[i].SpeedFlags&0x01) {
							Objects[i].pos+=(Objects[i].speed*JointInfo.FrameTime);
							MWSE_UpdateObjectMatrix(i,0);
						}
						if(Objects[i].SpeedFlags&0x02) {
							Objects[i].rot+=(Objects[i].AngSpeed*JointInfo.FrameTime);
							MWSE_UpdateObjectMatrix(i,1);
						}
						if(Objects[i].SpeedFlags&0x04) {
							Objects[i].scale+=(Objects[i].growth*JointInfo.FrameTime);
							MWSE_UpdateObjectMatrix(i,2);
						}
					}
					if(FogZooming) {
						FogRange=CLAMP(FogRange+(FogZoomRate*JointInfo.FrameTime),MINFOG,MAXFOG);
						UpdateFogRange();
					}
					if(FOVZooming) {
						FOV=CLAMP(FOV+(FOVZoomRate*JointInfo.FrameTime),MINZOOM,MAXZOOM);
					}
					if(RotSpinning) {
						Rotation+=(RotSpinRate*JointInfo.FrameTime);
						UpdateRotation();
					}
					if(CameraShakeChange) {
						CameraShakeAmp=CLAMP(CameraShakeAmp+(CameraShakeChangeRate*JointInfo.FrameTime),MINSHAKE,MAXSHAKE);
					}
				}
			}
			if(CameraShake) UpdateScreenShake();
			MWSE_Update();
			ActiveLights=0;
		}

		if(SceneCount<=2) {
			UpdateScene=1;
			HUDScene=1;
			AlphaScene=-1;
			SolarScene=-1;
		} else if(UpdateScene==1||LastSceneCount!=SceneCount) {
			UpdateSceneNums();
			UpdateScene=0;
		}
		LastSceneCount=SceneCount;
		SceneCount=0;
		RenderDLand=true;
		return hr;
	}
	HRESULT _stdcall GetBackBuffer(UINT a,D3DBACKBUFFER_TYPE b,IDirect3DSurface8** c) {
		IDirect3DSurface9* c2=NULL;
		*c=NULL;
		DWORD unused=4;
		HRESULT hr=RealDevice->GetBackBuffer(0,a,b,&c2);
		if(hr!=D3D_OK||c2==NULL) return hr;
		hr=c2->GetPrivateData(guid,(void*)c,&unused);
		if(hr!=D3D_OK) *c=new FakeSurface(c2);
		return D3D_OK;
	}
	HRESULT _stdcall GetRasterStatus(D3DRASTER_STATUS* a) {
		return RealDevice->GetRasterStatus(0,a);
	}
	void _stdcall SetGammaRamp(DWORD a,const D3DGAMMARAMP* b) {
		RealDevice->SetGammaRamp(0,a,b);
	}
	void _stdcall GetGammaRamp(D3DGAMMARAMP* a) {
		RealDevice->GetGammaRamp(0,a);
	}
	HRESULT _stdcall CreateTexture(UINT a,UINT b,UINT c,DWORD d,D3DFORMAT e,D3DPOOL f,IDirect3DTexture8** g) {
		if(HookNextTexture) {
			HookNextTexture=false;
			HookedTextures[NextHookIndex].tHook=new TextureHook(NextHookIndex,a,b);
			*g=(IDirect3DTexture8*)HookedTextures[NextHookIndex].tHook;
			return D3D_OK;
		}
		IDirect3DTexture9* g2=NULL;
		HRESULT hr=RealDevice->CreateTexture(a,b,c,d,e,f,&g2,NULL);
		if(hr!=D3D_OK||g2==NULL) return hr;
		*g=new FakeTexture(g2);
		if(NextTexIsWater) {
			HookWaterLoad(&((FakeTexture*)*g)->RealObject);
			NextTexIsWater=false;
		}
		return D3D_OK;
	}
	HRESULT _stdcall CreateVolumeTexture(UINT a,UINT b,UINT c,UINT d,DWORD e,D3DFORMAT f,D3DPOOL g,IDirect3DVolumeTexture8** h) {
		return UnusedFunction();
	}
	HRESULT _stdcall CreateCubeTexture(UINT a,UINT b,DWORD c,D3DFORMAT d,D3DPOOL e,IDirect3DCubeTexture8** f) {
		return UnusedFunction();
	}
	HRESULT _stdcall CreateVertexBuffer(UINT a,DWORD b,DWORD c,D3DPOOL d,IDirect3DVertexBuffer8** e) {
		return RealDevice->CreateVertexBuffer(a,b,c,d,(IDirect3DVertexBuffer9**)e,NULL);
	}
	HRESULT _stdcall CreateIndexBuffer(UINT a,DWORD b,D3DFORMAT c,D3DPOOL d,IDirect3DIndexBuffer8** e) {
		return RealDevice->CreateIndexBuffer(a,b,c,d,(IDirect3DIndexBuffer9**)e,NULL);
	}
	HRESULT _stdcall CreateRenderTarget(UINT a,UINT b,D3DFORMAT c,D3DMULTISAMPLE_TYPE d,BOOL e,IDirect3DSurface8** f) {
		IDirect3DSurface9* f2=NULL;
		HRESULT hr=RealDevice->CreateRenderTarget(a,b,c,d,0,e,&f2,NULL);
		if(hr!=D3D_OK||f2==NULL) return hr;
		*f=new FakeSurface(f2);
		return D3D_OK;
	}
	HRESULT _stdcall CreateDepthStencilSurface(UINT a,UINT b,D3DFORMAT c,D3DMULTISAMPLE_TYPE d,IDirect3DSurface8** e) {
		//Not sure if Discard should be true or false
		IDirect3DSurface9* e2=NULL;
		HRESULT hr=RealDevice->CreateDepthStencilSurface(a,b,c,d,0,false,&e2,NULL);
		if(hr!=D3D_OK) return hr;
		*e=new FakeSurface(e2);
		return D3D_OK;
	}
	HRESULT _stdcall CreateImageSurface(UINT a,UINT b,D3DFORMAT c,IDirect3DSurface8** d) {
		IDirect3DSurface9* d2=NULL;
		//Use default to get stretchrect to work, scratch otherwise
		HRESULT hr=RealDevice->CreateOffscreenPlainSurface(a,b,c,D3DPOOL_SYSTEMMEM,&d2,NULL);
		//HRESULT hr=RealDevice->CreateOffscreenPlainSurface(a,b,c,D3DPOOL_DEFAULT,&d2,NULL);
		if(hr!=D3D_OK||d2==NULL) return hr;
		*d=new FakeSurface(d2);
		return D3D_OK;
	}
	HRESULT _stdcall CopyRects(IDirect3DSurface8* a,const RECT* b,UINT c,IDirect3DSurface8* d,const POINT* e) {
		IDirect3DSurface9* a2=((FakeSurface*)a)->RealObject;
		IDirect3DSurface9* d2=((FakeSurface*)d)->RealObject;
		if(b==NULL&&e==NULL) {
			D3DSURFACE_DESC9 source;
			D3DSURFACE_DESC9 dest;
			if(a2->GetDesc(&source)!=D3D_OK||d2->GetDesc(&dest)!=D3D_OK) {
				return UnusedFunction();
			} else if(source.Usage==1&&dest.Usage==0) {
				return RealDevice->GetRenderTargetData(a2,d2);
			} else if(source.Usage==0&&dest.Usage==1) {
				return RealDevice->UpdateSurface(a2,NULL,d2,NULL);
			} else {
				return RealDevice->StretchRect(a2,NULL,d2,NULL,D3DTEXF_NONE);
			}
		} else {
			//Really cant be bothered to get this working unless absolutely neccessery
			return UnusedFunction();
			/*HRESULT hr;
			RECT r;
			r.top=e->y;
			r.left=e->x;
			for(UINT i=0;i<c;i++) {
			r.bottom=r.top+b[i].bottom-b[i].top;
			r.right=r.left+b[i].right-b[i].left;
			hr=RealDevice->StretchRect(a2,&b[i],d2,&r,D3DTEXF_LINEAR);
			}
			return hr;*/
		}
	}
	HRESULT _stdcall UpdateTexture(IDirect3DBaseTexture8* a,IDirect3DBaseTexture8* b) {
		IDirect3DTexture9* a2=((FakeTexture*)a)->RealObject;
		IDirect3DTexture9* b2=((FakeTexture*)b)->RealObject;
		return RealDevice->UpdateTexture(a2,b2);
	}
	HRESULT _stdcall GetFrontBuffer(IDirect3DSurface8* a) {
		return UnusedFunction();
	}
	HRESULT _stdcall SetRenderTarget(IDirect3DSurface8* a,IDirect3DSurface8* b) {
		IDirect3DSurface9* a2=NULL;
		IDirect3DSurface9* b2=NULL;
		if(a!=NULL) a2=((FakeSurface*)a)->RealObject;
		if(b!=NULL) b2=((FakeSurface*)b)->RealObject;
		MyBackbuffer=(a2==BackBuffer);
		if(MyBackbuffer&&SceneCount==1) SceneCount=0;
		if(a2==BackBuffer||a2==NULL) {
			RealDevice->SetRenderState(D3DRS_RANGEFOGENABLE,Options.RangedFog);
		} else {
			RealDevice->SetRenderState(D3DRS_RANGEFOGENABLE,0);
		}
		HRESULT hr=D3D_OK;
		if(a2!=NULL) hr|=RealDevice->SetRenderTarget(0,a2);
		hr|=RealDevice->SetDepthStencilSurface(b2);
		return hr;
	}
	HRESULT _stdcall GetRenderTarget(IDirect3DSurface8** a) {
		IDirect3DSurface9* a2=NULL;
		*a=NULL;
		DWORD unused=4;
		HRESULT hr=RealDevice->GetRenderTarget(0,&a2);
		if(hr!=D3D_OK||a2==NULL) return hr;
		hr=a2->GetPrivateData(guid,(void*)a,&unused);
		if(hr!=D3D_OK) *a=new FakeSurface(a2);
		return D3D_OK;
	}
	HRESULT _stdcall GetDepthStencilSurface(IDirect3DSurface8** a) {
		IDirect3DSurface9* a2=NULL;
		*a=NULL;
		DWORD unused=4;
		HRESULT hr=RealDevice->GetDepthStencilSurface(&a2);
		if(hr!=D3D_OK||a2==NULL) return hr;
		hr=a2->GetPrivateData(guid,(void*)a,&unused);
		if(hr!=D3D_OK) *a=new FakeSurface(a2);
		return D3D_OK;
	}
	HRESULT _stdcall BeginScene() {
		if(SunManual) {
			if(HUDScene>1) {
				if(SceneCount==HUDScene-1) RealDevice->SetRenderState(D3DRS_AMBIENT,0xFFFFFFFF);
				else if(SceneCount==HUDScene) {
					if(HUDManual) RealDevice->SetRenderState(D3DRS_AMBIENT,HUDColour);
				}
				if(SceneCount==0) {
					if(SkyManual) RealDevice->SetRenderState(D3DRS_AMBIENT,SkyColour);
					else RealDevice->SetRenderState(D3DRS_AMBIENT,0xFFFFFFFF);
				}
			} else {
				RealDevice->SetRenderState(D3DRS_AMBIENT,0xFFFFFFFF);
			}
		}
		return RealDevice->BeginScene();
	}
	HRESULT _stdcall EndScene(void) {
		if(MyBackbuffer) SceneCount++;
		if(LoadedMemMap&&IsExterior()&&UsingDistantLand&&!LoadingScreen&&MyBackbuffer) {
			if(SceneCount==1&&!RenderedLand) {
				DL_RenderLand();
				RenderedLand=true;
			}
			if(SceneCount==AlphaScene) {
				DL_RenderWater();
			}
		}
		//Draw any new models loaded by mods
		if(UsedMWSEobjects&&SceneCount==AlphaScene) {
			StateBlock->Capture();
			RealDevice->SetRenderState(D3DRS_ALPHABLENDENABLE,0);
			RealDevice->SetRenderState(D3DRS_DIFFUSEMATERIALSOURCE,D3DMCS_MATERIAL);
			RealDevice->SetRenderState(D3DRS_SPECULARMATERIALSOURCE,D3DMCS_MATERIAL);
			RealDevice->SetRenderState(D3DRS_AMBIENTMATERIALSOURCE,D3DMCS_MATERIAL);
			RealDevice->SetRenderState(D3DRS_SRCBLEND,D3DBLEND_ONE);
			RealDevice->SetRenderState(D3DRS_DESTBLEND,D3DBLEND_ZERO);
			RealDevice->SetRenderState(D3DRS_ALPHATESTENABLE,false);
			for(DWORD i=0;i<ActiveLights;i++) {
				RealDevice->LightEnable(LightIDs[i],false);
			}
			for(int i=0;i<MAXMWSEOBJECTS;i++) {
				if(!Objects[i].Enabled) continue;
				RealDevice->SetTransform(D3DTS_WORLDMATRIX(0),&Objects[i].WorldMatrix);
				for(DWORD j=0;j<Objects[i].NumLights;j++) {
					RealDevice->LightEnable(Objects[i].lights[j],true);
				}
				if(Objects[i].IsMesh) {
					for(int j=1;j<8;j++) RealDevice->SetTexture(j,NULL);  //Clear unused textures
					if(Objects[i].effect!=NULL) {
						UINT passes;
						D3DXMATRIX worldviewproj=Objects[i].WorldMatrix*TransView*TransProj;
						Objects[i].effect->SetMatrix("worldviewproj",&worldviewproj);
						Objects[i].effect->Begin(&passes,0);
						for(UINT j=0;j<passes;j++) {
							Objects[i].effect->BeginPass(j);
							for(UINT k=0;k<Objects[i].subsets;k++) {
								RealDevice->SetTexture(0,Objects[i].textures[k]);
								Objects[i].mesh->DrawSubset(k);
							}
							Objects[i].effect->EndPass();
						}
						Objects[i].effect->End();
					} else {
						for(UINT k=0;k<Objects[i].subsets;k++) {
							RealDevice->SetTexture(0,Objects[i].textures[k]);
							Objects[i].mesh->DrawSubset(k);
						}
					}
				} else {
					for(int j=0;j<8;j++) RealDevice->SetTexture(j,Objects[i].textures[j]);
					RealDevice->SetStreamSource(0,Objects[i].vBuffer,0,Objects[i].stride);
					for(DWORD j=1;j<d3d8Caps.MaxStreams;j++) RealDevice->SetStreamSource(j,NULL,0,0);
					RealDevice->SetIndices(Objects[i].iBuffer);
					RealDevice->SetFVF(Objects[i].FVF);
					if(Objects[i].effect!=NULL) {
						UINT passes;
						D3DXMATRIX worldviewproj=Objects[i].WorldMatrix*TransView*TransProj;
						Objects[i].effect->SetMatrix("worldviewproj",&worldviewproj);
						Objects[i].effect->Begin(&passes,0);
						for(UINT j=0;j<passes;j++) {
							Objects[i].effect->BeginPass(j);
							RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,Objects[i].vertices,0,0,
								Objects[i].faces);
							Objects[i].effect->EndPass();
						}
						Objects[i].effect->End();
					} else {
						RealDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,Objects[i].vertices,0,0,
							Objects[i].faces);
					}
				}
				for(DWORD j=0;j<Objects[i].NumLights;j++) {
					RealDevice->LightEnable(Objects[i].lights[j],false);
				}
			}
			StateBlock->Apply();
		}
		if(Options.UseLagFix) {
			Font->DrawTextA(0," ",-1,&StatusRect,0,fpsColor);
		}
		return RealDevice->EndScene();
	}
	HRESULT _stdcall Clear(DWORD a,const D3DRECT* b,DWORD c,D3DCOLOR d,float e,DWORD f) {
		if(SunManual) {
			if(ClsManual) d=ClsColour; else ClsColour=d;
		}
		return RealDevice->Clear(a,b,c,d,e,f);
	}
	HRESULT _stdcall SetTransform(D3DTRANSFORMSTATETYPE a,const D3DMATRIX* b) {
		if(a==D3DTS_VIEW||a==D3DTS_PROJECTION) {
			if(!Options.ZoomUI&&(SceneCount<HUDScene)) {
				D3DXMATRIX matrix=*b;
				if(Zoom&&a==D3DTS_PROJECTION) {
					if(Options.CorrectAspectOnZoom) {
						D3DXMatrixPerspectiveFovLH(&matrix,D3DX_PI/FOV,JointInfo.aspect,1,100);
					} else {
						D3DXMatrixPerspectiveFovLH(&matrix,D3DX_PI/FOV,4.0f/3.0f,1,100);
					}
					matrix._33=b->_33;
					matrix._43=b->_43;
				}
				if(Rotation&&a==D3DTS_VIEW) {
					matrix*=RotateMatrix;
				}
				if(CameraShake&&a==D3DTS_VIEW) {
					matrix*=CameraShakeMatrix;
				}
				if(SceneCount==0&&MyBackbuffer) {
					D3DXVECTOR4 blank;
					D3DXMATRIX invmat;
					switch(a) {
					case D3DTS_VIEW:
						if((matrix._11==0||matrix._11==1)&&
							matrix._12==0&&
							(matrix._13==1||matrix._13==0)&&
							matrix._14==0&&
							matrix._21==0&&
							(matrix._22==1||matrix._22==0)&&
							(matrix._23==0||matrix._23==1)&&
							matrix._24==0&&
							(matrix._31==1||matrix._31==0)&&
							(matrix._32==0||matrix._32==1)&&
							matrix._33==0&&
							matrix._34==0&&
							matrix._41==0&&
							(matrix._42==0||matrix._42==-600)&&
							matrix._43==0&&
							matrix._44==1) {
								LoadingScreen=true;
						} else {
							TransView=matrix;
							LoadingScreen=false;
						}
						EyeVec.x=b->_13;
						EyeVec.y=b->_23;
						EyeVec.z=b->_33;
						ZeroStruct(blank);
						blank.w=1;
						D3DXMatrixInverse(&invmat,0,&matrix);
						D3DXVec4Transform(&EyePos,&blank,&invmat);
						if(LoadedMemMap&&IsExterior()&&!IsMenu()&&UsingDistantLand&&LastSceneCount>1&&!LoadingScreen) {
							DL_PlayerPos(EyePos.x,EyePos.y);
						}

						break;
					case D3DTS_PROJECTION:
						TransProj=matrix;
						break;
					}
				}
				return RealDevice->SetTransform(a,&matrix);
			}
		}
		return RealDevice->SetTransform(a,b);
	}
	HRESULT _stdcall GetTransform(D3DTRANSFORMSTATETYPE a,D3DMATRIX* b) {
		return RealDevice->GetTransform(a,b);
	}
	HRESULT _stdcall MultiplyTransform(D3DTRANSFORMSTATETYPE a,const D3DMATRIX* b) {
		return RealDevice->MultiplyTransform(a,b);
	}
	HRESULT _stdcall SetViewport(const D3DVIEWPORT8* a) {
		return RealDevice->SetViewport(a);
	}
	HRESULT _stdcall GetViewport(D3DVIEWPORT8* a) {
		return RealDevice->GetViewport(a);
	}
	HRESULT _stdcall SetMaterial(const D3DMATERIAL8* a) {
		return RealDevice->SetMaterial(a);
	}
	HRESULT _stdcall GetMaterial(D3DMATERIAL8* a) {
		return RealDevice->GetMaterial(a);
	}
	HRESULT _stdcall SetLight(DWORD a,const D3DLIGHT8* b) {
		if(a!=6) {
			const_cast<D3DLIGHT9*>(b)->Ambient.r=0; //Required so that distant land lighting matches with standard lighting
			const_cast<D3DLIGHT9*>(b)->Ambient.g=0;
			const_cast<D3DLIGHT9*>(b)->Ambient.b=0;
		}
		if(ActiveLights<MAXACTIVELIGHTS) {
			LightIDs[ActiveLights++]=a;
		}
		if(a==6) {
			if(Options.SetShaderVars||UsingDistantLand) {
				SunVec.x=b->Direction.x;
				SunVec.y=b->Direction.y;
				SunVec.z=b->Direction.z;
				D3DXVec3Normalize((D3DXVECTOR3*)&SunVec,(D3DXVECTOR3*)&SunVec);
				SunCol=*(D3DXVECTOR4*)&b->Diffuse;
				SunAmb=*(D3DXVECTOR4*)&b->Ambient;
				SunAmbD=D3DXCOLOR(b->Ambient);
			}
			if(SunManual) {
				D3DLIGHT8 light=*b;
				if(SunManDir) light.Direction=SunDirection; else SunDirection=light.Direction;
				if(SunManColour) light.Diffuse=SunColour; else SunColour=light.Diffuse;
				return RealDevice->SetLight(6,&light);
			}
		}
		return RealDevice->SetLight(a,b);
	}
	HRESULT _stdcall GetLight(DWORD a,D3DLIGHT8* b) {
		return RealDevice->GetLight(a,b);
	}
	HRESULT _stdcall LightEnable(DWORD a,BOOL b) {
		return RealDevice->LightEnable(a,b);
	}
	HRESULT _stdcall GetLightEnable(DWORD a,BOOL* b) {
		return RealDevice->GetLightEnable(a,b);
	}
	HRESULT _stdcall SetClipPlane(DWORD a,const float* b) {
		return RealDevice->SetClipPlane(a,b);
	}
	HRESULT _stdcall GetClipPlane(DWORD a,float* b) {
		return RealDevice->GetClipPlane(a,b);
	}
	HRESULT _stdcall SetRenderState(D3DRENDERSTATETYPE a,DWORD b) {
		if(a==D3DRS_ZWRITEENABLE&&b==1&&SceneCount==0&&LoadedMemMap&&UsingDistantLand&&!RenderedLand&&MyBackbuffer) {
			if(IsExterior()) {
				RenderedLand=true;
				DL_RenderLand();
			} else if(DistantLandActive) {
				DL_Deactivate();
			}
		}
		//Skip this render state if this is an option MGE controls
		if(a==D3DRS_FOGVERTEXMODE||a==D3DRS_FOGTABLEMODE)
			return D3D_OK;
		//Fogging is special. Setting it with the other render states causes a black screen
		if(a==D3DRS_FOGENABLE&&!Options.Fog) return D3D_OK;
		if(a==D3DRS_FOGSTART) {
			FogStart=DW2F(b);
			if(DistantLandActive) b=F2DW(Options.DL_Distance*8192*FogRange);
			else b=F2DW(FogStart*FogRange);
		} else if(a==D3DRS_FOGEND) {
			FogEnd=DW2F(b);
			if(DistantLandActive) b=F2DW(Options.DL_Fog*8192*FogRange);
			else b=F2DW(FogEnd*FogRange);
		}
		if(SceneCount<AlphaScene) {
			if(SunManual) {
				if(SceneCount<AlphaScene) {
					if(a==D3DRS_AMBIENT)  { 
						if(AmbientManual) {
							b=AmbientColour;
							AmbCol=(D3DXVECTOR4)D3DXCOLOR(b);
						} else {
							AmbientColour=b;
							AmbCol=(D3DXVECTOR4)D3DXCOLOR(b);
						}
					}
					if(a==D3DRS_FOGCOLOR) { if(FogManual)     b=FogColour;     else { D3DXCOLOR cv=(D3DXCOLOR)b; if(cv.r!=0&&cv.g!=0&&cv.b!=0) FogColour=b; } }
				} 
			} else {
				if(a==D3DRS_FOGCOLOR) {
					D3DXCOLOR cv=(D3DXCOLOR)b;
					if(cv.r!=0&&cv.g!=0&&cv.b!=0) FogColour=b;
				}
			}
		}
		if(SceneCount==HUDScene) {
			if(a==D3DRS_AMBIENT&&HUDManual) b=HUDColour;
		}
		return RealDevice->SetRenderState(a,b);
	}
	HRESULT _stdcall GetRenderState(D3DRENDERSTATETYPE a,DWORD* b) {
		return RealDevice->GetRenderState(a,b);
	}
	HRESULT _stdcall BeginStateBlock(void) {
		return UnusedFunction();
	}
	HRESULT _stdcall EndStateBlock(DWORD* a) {
		return UnusedFunction();
	}
	HRESULT _stdcall ApplyStateBlock(DWORD a) {
		return UnusedFunction();
	}
	HRESULT _stdcall CaptureStateBlock(DWORD a) {
		return UnusedFunction();
	}
	HRESULT _stdcall DeleteStateBlock(DWORD a) {
		return UnusedFunction();
	}
	HRESULT _stdcall CreateStateBlock(D3DSTATEBLOCKTYPE a,DWORD* b) {
		return UnusedFunction();
	}
	HRESULT _stdcall SetClipStatus(const D3DCLIPSTATUS8* a) {
		return UnusedFunction();
	}
	HRESULT _stdcall GetClipStatus(D3DCLIPSTATUS8* a) {
		return UnusedFunction();
	}
	HRESULT _stdcall GetTexture(DWORD a,IDirect3DBaseTexture8** b) {
		return UnusedFunction();
	}
	HRESULT _stdcall SetTexture(DWORD a,IDirect3DBaseTexture8* b) {
		if(Options.HookTextures) {
			if(SetHookedTextures[a]!=255) {
				HookedTextures[SetHookedTextures[a]].tHook->Release();
				SetHookedTextures[a]=255;
			}
			if(HookedShaderActive&&a==0) {
				ShaderStruct.MWSEeffects[ActiveHookedShader].effect->EndPass();
				ShaderStruct.MWSEeffects[ActiveHookedShader].effect->End();
				HookedShaderActive=0;
			}
			if(b!=NULL&&b->GetLOD()==HOOKEDTEXTUREID) {
				BYTE d=((TextureHook*)b)->HookIndex;
				ActiveHookedShader = HookedTextures[d].ShaderIndex;
				SetHookedTextures[a]=d;
				HookedTextures[d].tHook->AddRef();

				if(a==0&&HookedTextures[d].ShaderIndex!=255) {
					ShaderStruct.MWSEeffects[ActiveHookedShader].effect->Begin(&HookedShaderActive,0);
					if(HookedShaderActive==1) {
						ShaderStruct.MWSEeffects[ActiveHookedShader].effect->BeginPass(0);
					}
				}
				return RealDevice->SetTexture(a,LoadedTextures[HookedTextures[d].TextureIndex].texture);
			}
		}
		if(SunManual&&SunTexture&&!Indoors&&(SceneCount==SolarScene)) return RealDevice->SetTexture(a,SunTexture);
		IDirect3DTexture9* b2=NULL;
		if(b!=NULL) b2=((FakeTexture*)b)->RealObject;
		return RealDevice->SetTexture(a,b2);
	}
	HRESULT _stdcall GetTextureStageState(DWORD a,D3DTEXTURESTAGESTATETYPE b,DWORD* c) {
		//Some bugger decided to split this into two functions :(
		D3DSAMPLERSTATETYPE sampler;
		switch(b) {
			case D3DTSS_ADDRESSU:
				sampler=D3DSAMP_ADDRESSU;
				break;
			case D3DTSS_ADDRESSV:
				sampler=D3DSAMP_ADDRESSV;
				break;
			case D3DTSS_BORDERCOLOR:
				sampler=D3DSAMP_BORDERCOLOR;
				break;
			case D3DTSS_MAGFILTER:
				sampler=D3DSAMP_MAGFILTER;
				break;
			case D3DTSS_MINFILTER:
				sampler=D3DSAMP_MINFILTER;
				break;
			case D3DTSS_MIPFILTER:
				sampler=D3DSAMP_MIPFILTER;
				break;
			case D3DTSS_MIPMAPLODBIAS:
				sampler=D3DSAMP_MIPMAPLODBIAS;
				break;
			case D3DTSS_MAXMIPLEVEL:
				sampler=D3DSAMP_MAXMIPLEVEL;
				break;
			case D3DTSS_MAXANISOTROPY:
				sampler=D3DSAMP_MAXANISOTROPY;
				break;
			case D3DTSS_ADDRESSW:
				sampler=D3DSAMP_ADDRESSW;
				break;
			default:return RealDevice->GetTextureStageState(a,b,c);
		}
		return RealDevice->GetSamplerState(a,sampler,c);
	}
	HRESULT _stdcall SetTextureStageState(DWORD a,D3DTEXTURESTAGESTATETYPE b,DWORD c) {
		//if(b==11) return D3DERR_INVALIDCALL;
		//If it's something MGE controls then override it
		if(a==0) {
			if((b==D3DTSS_MINFILTER/*||b==D3DTSS_MAGFILTER*/)&&c==2) {
				D3DSAMP_MINFILTER;
				return RealDevice->SetSamplerState(0,(D3DSAMPLERSTATETYPE)(b-11),Options.MagMinFilter);
			}
			if(b==D3DTSS_MIPFILTER&&c==2) {
				return RealDevice->SetSamplerState(0,D3DSAMP_MIPFILTER,Options.MipFilter);
			}
			if(b==D3DTSS_MIPMAPLODBIAS&&Options.LodBias) {
				float f=Options.LodBias;
				return RealDevice->SetSamplerState(0,D3DSAMP_MIPMAPLODBIAS,*(DWORD*)&f);
			}
		}
		//Some bugger decided to split this into two functions :(
		D3DSAMPLERSTATETYPE sampler;
		switch(b) {
			case D3DTSS_ADDRESSU:
				sampler=D3DSAMP_ADDRESSU;
				break;
			case D3DTSS_ADDRESSV:
				sampler=D3DSAMP_ADDRESSV;
				break;
			case D3DTSS_BORDERCOLOR:
				sampler=D3DSAMP_BORDERCOLOR;
				break;
			case D3DTSS_MAGFILTER:
				sampler=D3DSAMP_MAGFILTER;
				break;
			case D3DTSS_MINFILTER:
				sampler=D3DSAMP_MINFILTER;
				break;
			case D3DTSS_MIPFILTER:
				sampler=D3DSAMP_MIPFILTER;
				break;
			case D3DTSS_MIPMAPLODBIAS:
				sampler=D3DSAMP_MIPMAPLODBIAS;
				break;
			case D3DTSS_MAXMIPLEVEL:
				sampler=D3DSAMP_MAXMIPLEVEL;
				break;
			case D3DTSS_MAXANISOTROPY:
				sampler=D3DSAMP_MAXANISOTROPY;
				break;
			case D3DTSS_ADDRESSW:
				sampler=D3DSAMP_ADDRESSW;
				break;
			default:return RealDevice->SetTextureStageState(a,b,c);
		}
		return RealDevice->SetSamplerState(a,sampler,c);
	}
	HRESULT _stdcall ValidateDevice(DWORD* a) {
		return RealDevice->ValidateDevice(a);
	}
	HRESULT _stdcall GetInfo(DWORD a,void* b,DWORD c) {
		//Doesn't seem to have a DX9 equivilent
		return UnusedFunction();
	}
	HRESULT _stdcall SetPaletteEntries(UINT a,const PALETTEENTRY* b) {
		return RealDevice->SetPaletteEntries(a,b);
	}
	HRESULT _stdcall GetPaletteEntries(UINT a,PALETTEENTRY* b) {
		return RealDevice->GetPaletteEntries(a,b);
	}
	HRESULT _stdcall SetCurrentTexturePalette(UINT a) {
		return RealDevice->SetCurrentTexturePalette(a);
	}
	HRESULT _stdcall GetCurrentTexturePalette(UINT *a) {
		return RealDevice->GetCurrentTexturePalette(a);
	}
	HRESULT _stdcall DrawPrimitive(D3DPRIMITIVETYPE a,UINT b,UINT c) {
		if(HookedShaderActive>1) {
			HRESULT hr=0;
			for(DWORD i=0;i<HookedShaderActive;i++) {
				ShaderStruct.MWSEeffects[ActiveHookedShader].effect->BeginPass(i);
				hr|=RealDevice->DrawPrimitive(a,b,c);
			}
			return hr;
		}
		return RealDevice->DrawPrimitive(a,b,c);
	}
	HRESULT _stdcall DrawIndexedPrimitive(D3DPRIMITIVETYPE a,UINT b,UINT c,UINT d,UINT e) {
		if(HookedShaderActive>1) {
			HRESULT hr=0;
			for(DWORD i=0;i<HookedShaderActive;i++) {
				ShaderStruct.MWSEeffects[ActiveHookedShader].effect->BeginPass(i);
				return RealDevice->DrawIndexedPrimitive(a,BaseVertexIndex,b,c,d,e);
			}
			return hr;
		}
		return RealDevice->DrawIndexedPrimitive(a,BaseVertexIndex,b,c,d,e);
	}
	HRESULT _stdcall DrawPrimitiveUP(D3DPRIMITIVETYPE a,UINT b,const void* c,UINT d) {
		if(HookedShaderActive>1) {
			HRESULT hr=0;
			for(DWORD i=0;i<HookedShaderActive;i++) {
				ShaderStruct.MWSEeffects[ActiveHookedShader].effect->BeginPass(i);
				hr|=RealDevice->DrawPrimitiveUP(a,b,c,d);
			}
			return hr;
		}
		return RealDevice->DrawPrimitiveUP(a,b,c,d);
	}
	HRESULT _stdcall DrawIndexedPrimitiveUP(D3DPRIMITIVETYPE a,UINT b,UINT c,UINT d,const void* e,D3DFORMAT f,const void* g,UINT h) {
		if(HookedShaderActive>1) {
			HRESULT hr=0;
			for(DWORD i=0;i<HookedShaderActive;i++) {
				ShaderStruct.MWSEeffects[ActiveHookedShader].effect->BeginPass(i);
				hr|=RealDevice->DrawIndexedPrimitiveUP(a,b,c,d,e,f,g,h);
			}
			return hr;
		}
		return RealDevice->DrawIndexedPrimitiveUP(a,b,c,d,e,f,g,h);
	}
	HRESULT _stdcall ProcessVertices(UINT a,UINT b,UINT c,IDirect3DVertexBuffer8* d,DWORD e) {
		return RealDevice->ProcessVertices(a,b,c,(IDirect3DVertexBuffer9*)d,NULL,e);
	}
	HRESULT _stdcall CreateVertexShader(const DWORD* a,const DWORD* b,DWORD* c,DWORD d) {
		*c=(vShaderCount+=2);
		ID3DXBuffer* ShaderData=NULL;
		switch(*c) {
			case 3:D3DXAssembleShaderFromFileA("Data Files\\shaders\\water\\vs1.txt",NULL,NULL,0,&ShaderData,NULL);break;
			case 5:D3DXAssembleShaderFromFileA("Data Files\\shaders\\water\\vs2.txt",NULL,NULL,0,&ShaderData,NULL);break;
			case 7:D3DXAssembleShaderFromFileA("Data Files\\shaders\\water\\vs3.txt",NULL,NULL,0,&ShaderData,NULL);break;
			case 9:D3DXAssembleShaderFromFileA("Data Files\\shaders\\water\\vs4.txt",NULL,NULL,0,&ShaderData,NULL);break;
			default:
				ERRORMSG("Morrowind attempted to create a vertex shader which MGE did not expect");
				/*ID3DXBuffer* output;
				HANDLE h;
				DWORD unused;
				D3DXDisassembleShader(b,false,NULL,&output);
				h=CreateFileA("vout.txt",GENERIC_WRITE,0,0,CREATE_ALWAYS,0,0);
				WriteFile(h,output->GetBufferPointer(),output->GetBufferSize()-1,&unused,NULL);
				CloseHandle(h);
				D3DXAssembleShaderFromFileA("vout.txt",NULL,NULL,0,&ShaderData,&Errors);break;
				break;*/
		}
		HRESULT hr=RealDevice->CreateVertexShader((DWORD*)ShaderData->GetBufferPointer(),&NewVShaders[(*c-3)/2]);
		if(ShaderData) ShaderData->Release();
		return hr;
	}
	HRESULT _stdcall SetVertexShader(DWORD a) {
		CurrentVShader=a;
		switch(a) {
			case 0: RealDevice->SetVertexShader(NULL);
			case 3:
			case 5:
			case 9: RealDevice->SetVertexDeclaration(decl[0]); break; 
			case 7: RealDevice->SetVertexDeclaration(decl[1]); break;
			default:
				CurrentVShader=0;
				RealDevice->SetVertexShader(NULL);
				return RealDevice->SetFVF(a);
		}
		return RealDevice->SetVertexShader(NewVShaders[(a-3)/2]);
	}
	HRESULT _stdcall GetVertexShader(DWORD* a) {
		if(CurrentVShader!=0) {
			*a=CurrentVShader;
			return D3D_OK;
		} else {
			return RealDevice->GetFVF(a);
		}
	}
	HRESULT _stdcall DeleteVertexShader(DWORD a) {
		if(NewVShaders[(a-3)/2]!=NULL) {
			while(NewVShaders[(a-3)/2]->Release()) { }
			NewVShaders[(a-3)/2]=NULL;
		}
		return D3D_OK;
	}
	HRESULT _stdcall SetVertexShaderConstant(DWORD a,const void* b,DWORD c) {
		return RealDevice->SetVertexShaderConstantF(a,(float*)b,c);
	}
	HRESULT _stdcall GetVertexShaderConstant(DWORD a,void* b,DWORD c) {
		return RealDevice->GetVertexShaderConstantF(a,(float*)b,c);
	}
	HRESULT _stdcall GetVertexShaderDeclaration(DWORD a,void* b,DWORD* c) {
		return UnusedFunction();
	}
	HRESULT _stdcall GetVertexShaderFunction(DWORD a,void* b,DWORD* c) {
		return UnusedFunction();
	}
	HRESULT _stdcall SetStreamSource(UINT a,IDirect3DVertexBuffer8* b,UINT c) {
		return RealDevice->SetStreamSource(a,(IDirect3DVertexBuffer9*)b,0,c);
	}
	HRESULT _stdcall GetStreamSource(UINT a,IDirect3DVertexBuffer8** b,UINT* c) {
		UINT unused;
		return RealDevice->GetStreamSource(a,(IDirect3DVertexBuffer9**)b,&unused,c);
	}
	HRESULT _stdcall SetIndices(IDirect3DIndexBuffer8* a,UINT b) {
		BaseVertexIndex=b;  //BaseVertexIndex has been moved to ProcessVerticies()
		return RealDevice->SetIndices((IDirect3DIndexBuffer9*)a);
	}
	HRESULT _stdcall GetIndices(IDirect3DIndexBuffer8** a,UINT* b) {
		*b=BaseVertexIndex;
		return RealDevice->GetIndices((IDirect3DIndexBuffer9**)a);
	}
	HRESULT _stdcall CreatePixelShader(const DWORD* a,DWORD* b) {
		PixelShading=true;
		*b=++pShaderCount;
		ID3DXBuffer* ShaderData=NULL;
		switch(*b) {
			case 1:D3DXAssembleShaderFromFileA("Data Files\\shaders\\water\\ps1.txt",NULL,NULL,0,&ShaderData,NULL);break;
			case 2:D3DXAssembleShaderFromFileA("Data Files\\shaders\\water\\ps2.txt",NULL,NULL,0,&ShaderData,NULL);break;
			case 3:D3DXAssembleShaderFromFileA("Data Files\\shaders\\water\\ps3.txt",NULL,NULL,0,&ShaderData,NULL);break;
			case 4:D3DXAssembleShaderFromFileA("Data Files\\shaders\\water\\ps4.txt",NULL,NULL,0,&ShaderData,NULL);break;
			case 5:D3DXAssembleShaderFromFileA("Data Files\\shaders\\water\\ps5.txt",NULL,NULL,0,&ShaderData,NULL);break;
			case 6:D3DXAssembleShaderFromFileA("Data Files\\shaders\\water\\ps6.txt",NULL,NULL,0,&ShaderData,NULL);break;
			case 7:D3DXAssembleShaderFromFileA("Data Files\\shaders\\water\\ps7.txt",NULL,NULL,0,&ShaderData,NULL);break;
			case 8:D3DXAssembleShaderFromFileA("Data Files\\shaders\\water\\ps8.txt",NULL,NULL,0,&ShaderData,NULL);break;
			case 9:D3DXAssembleShaderFromFileA("Data Files\\shaders\\water\\ps9.txt",NULL,NULL,0,&ShaderData,NULL);break;
			default:
				ERRORMSG("Morrowind tried to create an unexpected pixel shader");
		}
		HRESULT hr=RealDevice->CreatePixelShader((DWORD*)ShaderData->GetBufferPointer(),&NewPShaders[*b-1]);
		if(hr!=D3D_OK) {
			DebugBreak();
		}
		if(ShaderData!=NULL) ShaderData->Release();
		return hr;
	}
	HRESULT _stdcall SetPixelShader(DWORD a) {
		CurrentPShader=a;
		if(a==0) return RealDevice->SetPixelShader(NULL);
		else return RealDevice->SetPixelShader(NewPShaders[a-1]);
	}
	HRESULT _stdcall GetPixelShader(DWORD* a) {
		*a=CurrentPShader;
		return D3D_OK;
	}
	HRESULT _stdcall DeletePixelShader(DWORD a) {
		if(NewPShaders[a-1]!=NULL) {
			while(NewPShaders[a-1]->Release()) { }
			NewPShaders[a-1]=NULL;
		}
		return D3D_OK;
	}
	HRESULT _stdcall SetPixelShaderConstant(DWORD a,const void* b,DWORD c) {
		return RealDevice->SetPixelShaderConstantF(a,(float*)b,c);
	}
	HRESULT _stdcall GetPixelShaderConstant(DWORD a,void* b,DWORD c) {
		return RealDevice->GetPixelShaderConstantF(a,(float*)b,c);
	}
	HRESULT _stdcall GetPixelShaderFunction(DWORD a,void* b,DWORD* c) {
		return UnusedFunction();
	}
	HRESULT _stdcall DrawRectPatch(UINT a,const float* b,const D3DRECTPATCH_INFO* c) {
		return RealDevice->DrawRectPatch(a,b,c);
	}
	HRESULT _stdcall DrawTriPatch(UINT a,const float* b,const D3DTRIPATCH_INFO* c) {
		return RealDevice->DrawTriPatch(a,b,c);
	}
	HRESULT _stdcall DeletePatch(UINT a) {
		return RealDevice->DeletePatch(a);
	}
};
class FakeObject : public IDirect3D8 {
public:
	/*** Fields ***/
	IDirect3D* RealObject;
	DWORD Refs;
	/*** Misc methods ***/
	FakeObject(IDirect3D* Real) {
		RealObject=Real;
		Refs=1;
	}
	/*** IUnknown methods ***/
	HRESULT _stdcall QueryInterface (REFIID a, LPVOID * b) {return RealObject->QueryInterface(a,b);}
	ULONG _stdcall AddRef(void) {
		return ++Refs;
	}
	ULONG _stdcall Release(void) {
		if(--Refs==0) {
			RealObject->Release();
			delete this;
			return 0;
		}
		return Refs;
	}
	/*** IDirect3D8 methods ***/
	HRESULT _stdcall RegisterSoftwareDevice(void* a) {
		return RealObject->RegisterSoftwareDevice(a);
	}
	UINT _stdcall GetAdapterCount(void) {
		return RealObject->GetAdapterCount();
	}
	HRESULT _stdcall GetAdapterIdentifier(UINT a,DWORD b,D3DADAPTER_IDENTIFIER8* c) {
		D3DADAPTER_IDENTIFIER9 c2;
		HRESULT hr=RealObject->GetAdapterIdentifier(a,0,&c2);
		if(hr==D3D_OK) {
			//Strings, so use copymemory instead of assignment
			::CopyMemory(c,&c2,1024);
			c->DeviceId=c2.DeviceId;
			c->DeviceIdentifier=c2.DeviceIdentifier;
			c->DriverVersion=c2.DriverVersion;
			c->Revision=c2.Revision;
			c->SubSysId=c2.SubSysId;
			c->VendorId=c2.VendorId;
			c->WHQLLevel=c2.WHQLLevel;
		}
		return hr;
	}
	UINT _stdcall GetAdapterModeCount(UINT a) {
		return RealObject->GetAdapterModeCount(a,D3DFMT_X8R8G8B8);
	}
	HRESULT _stdcall EnumAdapterModes(UINT a,UINT b,D3DDISPLAYMODE* c) {
		return RealObject->EnumAdapterModes(a,D3DFMT_X8R8G8B8,b,c);
	}
	HRESULT _stdcall GetAdapterDisplayMode(UINT a,D3DDISPLAYMODE* b) {
		//Hide the fact that windows is using a 16bit colour depth
		HRESULT hr=RealObject->GetAdapterDisplayMode(a,b);
		b->Format=D3DFMT_X8R8G8B8;
		return hr;
	}
	HRESULT _stdcall CheckDeviceType(UINT a,D3DDEVTYPE b,D3DFORMAT c,D3DFORMAT d,BOOL e) {
		return RealObject->CheckDeviceType(a,b,c,d,e);
	}
	HRESULT _stdcall CheckDeviceFormat(UINT a,D3DDEVTYPE b,D3DFORMAT c,DWORD d,D3DRESOURCETYPE e,D3DFORMAT f) {
		return RealObject->CheckDeviceFormat(a,b,c,d,e,f);
	}
	HRESULT _stdcall CheckDeviceMultiSampleType(UINT a,D3DDEVTYPE b,D3DFORMAT c,BOOL d,D3DMULTISAMPLE_TYPE e) {
		return RealObject->CheckDeviceMultiSampleType(a,b,c,d,e,NULL);
	}
	HRESULT _stdcall CheckDepthStencilMatch(UINT a,D3DDEVTYPE b,D3DFORMAT c,D3DFORMAT d,D3DFORMAT e) {
		return RealObject->CheckDepthStencilMatch(a,b,c,d,e);
	}
	HRESULT _stdcall GetDeviceCaps(UINT a,D3DDEVTYPE b,D3DCAPS8* c) {
		*c=d3d8Caps;
		return D3D_OK;
	}
	HMONITOR _stdcall GetAdapterMonitor(UINT a) {
		return RealObject->GetAdapterMonitor(a);
	}
	HRESULT _stdcall CreateDevice(UINT a,D3DDEVTYPE b,HWND c,DWORD d,D3DPRESENT_PARAMETERS8* e,IDirect3DDevice8** f) {
		D3DSURFACE_DESC BackBufferDesc;
		JointInfo.GraphicsWidth=(WORD)e->BackBufferWidth;
		JointInfo.GraphicsHeight=(WORD)e->BackBufferHeight;
		JointInfo.aspect=(float)e->BackBufferWidth/(float)e->BackBufferHeight;
		RcpRes.x=(1/(float)e->BackBufferWidth);
		RcpRes.y=(1/(float)e->BackBufferHeight);
		RcpRes.z=0;
		RcpRes.w=1;
		MaxViewRange=(float)(Options.MaxViewDistance*100);
		MinViewRange=(float)(Options.MinViewDistance*100);
		if(e->Windowed) {
			JointInfo.WNDwindowed=true;
		} else {
			JointInfo.WNDwindowed=false;
		}
		e->MultiSampleType=(D3DMULTISAMPLE_TYPE)Options.AA;
		if(Options.UseDistantLand) {
			e->AutoDepthStencilFormat=D3DFMT_D24X8;
		} else {
			e->AutoDepthStencilFormat=(D3DFORMAT)Options.ZBuffer;
		}
		if(Options.VWait==3) {
			e->FullScreen_PresentationInterval=0x80000000L;
		} else {
			e->FullScreen_PresentationInterval=Options.VWait;
		}
		if(!e->Windowed) {
			e->FullScreen_RefreshRateInHz=Options.RefreshRate;
		}
		if(e->Flags&D3DPRESENTFLAG_LOCKABLE_BACKBUFFER) e->Flags^=D3DPRESENTFLAG_LOCKABLE_BACKBUFFER;
		//Create some new presentation parameters if using DX9
		D3DPRESENT_PARAMETERS9 e2;
		e2.AutoDepthStencilFormat=e->AutoDepthStencilFormat;
		e2.BackBufferCount=e->BackBufferCount;
		e2.BackBufferFormat=e->BackBufferFormat;
		e2.BackBufferHeight=e->BackBufferHeight;
		e2.BackBufferWidth=e->BackBufferWidth;
		e2.EnableAutoDepthStencil=e->EnableAutoDepthStencil;
		e2.Flags=e->Flags;
		e2.FullScreen_RefreshRateInHz=e->FullScreen_RefreshRateInHz;
		e2.PresentationInterval=e->FullScreen_PresentationInterval;
		e2.hDeviceWindow=e->hDeviceWindow;
		e2.MultiSampleQuality=0;
		e2.MultiSampleType=e->MultiSampleType;
		e2.SwapEffect=e->SwapEffect;
		e2.Windowed=e->Windowed;
		HRESULT hr = RealObject->CreateDevice(a,b,c,d,&e2,&RealDevice);

		if(hr!=D3D_OK||RealObject==NULL) return hr;

		//Setup any render options we need (Do _NOT_ enable fog here!)
		FOV=Options.InitialZoom;
		//Setup render states
		RealDevice->SetSamplerState(0,D3DSAMP_MAXANISOTROPY,Options.AnsiLevel);
		RealDevice->SetSamplerState(0,D3DSAMP_MINFILTER,Options.MagMinFilter);
		//RealDevice->SetSamplerState(0,D3DSAMP_MAGFILTER,Options.MagMinFilter); //Only min filter needs to be changed
		{
			float ff=Options.LodBias;
			RealDevice->SetSamplerState(0,D3DSAMP_MIPMAPLODBIAS,*(DWORD*)&ff);
		}
		RealDevice->SetRenderState(D3DRS_FOGTABLEMODE,(D3DTEXTURESTAGESTATETYPE)Options.FogPixelMode);
		RealDevice->SetRenderState(D3DRS_FOGVERTEXMODE,(D3DTEXTURESTAGESTATETYPE)Options.FogVertexMode);
		RealDevice->SetRenderState(D3DRS_RANGEFOGENABLE,(D3DTEXTURESTAGESTATETYPE)Options.RangedFog);
		//RealDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_USEW);
		//RealDevice->SetRenderState(D3DRS_CLIPPING,(D3DTEXTURESTAGESTATETYPE)0);
		//Get a pointer to the default back buffer and set up stuff for fullscreen shaders
		hr=RealDevice->GetBackBuffer(0,0,D3DBACKBUFFER_TYPE_MONO,&BackBuffer);
		BackBuffer->GetDesc(&BackBufferDesc);
		hr=RealDevice->GetDepthStencilSurface(&DepthBuffer);
		//Create texture surfaces for fullscreen hardware shaders
		RealDevice->CreateTexture(BackBufferDesc.Width,BackBufferDesc.Height,1,D3DUSAGE_RENDERTARGET,BackBufferDesc.Format,
			D3DPOOL_DEFAULT,&ShaderStruct.thisframe,NULL);
		ShaderStruct.thisframe->GetSurfaceLevel(0,&ShaderStruct.thisframeSurface);
		RealDevice->CreateTexture(BackBufferDesc.Width,BackBufferDesc.Height,1,D3DUSAGE_RENDERTARGET,BackBufferDesc.Format,
			D3DPOOL_DEFAULT,&ShaderStruct.lastframe,NULL);
		ShaderStruct.lastframe->GetSurfaceLevel(0,&ShaderStruct.lastframeSurface);
		RealDevice->CreateTexture(BackBufferDesc.Width,BackBufferDesc.Height,1,D3DUSAGE_RENDERTARGET,BackBufferDesc.Format,
			D3DPOOL_DEFAULT,&ShaderStruct.lastpass,NULL);
		ShaderStruct.lastpass->GetSurfaceLevel(0,&ShaderStruct.lastpassSurface);
		RealDevice->CreateTexture(BackBufferDesc.Width,BackBufferDesc.Height,1,D3DUSAGE_RENDERTARGET,BackBufferDesc.Format,
			D3DPOOL_DEFAULT,&ShaderStruct.lastshader,NULL);
		ShaderStruct.lastshader->GetSurfaceLevel(0,&ShaderStruct.lastshaderSurface);
		ShaderVertices[3].x=ShaderVertices[2].x=(float)(BackBufferDesc.Width)-.5f;
		ShaderVertices[3].y=ShaderVertices[1].y=(float)(BackBufferDesc.Height)-.5f;
		RealDevice->CreateVertexBuffer(4*sizeof(MyVertex),D3DUSAGE_WRITEONLY,MYVERTEXFORMAT,D3DPOOL_DEFAULT,
			&ShaderStruct.buffer, 0);
		byte* VertexPointer;
		ShaderStruct.buffer->Lock(0,0,(void**)&VertexPointer,0);
		CopyMemory(VertexPointer,ShaderVertices,sizeof(ShaderVertices));
		ShaderStruct.buffer->Unlock();
		//Hardware shaders
		HANDLE InFile=CreateFileA("Data Files\\shaders\\default\\active.sav",GENERIC_READ,0,NULL,OPEN_EXISTING,0,NULL);
		if(InFile!=INVALID_HANDLE_VALUE) {
			BYTE b;
			DWORD read;
			DWORD count=0;
			char TexPath[256];
			while(true) {
				ReadFile(InFile,&b,1,&read,NULL);
				if(!read) break;
				if(count>255) { ERRORMSG("Shader name is greater than 256 characters long"); }
				TexPath[count++]=b;
				if(!b) {
					hr=D3DXCreateEffectFromFile(RealDevice,TexPath,0,0,0,0,
						&ShaderStruct.DefaultEffects[ShaderStruct.DefEffectCount],NULL);
					if(hr==D3D_OK) {
						ShaderStruct.DefaultEffects[ShaderStruct.DefEffectCount]->SetTechnique("T0");
						LoadShaderTextures(ShaderStruct.DefaultEffects[ShaderStruct.DefEffectCount]);
						ShaderStruct.DefEffectCount++;
					}
					count=0;
				}
			}
			CloseHandle(InFile);
		} else {
			Options.HWShader=false;
		}
		RealDevice->CreateStateBlock(D3DSBT_ALL,&StateBlock);
		//Create a font object for scribbling over the screen
		if(D3DXCreateFontA(RealDevice,12,0,400,1,false,ANSI_CHARSET,OUT_DEFAULT_PRECIS,
			DEFAULT_QUALITY,DEFAULT_PITCH|FF_DONTCARE,"arial",&Font)) {
				DEBUGSTRING("MGE: Could not create font."
					"If MGE is set to display messages, it will probably CTD shortly.");
		}
		//Create a sprite object for MGE HUD elements
		if(D3DXCreateSprite(RealDevice,&HUDsprite)!=D3D_OK) {
			DEBUGSTRING("MGE: Could not create HUD sprite."
				"Any mod that tries to add HUD elements will cause morrowind to CTD");
		}
		//if(Options.UseDistantLand) DL_Init();
		//Create and return the fake device
		fpsLastTime=GetTickCount();
		SETSTATUS(VERSION)6;
		*f=new FakeDevice(this);
		return D3D_OK;
	}
};

void* CreateD3DWrapper(UINT version) {
	//Load the options file
	HANDLE keyfile=0;
	keyfile=CreateFileA("MGE3\\settings",GENERIC_READ,0,0,OPEN_EXISTING,0,0);
	if(keyfile==INVALID_HANDLE_VALUE||keyfile==0) {
		DWORD error=GetLastError();
		char chars[512];
		char Message[256]; Message[0]='\0';
		_itoa_s(error,chars,512,10);
		strcat_s(chars,512,"\nCould not open the mge settings file\n");
		FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM,NULL,error,0,Message,256,NULL);
		strcat_s(chars,512,Message);
		strcat_s(chars,512,"Try running MGEgui.exe to recreate the save files.");
		ERRORMSG(chars);
		return 0;
	}
	DWORD unused;
	if(!ReadFile(keyfile,&Options,sizeof(Options),&unused,NULL)) {
		DWORD error=GetLastError();
		char chars[512];
		char Message[256]; Message[0]='\0';
		_itoa_s(error,chars,512,10);
		strcat_s(chars,512,"\nError reading the mge settings file\n");
		FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM,NULL,error,0,Message,256,NULL);
		strcat_s(chars,512,Message);
		strcat_s(chars,512,"Try running MGEgui.exe to recreate the save files.");
		ERRORMSG(chars);
		return 0;
	} else if(unused==0) {
		ERRORMSG("Unable to read the mge settings file as it appears to have 0 length.\n"
			"Try running MGEgui.exe to recreate the save files");
		return 0;
	}
	CloseHandle(keyfile);
	//Check that save file version is correct
	if(Options.V!=MGE_SAVE_VERSION) {
		ERRORMSG("The mge settings file is out of date.\n"
			"You need to run MGEgui.exe at least once to update the save files");
		return 0;
	}

	//Init texture hooks if using texture hooking
	if(Options.HookTextures||Options.UseDistantLand) InitHooks();

	//Local handle variables
	HMODULE d3ddll=NULL;
	D3DProc8 func=NULL;
	D3DProc9 func9=NULL;

	//Load the real d3d8 dll and get device caps
	char Path[MAX_PATH];
	GetSystemDirectoryA(Path,MAX_PATH);
	strcat_s(Path,MAX_PATH,"\\d3d8.dll");
	d3ddll=LoadLibraryA(Path);
	func=(D3DProc8)GetProcAddress(d3ddll,"Direct3DCreate8");
	if(func==NULL) {
		ERRORMSG("MGE: Unable to load real direct 3d dll");
		return 0;
	}
	IDirect3D8* TempObject = (func)(version);
	HRESULT hr=TempObject->GetDeviceCaps(0,D3DDEVTYPE_HAL,&d3d8Caps);
	if(hr!=D3D_OK) { ERRORMSG("Unable to get DX8 device caps") }
	TempObject->Release();
	FreeLibrary(d3ddll);

	//If using distant land, make sure that standard pixel shaders aren't in use
	if(Options.UseDistantLand) {
		d3d8Caps.PixelShaderVersion=0;
		d3d8Caps.VertexShaderVersion=0;
		UsingDistantLand=true;
	}

	//Load the dx9 dll and return
	d3ddll=LoadLibrary("d3d9.dll");
	func9=(D3DProc9)GetProcAddress(d3ddll,"Direct3DCreate9");
	DEBUGSTRING("MGE: The Direct3D8 device has been wrapped");
	return new FakeObject((func9)(D3D_SDK_VERSION));
}
